package org.openhabbo.catalog;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.items.Item;
import org.openhabbo.items.ItemManager;
import org.openhabbo.sessions.Session;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.users.User;
import org.openhabbo.users.UserManager;

public final class Catalog
{
	private Map<Integer, CatalogPage> pages;
	private Map<Integer, CatalogProduct> products;

	private Catalog()
	{
		this.pages = Collections.emptyMap();
		this.products = Collections.emptyMap();
	}

	public void reload()
	{
		// Create new collections
		Map<Integer, CatalogPage> tmpPages = new HashMap<Integer, CatalogPage>();
		Map<Integer, CatalogProduct> tmpProducts = new HashMap<Integer, CatalogProduct>();

		// Load all the pages
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM catalog_pages ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogPage page = new CatalogPage();
			page.id = result.getInt("id");
			page.parentId = result.getInt("parent_id");
			page.name = result.getString("name");
			page.requiredRight = result.getString("required_right");
			if(page.requiredRight == null || page.requiredRight.isEmpty())
			{
				page.requiredRight = "default";
			}
			page.isEnabled = result.getBoolean("enabled");
			page.isCategory = result.getBoolean("is_category");
			page.iconColor = result.getInt("icon_color");
			page.iconImage = result.getInt("icon_image");
			page.layout = result.getString("layout");
			page.headlineImage = result.getString("headline_image");
			page.teaserImage = result.getString("teaser_image");
			page.specialImage = result.getString("special_image");
			page.text1 = result.getString("text1");
			page.text2 = result.getString("text2");
			page.details = result.getString("details");
			page.teaserText = result.getString("teaser_text");
			tmpPages.put(page.id, page);
		}
		result.release();

		// Link the pages in parent/child fashion
		for(CatalogPage page : tmpPages.values())
		{
			if(page.parentId > 0)
			{
				CatalogPage parent = tmpPages.get(page.parentId);
				if(parent == null)
				{
					Logger.getInstance().warning("Catalog", String.format("Page #%d is a child of page #%d, which is missing. Page not attached to parent.", page.id, page.parentId));
				}
				else
				{
					parent.pages.add(page);
				}
			}
		}

		// Load the products, grab the item definitions and attach them to the pages
		result = OpenHABBO.getInstance().getDb().query("SELECT * FROM catalog_products ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogProduct product = new CatalogProduct();
			product.id = result.getInt("id");
			product.pageId = result.getInt("page_id");
			product.name = result.getString("name");
			product.itemDefinitionId = result.getInt("definition_id");
			product.itemAmount = result.getInt("amount");
			product.costCredits = result.getInt("cost_credits");
			product.costPixels = result.getInt("cost_pixels");
			product.costQuestPoints = result.getInt("cost_questpoints");
			product.itemDefinition = ItemManager.getInstance().getDefinition(product.itemDefinitionId);
			if(product.itemDefinition == null)
			{
				if(product.itemDefinitionId > 0)
				{
					Logger.getInstance().warning("Catalog", String.format("Product #%d ('%s') refers to the missing item definition #%d. Product ignored.", product.id, product.name, product.itemDefinitionId));
				}
			}
			else
			{
				if(product.pageId > 0)
				{
					CatalogPage page = tmpPages.get(product.pageId);
					if(page == null)
					{
						Logger.getInstance().warning("Catalog", String.format("Product #%d ('%s') is supposed to appear at page #%d, which is missing. Product not attached to page.", product.id, product.name, product.pageId));
					}
					else
					{
						product.page = page;
						page.products.add(product);
					}
				}
				tmpProducts.put(product.id, product);
			}
		}
		result.release();

		// Set instance fields
		this.pages = Collections.unmodifiableMap(tmpPages);
		this.products = Collections.unmodifiableMap(tmpProducts);

		// Done!
		Logger.getInstance().info("Catalog", String.format("Reloaded %d pages and %d products.", this.pages(), this.products()));
	}

	public CatalogPage getPage(int pageId)
	{
		return this.pages.get(pageId);
	}

	public List<CatalogPage> getVisibleIndexPagesForUser(User user)
	{
		List<CatalogPage> visible = new ArrayList<CatalogPage>();
		for(CatalogPage page : this.pages.values())
		{
			if(page.parentId == -1 && page.isVisibleTo(user))
			{
				visible.add(page);
			}
		}
		return visible;
	}

	public int pages()
	{
		return this.pages.size();
	}

	public CatalogProduct getProduct(int productId)
	{
		return this.products.get(productId);
	}

	public int products()
	{
		return this.products.size();
	}

	public void billUser(Session session, CatalogProduct product)
	{
		if(product.isPurchasableBy(session.getUser()))
		{
			// Payday!
			if(product.costCredits > 0)
			{
				session.getUser().credits -= product.costCredits;
				session.sendCreditBalance();
			}
			if(product.costPixels > 0)
			{
				session.getUser().pixels -= product.costPixels;
				session.sendPixelBalance();
			}
			session.getUser().saveCreditsAndPixels();

			// Write transaction log
			if(product.costCredits > 0)
			{
				UserManager.getInstance().logCreditTransaction(session.getUser().id, product.costCredits, "catalog product (#" + product.id + ")");
			}
		}
	}
	
	public List<Item> purchaseProduct(CatalogProduct product, User user, String purchaseData)
	{
		// Valid args?
		if(product == null || user == null || product.itemDefinition == null)
		{
			return null;
		}

		// Process purchase data
		try
		{
			purchaseData = product.itemDefinition.newInstance().processPurchaseDataToState(user, purchaseData);
		}
		catch(Exception ex)
		{
			purchaseData = Item.ERROR_DATA;
		}
		if(purchaseData != null && purchaseData.equals(Item.ERROR_DATA))
		{
			Logger.getInstance().warning("Catalog", String.format("Purchase of product #%d by user #%d failed, bad purchasedata.", product.id, user.id));
			return null;
		}

		// Create the shipping items
		List<Item> shipping = new ArrayList<Item>(1);
		for(int x = 0; x < product.itemAmount; x++)
		{
			/* 
			 * Create item instance, run purchase logic.
			 * Some products are actually not including items into the shipping, but performing services on the user etc
			 */
			Item item = product.itemDefinition.newInstance();
			item.ownerId = user.id;
			item.state = purchaseData;
			item.onPurchase(shipping);
		}    

		// These are the items that are shipped
		return shipping;
	}

	/* Begin singleton */
	private final static Catalog instance = new Catalog();

	public final static Catalog getInstance()
	{
		return instance;
	}
	/* End singleton */
}
