package org.openhabbo.items;

import org.openhabbo.OpenHABBO;
import org.openhabbo.net.messages.SerializableObject;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomTile;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public abstract class Item implements SerializableObject
{
	public int id;
	public ItemDefinition definition;
	public int ownerId;
	public int roomId;
	public int x;
	public int y;
	public float z;
	public byte rotation;
	public String wallPosition;
	public String customData;
	public int teleporterId;

	public RoomInstance room;

	private int pulseTimer;
	private boolean requiresDbUpdate;

	public boolean inRoom()
	{
		return(this.roomId > 0);
	}

	public boolean is(char flag)
	{
		return this.definition.is(flag);
	}

	public void requireDbUpdate()
	{
		requiresDbUpdate = true;
	}

	public void dbUpdated()
	{
		requiresDbUpdate = false;
	}

	public boolean requiresDbUpdate()
	{
		return requiresDbUpdate;
	}

	public boolean requiresPulse()
	{
		return(this.pulseTimer > 0);
	}

	protected boolean hasTimeout()
	{
		return this.requiresPulse();
	}

	protected void setTimeout(int pulses)
	{
		this.pulseTimer = pulses;
	}

	protected void clearTimeout()
	{
		this.pulseTimer = 0;
	}

	public void pulse()
	{
		// Pulse!
		this.pulseTimer--;
		this.onPulse();

		// End of timeout?
		if(this.pulseTimer <= 0)
		{
			this.clearTimeout();
			this.onTimeout();
		}
	}

	public void setPersistentState(String state)
	{
		this.customData = state;
		this.requireDbUpdate();
		this.setTemporaryState(state);
	}

	public void setTemporaryState(String state)
	{
		if(this.room != null)
		{
			ServerMessage msg = new ServerMessage();
			if(this.is(EItemBehaviour.FLOORITEM))
			{
				msg.init(EServerMessage.ObjectDataUpdate);
				msg.appendStr(Integer.toString(this.id));
				msg.appendStr(state);
			}
			else if(this.is(EItemBehaviour.WALLITEM))
			{
				// tmp swap variable
				String backup = this.customData;
				this.customData = state;

				msg.init(EServerMessage.ItemUpdate);
				msg.appendObject(this);

				this.customData = backup;
			}
			this.room.broadcast(msg);
		}
	}

	public RoomTile getTile()
	{
		return new RoomTile(this.x, this.y);
	}

	public RoomTile getTileInFront()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y--;
		}
		else if(this.rotation == 2)
		{
			x++;
		}
		else if(this.rotation == 4)
		{
			y++;
		}
		else if(this.rotation == 6)
		{
			x--;
		}
		return new RoomTile(x, y);
	}

	public RoomTile getTileBehind()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y++;
		}
		else if(this.rotation == 2)
		{
			x--;
		}
		else if(this.rotation == 4)
		{
			y--;
		}
		else if(this.rotation == 6)
		{
			x++;
		}
		return new RoomTile(x, y);
	}

	public void update()
	{
		if(this.is(EItemBehaviour.FLOORITEM))
		{
			OpenHABBO.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",x=" + this.x + ",y=" + this.y + ",z=" + this.z + ",rotation=" + this.rotation + ",customdata='" + SQLHelper.escape(this.customData) + "' WHERE id=" + this.id);
		}
		else if(this.is(EItemBehaviour.WALLITEM))
		{
			OpenHABBO.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",wallposition = '" + SQLHelper.escape(this.wallPosition) + "',customdata='" + SQLHelper.escape(this.customData) + "' WHERE id=" + this.id);
		}

		this.dbUpdated();
	}

	public void delete()
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM items WHERE id=" + this.id);
	}

	public boolean onPlace(Session session)
	{
		// Optionally overridden by impl class
		return true;
	}

	public boolean onPickup(Session session)
	{
		// Optionally overridden by impl class
		return true;
	}

	public void onDelete(Session session)
	{
		// Optionally overridden by impl class
	}

	public void onTrigger(Session session, int newState)
	{
		// Optionally overridden by impl class
	}

	protected void onPulse()
	{
		// Optionally overridden by impl class
	}

	protected void onTimeout()
	{
		// Optionally overridden by impl class
	}

	@Override
	public void serialize(ServerMessage msg)
	{
		if(this.is(EItemBehaviour.FLOORITEM))
		{
			msg.appendInt(this.id);
			msg.appendInt(this.definition.id);
			msg.appendInt(this.x);
			msg.appendInt(this.y);
			msg.appendInt(this.rotation);
			msg.appendStr(Float.toString(this.z).replace(',', '.'));
			msg.appendInt(0);
			msg.appendStr(this.customData);
			msg.appendInt(-1);
		}
		else if(this.is(EItemBehaviour.WALLITEM))
		{
			msg.appendStr(Integer.toString(this.id));
			msg.appendInt(this.definition.id);
			msg.appendStr(this.wallPosition);
			msg.appendStr(this.customData);
		}
	}

	public static Item parse(DatabaseResult row)
	{
		// Find definition
		ItemDefinition def = ItemManager.getInstance().getDefinition(row.getInt("definition_id"));
		if(def != null)
		{
			// Create instance
			Item item = def.newInstance();
			item.id = row.getInt("id");
			item.ownerId = row.getInt("owner_id");
			item.x = row.getInt("x");
			item.y = row.getInt("y");
			item.z = row.getFloat("z");
			item.rotation = row.getByte("rotation");
			item.wallPosition = row.getString("wallposition");
			item.customData = row.getString("customdata");
			if(item.parseExtraData(row))
			{
				return item;
			}
		}

		// Bad item
		return null;
	}

	protected boolean parseExtraData(DatabaseResult row)
	{
		// This method can be overridden by impl classes
		return true;
	}
}
