package org.openhabbo.rooms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public final class RoomNavigator
{
	private long lastRefreshTime;
	private final int maxAgeTime;

	private Map<Integer, RoomCategory> categories;
	private List<NavigatorFrontpageItem> frontpage;
	private final List<Room> busiestGuestRooms;
	private final List<Room> topRatedGuestRooms;

	private RoomNavigator(int maxAgeSeconds)
	{
		this.maxAgeTime = (maxAgeSeconds * 1000);
		this.categories = Collections.emptyMap();
		this.frontpage = Collections.emptyList();
		this.busiestGuestRooms = new ArrayList<Room>();
		this.topRatedGuestRooms = new ArrayList<Room>();
	}

	public void reload()
	{
		// Create new collections
		Map<Integer, RoomCategory> tmpCategories = new HashMap<Integer, RoomCategory>();
		List<NavigatorFrontpageItem> tmpFrontpage = new ArrayList<NavigatorFrontpageItem>();

		// Load categories from database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM navigator_categories ORDER BY ordernum ASC");
		while(result.row())
		{
			if(result.getBoolean("enabled"))
			{
				RoomCategory category = new RoomCategory();
				category.id = result.getInt("id");
				category.name = result.getString("name");
				category.requiredRight = result.getString("requiredright");
				if(category.requiredRight.isEmpty())
				{
					category.requiredRight = null;
				}
				tmpCategories.put(category.id, category);
			}
		}

		// Load frontpage items from database
		result.query("SELECT id,name,bannertype,type,inner_id,image,image_src,enabled FROM navigator_frontpage ORDER BY ordernum ASC");
		while(result.row())
		{
			// Enabled?
			if(result.getBoolean("enabled"))
			{
				// Parse entry
				NavigatorFrontpageItem fpi = new NavigatorFrontpageItem();
				fpi.id = result.getInt("id");
				fpi.name = result.getString("name");
				fpi.bannerType = result.getInt("bannertype");
				fpi.image = result.getString("image");
				fpi.imageIsExternal = result.getString("image_src").equals("external");

				// Get inner item
				String innerType = result.getString("type");
				int innerId = result.getInt("inner_id");
				if(innerType.equals("room"))
				{
					fpi.room = RoomManager.getInstance().getRoomInfo(innerId);
				}
				else
				{
					fpi.category = tmpCategories.get(innerId);
				}

				// Inner item exists? (don't enable invalid frontpage items)
				if(fpi.room != null || fpi.category != null)
				{
					tmpFrontpage.add(fpi);
				}
				else
				{
					Logger.getInstance().warning("RoomNavigator", String.format("Frontpage item #%d (\"%s\") refers to the missing %s #%d. Item ignored.", fpi.id, fpi.name, innerType, innerId));
				}
			}
		}
		result.release();

		// Set instance fields
		this.categories = Collections.unmodifiableMap(tmpCategories);
		this.frontpage = Collections.unmodifiableList(tmpFrontpage);

		// Done!
		Logger.getInstance().info("RoomNavigator", String.format("Reloaded %d guestroom categories and %d frontpage items.", this.categories(), this.frontpageItems()));
	}
	
	public void checkAge()
	{
		// Needs refresh?
		if((System.currentTimeMillis() - this.lastRefreshTime) > this.maxAgeTime)
		{
			this.refresh();
		}
	}
	
	public void refresh()
	{	
		// Refresh categories
		for(RoomCategory category : this.categories.values())
		{
			// Refresh rooms in this category
			category.rooms.clear();
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE category_id=" + category.id + " ORDER BY users_now DESC,rating DESC LIMIT 50");
			while(result.row())
			{
				category.rooms.add(Room.parse(result));
			}
			result.release();
		}
		
		// Refresh rooms that are on the frontpage
		for(NavigatorFrontpageItem fpi : this.frontpage)
		{
			// Is this a room?
			if(fpi.room != null)
			{
				// Get fresh room info, stick to the old info if the room has been deleted now
				Room newInfo = RoomManager.getInstance().getRoomInfo(fpi.room.id);
				if(newInfo != null)
				{
					fpi.room = newInfo;
				}
			}
		}
		
		// Refresh busiest rooms
		this.busiestGuestRooms.clear();
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id > 0 ORDER BY users_now DESC LIMIT 50");
		while(result.row())
		{
			this.busiestGuestRooms.add(Room.parse(result));
		}
		result.release();

		// Refresh toprated rooms
		this.topRatedGuestRooms.clear();
		result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id > 0 ORDER BY rating DESC LIMIT 50");
		while(result.row())
		{
			this.topRatedGuestRooms.add(Room.parse(result));
		}
		result.release();

		// This is for debug only
		this.lastRefreshTime = System.currentTimeMillis();
		Logger.getInstance().debug("RoomNavigator", "Refreshed successfully.");
	}

	public RoomCategory getCategory(int categoryId)
	{
		return this.categories.get(categoryId);
	}

	public int categories()
	{
		return this.categories.size();
	}

	public Collection<RoomCategory> getCategories()
	{
		return this.categories.values();
	}

	public int frontpageItems()
	{
		return this.frontpage.size();
	}

	public List<NavigatorFrontpageItem> getFrontpageItems()
	{
		return this.frontpage;
	}

	public List<Room> getBusiestGuestRooms()
	{
		return this.busiestGuestRooms;
	}

	public List<Room> getTopRatedGuestRooms()
	{
		return this.topRatedGuestRooms;
	}

	public List<Room> getGuestRoomSearchResult(String criteria, boolean byTags)
	{
		ArrayList<Room> rooms = new ArrayList<Room>();

		criteria = SQLHelper.escape(criteria);
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE u.username='" + criteria + "' OR r.name LIKE '" + criteria + "%' OR r.tags LIKE '%" + criteria + "%' LIMIT 30;");
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}

	/* Begin singleton */
	private static RoomNavigator instance;

	public final static RoomNavigator getInstance()
	{
		instance.checkAge();
		return instance;
	}

	public static void createInstance(int maxAgeSeconds)
	{
		instance = new RoomNavigator(maxAgeSeconds);
	}
	/* End singleton */
}
