package org.openhabbo.rooms.instances;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.AttachmentBox;
import org.openhabbo.core.Logger;
import org.openhabbo.events.Event;
import org.openhabbo.items.EItemBehaviour;
import org.openhabbo.items.Item;
import org.openhabbo.items.ItemManager;

import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.RoomModel;
import org.openhabbo.rooms.pathfinding.JoehPathfinderStaticImpl;
import org.openhabbo.rooms.pathfinding.PathfinderNode;
import org.openhabbo.rooms.pathfinding.RotationCalculator;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.storage.DatabaseConnection;
import org.openhabbo.util.ChatUtil;

public final class RoomInstance extends AttachmentBox
{
	/**
	 * Affecting this value will speed up movement of room users, bot AI, etc. The value is in milliseconds.
	 */
	public final static int PULSE_INTERVAL = 470;

	// Events
	public final static Event onLoad = new Event("RoomInstance.OnLoad");
	public final static Event onUnload = new Event("RoomInstance.OnUnload");
	public final static Event onSessionEnter = new Event("RoomInstance.OnSessionEnter");
	public final static Event onSessionExit = new Event("RoomInstance.OnSessionExit");

	private Room info;
	private RoomModel model;
	private RoomMap map;

	private int workerIndex;
	private int idleTimer;
	private boolean unloading;

	private Map<Integer, Item> items;

	private Map<Integer, Long> bans;
	private Map<Integer, String> rights;

	private int virtualIdCounter;
	private Map<Integer, RoomUser> users;
	private Map<String, RoomUser> usersByUsername;

	public RoomInstance(Room info, RoomModel model)
	{
		// Core room info
		this.info = info;
		this.model = model;
		this.map = new RoomMap(model.getMaxX(), model.getMaxY());

		// Worker
		this.workerIndex = -1;
		this.idleTimer = 0;
		this.unloading = false;

		// Content
		this.items = new ConcurrentHashMap<Integer, Item>(5);
		this.bans = new ConcurrentHashMap<Integer, Long>(0);
		this.rights = new ConcurrentHashMap<Integer, String>();

		// Users
		this.virtualIdCounter = 0;
		this.users = new ConcurrentHashMap<Integer, RoomUser>(5);
		this.usersByUsername = new ConcurrentHashMap<String, RoomUser>(5);
	}

	public final Room getInfo()
	{
		return this.info;
	}

	public final RoomModel getModel()
	{
		return this.model;
	}

	public final RoomMap getMap()
	{
		return this.map;
	}

	public final int getWorkerIndex()
	{
		return this.workerIndex;
	}

	public void setWorkerIndex(int index)
	{
		this.workerIndex = index;
	}

	public boolean requiresUnload()
	{
		final int maxIdleSeconds = 10;
		return(this.idleTimer > ((maxIdleSeconds * 1000) / RoomInstance.PULSE_INTERVAL));
	}

	@Override
	public final String toString()
	{
		return "RoomInstance #" + this.info.id;
	}

	public void load() throws Exception
	{
		// Reload bans & rights
		this.bans.clear();
		this.rights.clear();
		this.rights.putAll(RoomManager.getInstance().getRightsForRoom(this.info.id));

		// Reload items
		this.items.clear();
		for(Item item : ItemManager.getInstance().getItemsForRoom(this.info.id))
		{
			item.room = this;
			this.items.put(item.id, item);
		}

		// Reload other stuff
		RoomInstance.onLoad.fire(this);

		// Generate map
		this.generateFloorMap();
	}

	public void unload()
	{
		// Already unloading?
		if(this.unloading)
		{
			return;
		}

		// Flag that this room instance is unloading
		this.unloading = true;

		// Remove all users
		for(RoomUser user : this.users.values())
		{
			if(user.getSession() != null)
			{
				user.getSession().kickFromRoom();
			}
		}
		this.users.clear();
		this.usersByUsername.clear();
		this.rights.clear();
		this.bans.clear();
		if(this.info.usersNow > 0)
		{
			this.info.usersNow = 0;
			this.info.saveUserCount();
		}

		// Save & clear items
		this.saveChangedItems();
		this.items.clear();
	}

	private void saveChangedItems()
	{
		// Get dedicated database connection
		DatabaseConnection conn = OpenHABBO.getInstance().getDb().getConnMgr().getConnection();
		if(conn == null)
		{
			return;
		}

		// Prepare statement for updating items, batch everything and execute
		try
		{
			PreparedStatement query = conn.getInnerConnection().prepareStatement("UPDATE items SET x=?,y=?,z=?,rotation=?,wallposition=?,customdata=? WHERE id=?");
			for(Item item : this.items.values())
			{
				if(item.requiresDbUpdate())
				{
					query.setInt(1, item.x);
					query.setInt(2, item.y);
					query.setFloat(3, item.z);
					query.setByte(4, item.rotation);
					query.setString(5, item.wallPosition);
					query.setString(6, item.customData);
					query.setInt(7, item.id);
					query.addBatch();
					item.dbUpdated();
				}
			}
			query.executeBatch();
			query.close();
		}
		catch(SQLException ex)
		{
			Logger.getInstance().error(this, "Error saving changed items.", ex);
		}

		// Release the dedicated database connection
		OpenHABBO.getInstance().getDb().getConnMgr().releaseConnection(conn);
	}

	public RoomUser getUserByVirtualId(int virtualId)
	{
		for(RoomUser user : this.users.values())
		{
			if(user.virtualId == virtualId)
			{
				return user;
			}
		}
		return null;
	}

	public RoomUser getUserByUserId(int userId)
	{
		return this.users.get(userId);
	}

	public RoomUser getUserByUsername(String username)
	{
		return this.usersByUsername.get(username.toLowerCase());
	}

	public RoomUser getUserOnTile(int x, int y)
	{
		for(RoomUser user : this.users.values())
		{
			if(user.x == x && user.y == y)
			{
				return user;
			}
		}
		return null;
	}

	public void refreshUserOnTile(int x, int y)
	{
		RoomUser user = this.getUserOnTile(x, y);
		if(user != null)
		{
			// Remove old statuses and calculate new
			user.removeInteractiveStatuses();
			this.applyTileToUser(user);

			// Immediate refresh (no waiting on roominstance pulse)
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(1);
			user.appendStatusString(msg);
			this.broadcast(msg);
		}
	}

	public Collection<RoomUser> getAllUsers()
	{
		return this.users.values();
	}

	public RoomUser addSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.unloading)
		{
			return null;
		}

		// Create roomuser with unique virtual ID
		RoomUser user = new RoomUser(this.virtualIdCounter++, this);
		user.setSession(session);
		user.setUserInfo(session.getUser());

		// Position user at spawn position
		if(session.authenticatedTeleporterId == 0)
		{
			user.x = this.model.doorX;
			user.y = this.model.doorY;
			user.z = this.model.doorZ;
			user.headRotation = user.bodyRotation = this.model.doorRotation;
		}
		else
		{
			// Teleporter exists?
			Item tele = this.getItem(session.authenticatedTeleporterId);
			if(tele == null)
			{
				return null;
			}
			else
			{
				// Spawn in teleporter
				user.x = user.x;
				user.y = user.y;
				user.z = user.z;
				session.authenticatedTeleporterId = 0;

				// Broadcast activity
				// this.getInteractor().broadcastTeleporterActivity(obj.ID, obj.definition.sprite, user.getUserObject().name, false);
			}
		}

		// Determine privileges if in guestroom
		if(this.info.isGuestRoom())
		{
			boolean isAnyRoomOwner = user.getUser().role.hasRight("is_any_room_owner");
			user.hasRoomOwner = (isAnyRoomOwner || (user.getUser().id == this.info.ownerId));
			user.hasRoomRights = (user.hasRoomOwner || this.rights.containsKey(user.getUser().id));
			if(user.hasRoomRights)
			{
				user.addStatus("flatctrl", user.hasRoomOwner ? "useradmin" : null);
			}
		}

		// Add to user collection
		this.users.put(user.getUser().id, user);
		this.usersByUsername.put(user.getUser().username.toLowerCase(), user);

		// Announce to other users
		ServerMessage msg = new ServerMessage(EServerMessage.Users);
		msg.appendInt(1); // 1 item
		msg.appendObject(user);
		this.broadcast(msg);

		// Update current user count
		synchronized(this.info)
		{
			this.info.usersNow++;
			this.info.saveUserCount();
		}

		// Done!
		Logger.getInstance().debug(this, String.format("User #%d (%s) entered.", user.getUser().id, user.getUser().username));
		return user;
	}

	public void removeSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.unloading)
		{
			return;
		}

		// Remove user from collections
		RoomUser user = this.users.remove(session.getUser().id);
		this.usersByUsername.remove(session.getUser().username.toLowerCase());

		// Actually user removed?
		if(user != null)
		{
			// Invalidate user for any references floating around
			user.invalidate();

			// Unblock map
			this.map.users[user.x][user.y] = false;

			// Broadcast leaving
			ServerMessage msg = new ServerMessage(EServerMessage.UserRemove);
			msg.appendStr(Integer.toString(user.virtualId));
			this.broadcast(msg);

			// Update current user count
			synchronized(this.info)
			{
				this.info.usersNow--;
				this.info.saveUserCount();
			}

			// Exited successfully!
			Logger.getInstance().debug(this, String.format("User #%d (%s) exited.", user.getUser().id, user.getUser().username));
		}
	}

	public boolean ringDoorbell(Session ringing)
	{
		boolean heard = false;
		ServerMessage msg = null;
		for(RoomUser user : this.users.values())
		{
			if(user.getSession() != null && user.hasRoomRights)
			{
				// Notify of doorbelling user
				if(msg == null)
				{
					msg = new ServerMessage(EServerMessage.Doorbell);
					msg.appendStr(ringing.getUser().username);
				}
				user.getSession().send(msg);
				heard = true;
			}
		}

		// True if atleast one user heard the doorbell and is allowed to answer it
		return heard;
	}

	public void answerDoorbellAndAllowEnter(String username)
	{
		Session session = SessionManager.getInstance().getSessionByUsername(username);
		if(session != null && session.waitingForDoorbell == this.info.id)
		{
			session.waitingForDoorbell = 0;
			session.goToRoom(this.info.id);
		}
	}

	public void moderationKick(int issuerId, String message)
	{
		// Maintain collection with users to kick, expect none
		List<RoomUser> toKick = new ArrayList<RoomUser>(0);

		// Don't kick the issuer
		for(RoomUser user : this.usersByUsername.values())
		{
			if(user.getSession() != null && user.getUser().id != issuerId)
			{
				toKick.add(user);
			}
		}

		// Kick the targets
		for(RoomUser user : toKick)
		{
			user.getSession().kickFromRoom(message);
		}
	}

	public void broadcast(ServerMessage msg)
	{
		this.broadcast(msg.getResult());
	}

	public void broadcast(byte[] data)
	{
		Logger.getInstance().debug(this, "--> " + new String(data));
		for(RoomUser user : this.users.values())
		{
			if(user.getSession() != null)
			{
				user.getSession().getConnection().send(data);
			}
		}
	}

	public void addRoomBan(int userId, int seconds)
	{
		long expireTime = System.currentTimeMillis() + (seconds * 1000);
		this.bans.put(userId, expireTime);
	}

	public void removeRoomBan(int userId)
	{
		this.bans.remove(userId);
	}

	public boolean userIsBanned(int userId)
	{
		Long expireTime = this.bans.get(userId);
		if(expireTime != null)
		{
			if(System.currentTimeMillis() < expireTime)
			{
				return true;
			}
			else
			{
				this.removeRoomBan(userId);
			}
		}
		return false;
	}

	public boolean addRights(int userId, String username)
	{
		if(this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.put(userId, username);
			RoomManager.getInstance().addRoomRights(this.info.id, userId);
			return true;
		}
	}

	public boolean removeRights(int userId)
	{
		if(!this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.remove(userId);
			RoomManager.getInstance().removeRoomRights(this.info.id, userId);
			return true;
		}
	}

	public boolean userHasRights(int userId)
	{
		return this.rights.containsKey(userId);
	}

	public Set<Entry<Integer, String>> getUsersWithRights()
	{
		return this.rights.entrySet();
	}

	public void chat(RoomUser source, String text, boolean shout)
	{
		// Broadcast chat to room
		ServerMessage msg = new ServerMessage(shout ? EServerMessage.Shout : EServerMessage.Chat);
		msg.appendInt(source.virtualId);
		msg.appendStr(text);
		msg.appendInt(ChatUtil.getChatGesture(text));
		this.broadcast(msg);

		// Rotate all heads to the source of chat (if possible)
		for(RoomUser user : this.users.values())
		{
			if(user.virtualId != source.virtualId)
			{
				user.lookTo(source.x, source.y, true);
			}
		}

		// TODO: trigger bot/pet responses etc
	}

	private void generateFloorMap()
	{
		// Get default floormap from room model
		char[][] modelMap = this.model.getMap();

		// Determine map boundaries
		int maxX = modelMap.length;
		int maxY = modelMap[0].length;

		// Create default floor map
		for(int y = 0; y < maxY; y++)
		{
			for(int x = 0; x < maxX; x++)
			{
				// Grab tile
				char tile = modelMap[x][y];

				// Determine tile
				if(tile != 'x' && Character.isDigit(tile))
				{
					this.map.state[x][y] = RoomMap.OPEN;
					this.map.height[x][y] = Float.parseFloat(Character.toString(tile));
				}
				else
				{
					this.map.state[x][y] = RoomMap.CLOSED;
					this.map.height[x][y] = 0.0f;
				}
			}
		}

		// Apply floor items to map
		for(Item item : this.items.values())
		{
			if(item.is(EItemBehaviour.FLOORITEM))
			{
				this.applyFloorItemToMaps(item, this.map.state, this.map.height, null);
			}
		}

		// Restore door tile
		this.map.state[this.model.doorX][this.model.doorY] = RoomMap.OPEN;
		this.map.height[this.model.doorX][this.model.doorY] = this.model.doorZ;
	}

	private void applyFloorItemToMaps(Item obj, byte[][] tmpMap, float[][] tmpHeightmap, RoomTile[][] tmpRedirectMap)
	{
		// TRIGGERS ARE SIMPLE. K.
		if(obj.is(EItemBehaviour.TRIGGER))
		{
			tmpMap[obj.x][obj.y] = RoomMap.INTERACTIVE;
			return;
		}

		for(RoomTile tile : RoomMap.getAffectedTiles(obj, true))
		{
			// Out of range?
			if(!(tile.x >= 0 && tile.x < tmpMap.length && tile.y >= 0 && tile.y < tmpMap[0].length))
			{
				Logger.getInstance().warning(this, String.format("Floor item #%d at tile [%d,%d] was outside the boundaries of the map. Floor item not applied to maps.", obj.id, obj.x, obj.y));
				return;
			}

			// Increase heightmap
			if(obj.z >= tmpHeightmap[tile.x][tile.y])
			{
				tmpHeightmap[tile.x][tile.y] = obj.z;
			}

			// Is this object a door?
			if(obj.is(EItemBehaviour.DOOR))
			{
				// Is the door open?
				if(obj.customData != null && obj.customData.equals("O"))
				{
					tmpMap[tile.x][tile.y] = RoomMap.OPEN;
				}
				else
				{
					tmpMap[tile.x][tile.y] = RoomMap.CLOSED;
				}

				// Doors haven't got further handling
				return;
			}

			// Is this object the top item on the tile?
			if((obj.z >= tmpHeightmap[tile.x][tile.y]))
			{
				if(obj.is(EItemBehaviour.CAN_STAND_ON_TOP))
				{
					if(obj.z > tmpHeightmap[tile.x][tile.y])
					{
						tmpMap[tile.x][tile.y] = RoomMap.OPEN;
						tmpHeightmap[tile.x][tile.y] = (obj.z + obj.definition.height);
					}
				}
				else if(obj.is(EItemBehaviour.CAN_SIT_ON_TOP))
				{
					tmpMap[tile.x][tile.y] = RoomMap.INTERACTIVE;
				}
				else if(obj.is(EItemBehaviour.CAN_LAY_ON_TOP))
				{
					tmpMap[tile.x][tile.y] = RoomMap.INTERACTIVE;

					// Redirect to top
					if(obj.rotation == 0 || obj.rotation == 4)
					{
						tmpRedirectMap[tile.x][tile.y] = new RoomTile(tile.x, obj.y);
					}
					else
					{
						tmpRedirectMap[tile.x][tile.y] = new RoomTile(obj.x, tile.y);
					}
				}
				else
				{
					// Block the tile if there is no interactive object on top
					if(tmpMap[tile.x][tile.y] != RoomMap.INTERACTIVE)
					{
						tmpMap[tile.x][tile.y] = RoomMap.CLOSED;
					}
				}
			}
		}
	}

	public Item getItem(int itemId)
	{
		return this.items.get(itemId);
	}

	public List<Item> getItems(char behaviourFlag)
	{
		List<Item> items = new ArrayList<Item>();
		for(Item item : this.items.values())
		{
			if(item.is(behaviourFlag))
			{
				items.add(item);
			}
		}
		return items;
	}

	public List<Item> getFloorItemsOnTile(int X, int Y)
	{
		List<Item> objs = new ArrayList<Item>(0);
		for(Item obj : this.items.values())
		{
			if(obj.is(EItemBehaviour.FLOORITEM))
			{
				for(RoomTile tile : RoomMap.getAffectedTiles(obj, true))
				{
					if(tile.x == X && tile.y == Y)
					{
						objs.add(obj);
					}
				}
			}
		}
		return objs;
	}

	public boolean placeFloorItem(Item item, int x, int y, Session session)
	{
		return this.moveFloorItem(item, x, y, (byte) 0, true, session);
	}

	public boolean moveFloorItem(Item item, int x, int y, byte rotation, Session session)
	{
		return this.moveFloorItem(item, x, y, rotation, false, session);
	}

	private final boolean moveFloorItem(Item item, int newX, int newY, byte newRotation, boolean isNew, Session session)
	{
		// Valid item? (if new, should not be in room), (if not new, should be in room)
		if(item == null || ((isNew && this.items.containsKey(item.id) || !isNew && !this.items.containsKey(item.id))))
		{
			return false;
		}

		// Validate new rotation, this can really trouble client and server!
		if(newRotation != 0 && newRotation != 2 && newRotation != 4 && newRotation != 6)
		{
			return false;
		}

		// Get the tiles this object would reside on if this item was placed here, check them for 'freeness'
		List<RoomTile> newTiles = RoomMap.getAffectedTiles(item.definition.length, item.definition.width, newX, newY, newRotation, true);
		for(RoomTile tile : newTiles)
		{
			// Out of map range / bad tile?
			if(!this.map.exists(tile.x, tile.y))
			{
				return false;
			}

			// Is there a SpaceUser on this tile?
			if(this.getUserOnTile(tile.x, tile.y) != null)
			{
				// Can rotate on different tile than root?
				if(newRotation == item.rotation && (tile.x != item.x || tile.y != item.y))
				{
					return false;
				}
			}
		}

		// Calculate new height
		float newZ = this.map.height[newX][newY];

		// Trying to stack on itself? (by rotating an object that is already in the space)
		if(!isNew)
		{
			if(item.rotation == newRotation && item.x == newX && item.y == newY && item.z != newZ)
			{
				return false;
			}
		}

		// Now check all the new tiles to determine new height
		for(RoomTile tile : newTiles)
		{
			// Get the objects that are 'leaning' on this tile somehow
			List<Item> itemsOnTile = this.getFloorItemsOnTile(tile.x, tile.y);

			// Check those objects and determine top height
			for(Item tileItem : itemsOnTile)
			{
				// Is this not the object that is being moved now?
				if(tileItem.id != item.id)
				{
					// Is this the top item?
					if((tileItem.z + tileItem.definition.height) > newZ)
					{
						// Can stack on top?
						if(tileItem.is(EItemBehaviour.CAN_STACK_ON_TOP))
						{
							newZ = tileItem.z + tileItem.definition.height;
						}
						else
						{
							return false;
						}
					}
				}
			}
		}

		// Max stacking height
		final int maxItemZ = 15;
		newZ = (newZ > maxItemZ) ? maxItemZ : newZ;

		// Trigger the interactor
		if(!item.onPlace(session))
		{
			return false;
		}

		// Backup the old position
		int oldX = item.x;
		int oldY = item.y;
		byte oldRotation = item.rotation;

		// Set object on new position
		item.x = newX;
		item.y = newY;
		item.z = newZ;
		item.rotation = newRotation;

		// If item is new to room...
		if(isNew)
		{
			item.room = this;
			item.roomId = this.info.id;
			item.ownerId = this.info.ownerId;
			item.update();
			this.items.put(item.id, item);
		}
		else
		{
			item.requireDbUpdate();
		}

		// Broadcast placement
		ServerMessage msg = new ServerMessage(isNew ? EServerMessage.ObjectAdd : EServerMessage.ObjectUpdate);
		msg.appendObject(item);
		this.broadcast(msg);

		// Generate the new map
		this.generateFloorMap();

		// Restore RoomUsers on old RoomTiles?
		if(!isNew)
		{
			for(RoomTile tile : RoomMap.getAffectedTiles(item.definition.length, item.definition.width, oldX, oldY, oldRotation, true))
			{
				this.refreshUserOnTile(tile.x, tile.y);
			}
		}

		// Refresh RoomUsers on new RoomTiles
		for(RoomTile tile : RoomMap.getAffectedTiles(item, true))
		{
			this.refreshUserOnTile(tile.x, tile.y);
		}

		// Move OK!
		return true;
	}

	public boolean placeWallItem(Item item, String position, Session session)
	{
		// Valid item?
		if(item == null || this.items.containsKey(item.id))
		{
			return false;
		}

		// Trigger the interactor
		if(!item.onPlace(session))
		{
			return false;
		}

		// Update the item with new info
		item.room = this;
		item.roomId = this.info.id;
		item.ownerId = this.info.ownerId;
		item.wallPosition = position;
		item.update();
		this.items.put(item.id, item);

		// Broadcast update
		ServerMessage msg = new ServerMessage(EServerMessage.ItemAdd);
		msg.appendObject(item);
		this.broadcast(msg);

		// Move OK!
		return true;
	}

	public Item pickupItem(int itemId, Session session)
	{
		// Find the item, run pre-pickup checks
		Item item = this.getItem(itemId);
		if(item == null || !item.onPickup(session))
		{
			return null;
		}

		// Remove from collection
		this.items.remove(item.id);

		// Notify clients of removal
		ServerMessage msg = new ServerMessage();
		if(item.is(EItemBehaviour.FLOORITEM))
		{
			msg.init(EServerMessage.ObjectRemove);
			msg.appendStr(Integer.toString(item.id));
			msg.appendStr("");
			msg.appendBool(false);
		}
		else if(item.is(EItemBehaviour.WALLITEM))
		{
			msg.init(EServerMessage.ItemRemove);
			msg.appendStr(Integer.toString(item.id));
			msg.appendBool(false);
		}
		this.broadcast(msg);

		// Update floormaps if flooritem
		if(item.is(EItemBehaviour.FLOORITEM))
		{
			// Re-generate floormap
			this.generateFloorMap();

			// Refresh users on tile
			for(RoomTile tile : RoomMap.getAffectedTiles(item, true))
			{
				this.refreshUserOnTile(tile.x, tile.y);
			}
		}

		// Clear item
		item.room = null;
		item.roomId = 0;
		item.ownerId = 0;
		item.x = 0;
		item.y = 0;
		item.z = 0;
		item.rotation = 0;
		item.wallPosition = null;

		// Picked up successfully!
		return item;
	}

	public void pulse()
	{
		// Pulse the items, process timed items, etc
		this.pulseItems();

		// Pulse the users, kick idle sessions, etc
		this.pulseUsers();

		// Handle games? Or something? LOL!
	}

	private void pulseItems()
	{
		for(Item item : this.items.values())
		{
			if(item.requiresPulse())
			{
				item.pulse();
			}
		}
	}

	private void pulseUsers()
	{
		// These values are dealt with at the end of the pulse
		int updates = 0;
		boolean hasSessions = false;
		Session sessionInDoor = null;

		// Process all users (sessions/bots/pets
		for(RoomUser user : this.users.values())
		{
			// User with session?
			if(user.getSession() != null)
			{
				hasSessions = true;
			}

			// Any updates required?
			if(user.requiresUpdate)
			{
				// Atleast one update to be broadcasted
				updates++;
				
				// User in need of movement:phase1?
				if(user.isMoving())
				{
					// Destination reached?
					if(user.x == user.goalX && user.y == user.goalY)
					{
						// Trip complete!
						user.clearMovement();

						// Is SpaceUser on an interactive tile?
						if(this.map.state[user.x][user.y] == RoomMap.INTERACTIVE)
						{
							this.applyTileToUser(user);
						}
						else if(user.x == this.model.doorX && user.y == this.model.doorY)
						{
							sessionInDoor = user.getSession();
						}
					}
					else
					{
						// Determine next tile in path
						PathfinderNode next = null;
						if(user.overrideNextTile)
						{
							// Override used!
							user.overrideNextTile = false;

							// Substitute the next step
							next = new PathfinderNode();
							next.X = user.goalX;
							next.Y = user.goalY;
						}
						else
						{
							// Start new path?
							if(user.startNewTrip)
							{
								user.startNewTrip = false;
								this.startUserMovement(user, user.goalX, user.goalY, user.overrideNextTile);
							}

							// Pop the next tile from the path
							if(user.path.size() > 0)
							{
								next = user.path.remove(0);
								if(next == null || this.map.state[next.X][next.Y] == RoomMap.CLOSED)
								{
									// Invalid tile: tile blocked right now
								}
								else if(this.map.users[next.X][next.Y])
								{
									// Invalid tile: user here
									next = null;
								}
								else
								{
									final boolean CROSS_DIAGONAL = true;
									if(user.path.size() > 1 && CROSS_DIAGONAL)
									{
										// Peek + test next node (already popped one, so this is like path[1])
										PathfinderNode node = user.path.get(0);
										if(this.map.walkable(node.X, node.Y))
										{
											// Can't skip more than one tile!
											if(!(Math.abs(user.x - node.X) > 1 || Math.abs(user.y - node.Y) > 1))
											{
												byte X1 = 0;
												byte X2 = 0;
												if(node.X > user.x)
												{
													X1 = -1;
													X2 = 1;
												}
												else
												{
													X1 = 1;
													X2 = -1;
												}

												// Valid shortcut?
												if(this.map.walkable(node.X + X1, node.Y) && this.map.walkable(user.x + X2, user.y))
												{
													next = user.path.remove(0);
												}
											}
										}
									}
								}
							}
						}

						// Is there a next tile available?
						if(next == null)
						{
							// Retry?
							if(++user.movementTries < 3)
							{
								user.requestNewTrip(user.goalX, user.goalY);
							}
							else
							{
								user.clearMovement();
							}

							// Is RoomUser on an interactive tile?
							if(this.map.state[user.x][user.y] == RoomMap.INTERACTIVE)
							{
								this.applyTileToUser(user);
							}
						}
						else
						{
							// Moved a tile
							user.movementTries = 0;

							// Set next coordinate for the next phase
							user.nextX = next.X;
							user.nextY = next.Y;
							user.nextZ = this.map.height[next.X][next.Y];
							user.requiresUpdate = true;
						}
					}
				}
			}
		}

		// Any updates? This includes simple status updates, but also movement:phase2
		if(updates > 0)
		{
			// Build the status update
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(updates);
			for(RoomUser user : this.users.values())
			{
				// This user was one of the batch that required an update?
				if(user.requiresUpdate)
				{
					// Simple status update, or movement:phase2?
					if(user.nextX == -1)
					{
						// Broadcast status, and done!
						user.appendStatusString(msg);
						user.requiresUpdate = false;
					}
					else
					{
						// movement:phase2 for this user, pronto
						
						// Rotate users body
						byte rotation = RotationCalculator.calculateMoveDirection(user.x, user.y, user.nextX, user.nextY);

						// Apply rotation
						user.headRotation = rotation;
						user.bodyRotation = rotation;

						// Remove interactive statuses
						user.removeInteractiveStatuses();

						// Add the movement status for the status update
						user.statuses.put("mv", user.nextX + "," + user.nextY + "," + Float.toString(user.nextZ));
						user.appendStatusString(msg);
						user.statuses.remove("mv");

						// Unblock & block
						this.map.users[user.x][user.y] = false;
						this.map.users[user.nextX][user.nextY] = (this.info.allowWalkthrough ? false : true); //!walkthrough
						
						// Update position
						user.x = user.nextX;
						user.y = user.nextY;
						user.z = user.nextZ;
						user.nextX = -1;
					}
				}
			}
			
			// Broadcast the status update
			this.broadcast(msg);
		}

		// Update idle timer, kick session in door
		if(hasSessions)
		{
			this.idleTimer = 0;
			if(sessionInDoor != null)
			{
				sessionInDoor.kickFromRoom();
			}
		}
		else
		{
			this.idleTimer++;
		}
	}

	private void startUserMovement(RoomUser usr, int goalX, int goalY, boolean overrideNext)
	{
		// Tile exists?
		if(!this.map.exists(goalX, goalY))
		{
			// WTF TILE AINT ON MAP?!
			return;
		}

		// Override?
		if(overrideNext)
		{
			usr.goalX = goalX;
			usr.goalY = goalY;
			usr.overrideNextTile = true;
		}
		else
		{
			// One move to seat etc?
			boolean interactiveTile = false;
			if(this.map.state[goalX][goalY] == RoomMap.INTERACTIVE)
			{
				if(!this.map.users[goalX][goalY])
				{
					interactiveTile = true;
					this.map.state[goalX][goalY] = RoomMap.OPEN; // Free for one tick
				}
			}

			// Find path using the beloved Joe 'Joeh' Hegarty 2D A* pathfinder
			boolean noDescLimit = this.info.isGuestRoom();
			List<PathfinderNode> path = JoehPathfinderStaticImpl.findPath(usr.x, usr.y, goalX, goalY, this.map.state, this.map.users, this.map.height, 1.5f, noDescLimit ? 100.0f : 2.0f);

			// Restore map
			if(interactiveTile)
			{
				this.map.state[goalX][goalY] = RoomMap.INTERACTIVE;
			}

			// Is there a path found?
			if(path.size() > 0)
			{
				// Remove current tile node
				path.remove(0);

				// Set this path and destination to space user
				usr.path = path;
				usr.goalX = goalX;
				usr.goalY = goalY;
			}
		}
	}

	private void applyTileToUser(RoomUser user)
	{
		for(Item obj : this.getFloorItemsOnTile(user.x, user.y))
		{
			if(applyFloorItemToUser(user, obj))
			{
				return;
			}
		}
	}

	private boolean applyFloorItemToUser(RoomUser user, Item obj)
	{
		if(obj.is(EItemBehaviour.CAN_SIT_ON_TOP))
		{
			user.z = obj.z;
			user.headRotation = obj.rotation;
			user.bodyRotation = obj.rotation;
			user.danceId = 0;
			user.addStatus("sit", Float.toString(obj.z + obj.definition.height));
			return true;
		}

		return false;
	}
}
