package org.openhabbo.rooms.instances;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import org.openhabbo.core.AttachmentBox;
import org.openhabbo.net.messages.SerializableObject;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.pathfinding.PathfinderNode;
import org.openhabbo.rooms.pathfinding.RotationCalculator;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.users.User;

public final class RoomUser extends AttachmentBox implements SerializableObject
{
	public RoomInstance room;
	private boolean invalidated;

	public final int virtualId;
	public int x;
	public int y;
	public float z;
	public byte headRotation;
	public byte bodyRotation;
	public HashMap<String, String> statuses;
	public boolean requiresUpdate;

	public int idleTime;
	public boolean isAsleep;
	public int danceId;
	public int carryItemId;
	public int carryItemTimer;
	// public int effectId = -1;

	private User userInfo;

	public int goalX;
	public int goalY;
	public int nextX;
	public int nextY;
	public float nextZ;
	public byte movementTries;
	private boolean moveLock;
	public boolean overrideNextTile;
	public List<PathfinderNode> path;
	public boolean startNewTrip;

	// 'Real' users (not bot/pet)
	private Session session;
	public boolean hasRoomOwner;
	public boolean hasRoomRights;

	public RoomUser(int virtualId)
	{
		this.virtualId = virtualId;
		this.statuses = new HashMap<String, String>();
		this.path = new ArrayList<PathfinderNode>();
		this.goalX = this.nextX = -1;
	}

	public void invalidate()
	{
		this.invalidated = true;
	}

	public boolean invalidated()
	{
		return this.invalidated;
	}

	public Session getSession()
	{
		return this.session;
	}

	public void setSession(Session session)
	{
		this.session = session;
	}

	public User getUser()
	{
		return this.userInfo;
	}

	public void setUserInfo(User info)
	{
		this.userInfo = info;
	}

	public boolean at(RoomTile tile)
	{
		return this.at(tile.x, tile.y);
	}

	public boolean at(int x, int y)
	{
		return(this.x == x && this.y == y);
	}

	public boolean touches(RoomTile tile)
	{
		return this.touches(tile.x, tile.y);
	}

	public boolean touches(int x, int y)
	{
		return RoomMap.tilesTouch(this.x, this.y, x, y);
	}

	public void requestNewTrip(RoomTile goal)
	{
		this.requestNewTrip(goal.x, goal.y);
	}

	public void requestNewTrip(int goalX, int goalY)
	{
		if(!(goalX == this.x && goalY == this.y))
		{
			if(this.path.size() > 0) this.path.clear();
			this.goalX = goalX;
			this.goalY = goalY;
			this.nextX = this.nextY = -1;
			this.requiresUpdate = startNewTrip = true;
		}
	}

	public void clearMovement()
	{
		if(this.path.size() > 0) this.path.clear();
		this.movementTries = 0;
		this.goalX = this.goalY = this.nextX = this.nextY = -1;
	}

	public boolean canMove()
	{
		return !this.moveLock;
	}

	public void lockMovement()
	{
		this.moveLock = true;
		this.clearMovement();
	}

	public void unlockMovement()
	{
		this.moveLock = false;
	}

	public void addStatus(String key)
	{
		this.statuses.put(key, null);
	}

	public void addStatus(String key, String value)
	{
		this.statuses.put(key, value);
		this.requiresUpdate = true;
	}

	public void removeStatus(String key)
	{
		this.statuses.remove(key);
		this.requiresUpdate = true;
	}

	public void removeInteractiveStatuses()
	{
		this.statuses.remove("sit");
		this.statuses.remove("lay");
	}

	public void lookTo(int tileX, int tileY, boolean lockBody)
	{
		if(this.nextX != -1 || (tileX == this.x && tileY == this.y) || this.statuses.containsKey("lay"))
		{
			return;
		}

		byte rotation = RotationCalculator.calculateHumanDirection(this.x, this.y, tileX, tileY);
		int diff = this.bodyRotation - rotation;
		this.headRotation = this.bodyRotation;

		if(lockBody || this.statuses.containsKey("sit"))
		{
			if(this.bodyRotation == 2 || this.bodyRotation == 4)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
			else if(this.bodyRotation == 0 || this.bodyRotation == 6)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
		}
		else if(diff <= -2 || diff >= 2)
		{
			this.headRotation = rotation;
			this.bodyRotation = rotation;
		}
		else
		{
			this.headRotation = rotation;
		}
		this.requiresUpdate = true;
	}

	public void carryItem(int itemId)
	{
		// Set the ID + timer
		this.carryItemId = itemId;
		if(itemId > 0)
		{
			this.carryItemTimer = 240;
		}
		else
		{
			this.carryItemTimer = 0;
		}

		// Notify the roominstance
		ServerMessage msg = new ServerMessage(EServerMessage.CarryObject);
		msg.appendInt(this.virtualId);
		msg.appendInt(this.carryItemId);
		this.room.broadcast(msg);
	}

	public void appendStatusString(ServerMessage msg)
	{
		msg.appendInt(this.virtualId);
		msg.appendInt(this.x);
		msg.appendInt(this.y);
		msg.appendStr(Float.toString(this.z).replace(',', '.'));
		msg.appendInt(this.headRotation);
		msg.appendInt(this.bodyRotation);
		msg.append('/');
		for(Entry<String, String> status : this.statuses.entrySet())
		{
			msg.append(status.getKey());
			if(status.getValue() != null)
			{
				msg.append(' ');
				msg.append(status.getValue());
			}
			msg.append('/');
		}
		msg.appendStr("/");
	}

	@Override
	public void serialize(ServerMessage msg)
	{
		msg.appendInt(this.userInfo.id);
		msg.appendStr(this.userInfo.username);
		msg.appendStr(this.userInfo.motto);
		msg.appendStr(this.userInfo.figure);
		msg.appendInt(this.virtualId);
		msg.appendInt(this.x);
		msg.appendInt(this.y);
		msg.appendStr(Float.toString(this.z).replace(',', '.'));
		msg.appendInt(2);
		msg.appendInt(1);
		msg.appendStr(Character.toString(this.userInfo.gender));
		msg.appendInt(-1);
		msg.appendInt(-1);
		msg.appendInt(-1);
		msg.appendStr("");
	}

	@Override
	public String toString()
	{
		return "RoomUser";
	}
}
