package org.openhabbo.sessions;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.openhabbo.access.AccessControl;
import org.openhabbo.core.Logger;
import org.openhabbo.net.NetConnection;
import org.openhabbo.users.User;

public final class SessionManager
{
	private long sessionCounter;
	private final Map<Long, Session> sessions;
	private final Map<Integer, Session> sessionsByUserId;
	private final Map<String, Session> sessionsByUsername;

	public SessionManager(int expectedSessions)
	{
		this.sessionCounter = 0;
		this.sessions = new ConcurrentHashMap<Long, Session>(expectedSessions);
		this.sessionsByUserId = new ConcurrentHashMap<Integer, Session>(expectedSessions);
		this.sessionsByUsername = new ConcurrentHashMap<String, Session>(expectedSessions);
		Logger.getInstance().info("SessionMgr", String.format("Optimized for %d concurrent sessions.", expectedSessions));
	}

	public void shutdown()
	{
		// Any sessions?
		if(!this.sessions.isEmpty())
		{
			Logger.getInstance().info("SessionMgr", "Ending all sessions...");
			for(Session session : this.sessions.values())
			{
				session.onEnd();
			}
			this.sessions.clear();
			this.sessionsByUserId.clear();
			this.sessionsByUsername.clear();
			Logger.getInstance().info("SessionMgr", "All sessions ended.");
		}
	}

	public void startSession(NetConnection conn)
	{
		// Create Session object, assign it a unique ID and assign it as ConnectionHandler for the network connection
		Session session = new Session(++this.sessionCounter, conn);
		conn.setHandler(session);
		this.sessions.put(session.id, session);

		// New session started!
		Logger.getInstance().info("SessionMgr", String.format("Session #%d joined from %s.", session.id, session.getConnection().getIpAddress()));
	}

	public void endSession(long sessionId, String reason)
	{
		// Remove session from map and end session
		Session session = this.sessions.remove(sessionId);
		if(session != null)
		{
			session.onEnd();
			Logger.getInstance().info("SessionMgr", String.format("Session #%d ended. (reason: %s)", sessionId, reason));
		}
	}

	public void addAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		// Already an authed session for this user ID? End it, don't tolerate concurrent logins
		Session concurrent = this.sessionsByUserId.get(user.id);
		if(concurrent != null)
		{
			this.endSession(concurrent.id, "concurrent user login");
		}

		// Add the session to the session user maps
		this.sessionsByUserId.put(user.id, session);
		this.sessionsByUsername.put(user.username.toLowerCase(), session);

		// Write the accesslog
		int accessLogId = AccessControl.getInstance().logUserLogin(user.id, session.getConnection().getIpAddress());
		session.setAccessLogId(accessLogId);

		// User logged in!
		Logger.getInstance().info("SessionMgr", String.format("User #%d (%s, role: %s) logged in on session #%d.", user.id, user.username, user.role.name, session.id));
	}

	public void removeAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		// Remove from session-by-user map
		if(this.sessionsByUserId.remove(user.id) != null)
		{
			this.sessionsByUsername.remove(user.username.toLowerCase());
		}

		// Write the accesslog
		AccessControl.getInstance().logUserLogout(session.getAccessLogId());

		// User logged out!
		Logger.getInstance().info("SessionMgr", String.format("User #%d (%s) logged out.", user.id, user.username));
	}

	public Session getSession(long sessionId)
	{
		return this.sessions.get(sessionId);
	}

	public Session getSessionByUserId(int userId)
	{
		return this.sessionsByUserId.get(userId);
	}

	public Session getSessionByUsername(String username)
	{
		return this.sessionsByUsername.get(username.toLowerCase());
	}

	public void sendNotifToAll(String text)
	{
		this.sendNotifToAll(text, null);
	}

	public void sendNotifToAll(String text, String url)
	{
		for(Session session : this.sessionsByUserId.values())
		{
			session.sendNotif(text, url);
		}
	}

	public int sessions()
	{
		return this.sessions.size();
	}

	public int authedSessions()
	{
		return this.sessionsByUserId.size();
	}

	/* Begin singleton */
	private static SessionManager instance;

	public final static SessionManager getInstance()
	{
		return instance;
	}

	public static void createInstance(int expectedSessions)
	{
		instance = new SessionManager(expectedSessions);
	}
	/* End singleton */

}
