package mod_effects;

import java.util.Iterator;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import org.openhabbo.OpenHABBO;
import org.openhabbo.rooms.instances.RoomUser;
import org.openhabbo.sessions.Session;
import org.openhabbo.sessions.messages.ServerMessage;
import org.openhabbo.storage.DatabaseResult;

public final class AvatarEffectInventory
{
	private final Session session;
	private int currentEffect;
	private final List<AvatarEffect> effects;

	public AvatarEffectInventory(Session session)
	{
		this.session = session;
		this.currentEffect = -1;
		this.effects = new CopyOnWriteArrayList<AvatarEffect>();
	}

	public void clear()
	{
		this.effects.clear();
	}

	public void reload()
	{
		// Clear old effects
		this.clear();

		// Load new ones from database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM user_effects WHERE user_id=" + this.session.getUser().id);
		while(result.row())
		{
			AvatarEffect effect = new AvatarEffect();
			effect.userId = result.getInt("user_id");
			effect.effectId = result.getInt("effect_id");
			effect.totalDuration = result.getInt("total_duration");
			effect.enabled = result.getBoolean("enabled");
			if(effect.enabled)
			{
				effect.enabledAt = result.getTime("enabled_at");
			}
			if(effect.hasExpired())
			{
				effect.delete();
			}
			else
			{
				this.effects.add(effect);
			}
		}
		result.release();
	}

	public void addEffect(int effectId, int totalDuration)
	{
		// Create effect
		AvatarEffect effect = new AvatarEffect();
		effect.userId = this.session.getUser().id;
		effect.effectId = effectId;
		effect.totalDuration = totalDuration;
		effect.insert();

		// Add to list
		this.effects.add(effect);
		
		// Notify user
		ServerMessage msg = new ServerMessage(EEffectMessage.AvatarEffectAdded);
		msg.appendInt(effect.effectId);
		msg.appendInt(effect.totalDuration);
		this.session.send(msg);
	}

	public int getCurrentEffect()
	{
		return this.currentEffect;
	}

	public void setCurrentEffect(int effectId)
	{
		this.currentEffect = effectId;
		this.setEffectToRoomUser();
	}
	
	public void setEffectToRoomUser()
	{
		// Update room user
		RoomUser user = this.session.getRoomUser();
		if(user != null)
		{
			// Set/unset attachment
			if(this.currentEffect == AvatarEffect.NO_EFFECT)
			{
				user.unset("effect");
			}
			else
			{
				user.set("effect", this.currentEffect);
			}

			// Notify the room
			ServerMessage msg = new ServerMessage(EEffectMessage.AvatarEffect);
			msg.appendInt(user.avatarId);
			msg.appendInt(this.currentEffect);
			this.session.getRoomInstance().broadcast(msg);
		}
	}

	public void checkAndStopExpiredEffects()
	{
		// Has effects?
		if(!this.effects.isEmpty())
		{
			// Process all effects
			Iterator<AvatarEffect> effects = this.effects.iterator();
			while(effects.hasNext())
			{
				// Effect expired?
				AvatarEffect effect = effects.next();
				if(effect.hasExpired())
				{
					// Delete data
					effect.delete();
					
					// Remove from iterator
					effects.remove();

					// Notify the user that the effect expired
					ServerMessage msg = new ServerMessage(EEffectMessage.AvatarEffectExpired);
					msg.appendInt(effect.effectId);
					this.session.send(msg);

					// Current effect? Disable it!
					if(effect.effectId == this.currentEffect)
					{
						this.setCurrentEffect(-1);
					}
				}
			}
		}
	}

	public AvatarEffect getEffect(int effectId)
	{
		for(AvatarEffect effect : this.effects)
		{
			if(effect.effectId == effectId && !effect.hasExpired())
			{
				return effect;
			}
		}
		return null;
	}

	public List<AvatarEffect> getEffects()
	{
		return this.effects;
	}
}
