package mod_remote;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.List;

import org.openhabbo.core.Logger;

public class RemoteServer implements Runnable
{
	private final static int PACKET_SIZE = 1024;

	private int port;
	private DatagramSocket socket;

	private boolean alive;
	private Thread thread;

	private List<String> allowedIps;

	public RemoteServer(int port)
	{
		this.port = port;
		this.allowedIps = new ArrayList<String>();
	}

	public boolean start()
	{
		// Not already started?
		if(this.alive)
		{
			return false;
		}
		else
		{
			// Bind socket
			try
			{
				this.socket = new DatagramSocket(this.port);
			}
			catch(SocketException ex)
			{
				Logger.getInstance().error("RemoteServer", String.format("Error binding to UDP/%d.", this.port), ex);
				return false;
			}

			// Create thread
			this.thread = new Thread(this, "RemoteServer (UDP/" + this.port + ")");

			// Start thread
			this.alive = true;
			this.thread.start();

			// Started!
			Logger.getInstance().info("RemoteServer", String.format("Listening for remote commands on UDP port %d.", this.port));
			return true;
		}
	}

	public void stop()
	{
		if(this.alive)
		{
			// Close socket
			this.alive = false;
			this.thread.interrupt();

			// Close the socket
			this.socket.close();
		}
	}

	public void addAllowedIp(String ip)
	{
		synchronized(this.allowedIps)
		{
			this.allowedIps.add(ip);
		}
	}

	public void removeAllowedIp(String ip)
	{
		synchronized(this.allowedIps)
		{
			this.allowedIps.remove(ip);
		}
	}

	@Override
	public void run()
	{
		// This packet is re-used all the time
		DatagramPacket packet = new DatagramPacket(new byte[RemoteServer.PACKET_SIZE], RemoteServer.PACKET_SIZE);

		// While the server is running...
		while(this.alive)
		{
			// Receive the next packet
			try
			{
				this.socket.receive(packet);
			}
			catch(IOException ex)
			{
				// Logger.getInstance().error("RemoteServer", "Error reading new UDP packet.", ex);
				continue;
			}

			// Sender is allowed?
			String senderIp = packet.getAddress().getHostAddress();
			if(this.allowedIps.contains(senderIp))
			{
				// Get data as a string and parse "COMMAND arg1 arg2 arg3 etc"
				String data = new String(packet.getData(), 0, packet.getLength());
				String parts[] = data.split(" ", 2);

				// Dispatch command to command handler
				this.handleCommand(parts[0], (parts.length == 2) ? parts[1] : null);
			}
			else
			{
				// Lol hax
				Logger.getInstance().warning("RemoteServer", String.format("IP '%s' is not allowed to issue remote commands. Command ignored.", senderIp));
			}
		}
	}
	
	private boolean handleCommand(String cmd, String payload)
	{
		// Handler registered?
		RemoteCommandHandler handler = RemoteCmdRegistry.lookup(cmd);
		if(handler == null)
		{
			Logger.getInstance().warning("RemoteServer", String.format("Unknown command '%s'. Command ignored.", cmd));
			return false;
		}

		// Invoke handler and catch exceptions
		try
		{
			handler.handle(payload);
		}
		catch(Exception ex)
		{
			Logger.getInstance().error("RemoteServer", String.format("Error processing command '%s'. (payload=%s)", cmd, payload), ex);
		}

		// Handled
		return true;
	}

	public int getPort()
	{
		return this.port;
	}
	
	/* Begin singleton */
	private static RemoteServer instance;
	
	public final static RemoteServer getInstance()
	{
		return instance;
	}
	
	public static void createInstance(int port)
	{
		instance = new RemoteServer(port);
	}
	/* End singleton */
}
