package org.openhabbo.catalog;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.items.EItemBehaviour;
import org.openhabbo.items.Item;
import org.openhabbo.items.ItemDefinition;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.users.User;
import org.openhabbo.util.SecurityUtil;

public class Catalog
{
	private final static SimpleDateFormat PRIZETROPHY_DATEFORMAT = new SimpleDateFormat("dd-MM-yyyy");
	private final static String ERROR_CUSTOMDATA = Character.toString((char) 1);

	private Object lock;
	private Map<Integer, CatalogPage> pages;
	private Map<Integer, CatalogProduct> products;

	public Catalog()
	{
		this.lock = new Object();
		this.pages = new HashMap<Integer, CatalogPage>();
		this.products = new HashMap<Integer, CatalogProduct>();
	}

	public void reload()
	{
		synchronized(this.lock)
		{
			this.pages.clear();
			this.products.clear();

			// Load all the pages
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM catalog_pages ORDER BY ordernum ASC");
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					CatalogPage page = CatalogPage.parse(result);
					this.pages.put(page.id, page);
				}
			}
			result.release();

			// Link the pages in parent/child fashion
			for(CatalogPage page : this.pages.values())
			{
				if(page.parentId > 0)
				{
					CatalogPage parent = this.pages.get(page.parentId);
					if(parent == null)
					{
						OpenHABBO.getInstance().getLogger().warning("Catalog", String.format("Page #%d is a child of page #%d, which is missing. Page not attached to parent.", page.id, page.parentId));
					}
					else
					{
						parent.pages.add(page);
					}
				}
			}

			// Load the products, grab the item definitions and attach them to the pages
			result = OpenHABBO.getInstance().getDb().query("SELECT * FROM catalog_products ORDER BY ordernum");
			while(result.row())
			{
				CatalogProduct product = CatalogProduct.parse(result);
				product.itemDefinition = OpenHABBO.getInstance().getItemMgr().getDefinition(product.itemDefinitionId);
				if(product.itemDefinition == null)
				{
					if(product.itemDefinitionId > 0)
					{
						OpenHABBO.getInstance().getLogger().warning("Catalog", String.format("Product #%d ('%s') refers to the missing item definition #%d. Product ignored.", product.id, product.name, product.itemDefinitionId));
					}
				}
				else
				{
					if(product.pageId > 0)
					{
						CatalogPage page = this.pages.get(product.pageId);
						if(page == null)
						{
							OpenHABBO.getInstance().getLogger().warning("Catalog", String.format("Product #%d ('%s') is supposed to appear at page #%d, which is missing. Product not attached to page.", product.id, product.name, product.pageId));
						}
						else
						{
							product.page = page;
							page.products.add(product);
						}
					}
					this.products.put(product.id, product);
				}
			}
			result.release();

			// Done!
			OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Reloaded %d pages and %d products.", this.pages(), this.products()));
		}
	}

	public CatalogPage getPage(int pageId)
	{
		return this.pages.get(pageId);
	}

	public List<CatalogPage> getVisibleIndexPagesForAccessLevel(int accessLevel)
	{
		List<CatalogPage> visible = new ArrayList<CatalogPage>();
		synchronized(this.lock)
		{
			for(CatalogPage page : this.pages.values())
			{
				if(page.isVisible && page.parentId == -1 && page.minAccessLevel <= accessLevel)
				{
					visible.add(page);
				}
			}
		}
		return visible;
	}

	public int pages()
	{
		return this.pages.size();
	}

	public CatalogProduct getProduct(int productId)
	{
		return this.products.get(productId);
	}

	public int products()
	{
		return this.products.size();
	}

	public boolean purchase(CatalogProduct product, String customData, User user, int receiverId, boolean asGift, String giftNote)
	{
		// Run pre-purchase logic, filter & apply customdata
		customData = this.handlePrePurchase(user, product, customData);
		if(customData != null && customData.equals(ERROR_CUSTOMDATA))
		{
			OpenHABBO.getInstance().getLogger().warning("Catalog", String.format("Purchase of product #%d by user #%d failed, bad customdata.", product.id, user.id));
			return false;
		}

		// Produce what kind of item?
		Item item;
		if(!asGift)
		{
			item = product.itemDefinition.newInstance();
		}
		else
		{
			// Attempt to get a presentbox definition
			ItemDefinition presentBoxDef = OpenHABBO.getInstance().getItemMgr().getRandomPresentBoxDefinition();
			if(presentBoxDef == null)
			{
				return false;
			}

			// Apply giftnote
			item = presentBoxDef.newInstance();
			item.customData = "!" + giftNote;
		}

		// Insert the item in the database and into the shipping
		if(!OpenHABBO.getInstance().getItemMgr().storeItem(item))
		{
			return false;
		}

		// Finally create the items 'to ship'
		List<Item> items = new ArrayList<Item>(1);
		items.add(item);

		if(asGift)
		{
			// Store the gift payload, under the ID of the presentbox item
			OpenHABBO.getInstance().getItemMgr().createGiftInfo(item.id, product.id, customData);
		}
		else
		{
			// Add this item to the shipping and run post-purchase logic, eventually adding more items
			this.handlePostPurchase(item, customData, items);
		}

		// Is the receiver online?
		Session receiver = OpenHABBO.getInstance().getSessionMgr().getSessionByUserId(receiverId);
		if(receiver != null)
		{
			// Add the items to inventory
			for(Item tmpItem : items)
			{
				receiver.getItemInventory().addItem(tmpItem, false);
			}

			// Trigger refresh
			receiver.send(new ServerMessage(EServerMessage.InventoryUpdate));

			// Send out a notification (TODO: localize this?)
			if(asGift)
			{
				receiver.sendNotif("You have received a gift!");
			}
		}

		// Purchase OK!
		return true;
	}

	private String handlePrePurchase(User user, CatalogProduct product, String customData)
	{
		if(product.itemDefinition.is(EItemBehaviour.DECORATION))
		{
			try
			{
				// 5.5
				return Double.toString(Double.parseDouble(customData));
			}
			catch(Exception ex)
			{
				// Can't be processed!
				return ERROR_CUSTOMDATA;
			}
		}
		else if(product.itemDefinition.sprite.startsWith("prizetrophy") || product.itemDefinition.is(EItemBehaviour.PRIZETROPHY))
		{
			return(user.username + (char) 9 + PRIZETROPHY_DATEFORMAT.format(new Date()) + (char) 9 + SecurityUtil.filterStr(customData, false));
		}
		else
		{
			// ... Way more!
		}

		// No customdata!
		return null;
	}

	private void handlePostPurchase(Item item, String customData, List<Item> extraItems)
	{
		if(item.definition.sprite.startsWith("door") || item.is(EItemBehaviour.TELEPORTER))
		{
			// Store teleporter 2
			Item tele2 = item.definition.newInstance();
			tele2.ownerId = item.ownerId;
			tele2.customData = Integer.toString(item.id);
			OpenHABBO.getInstance().getItemMgr().storeItem(tele2);

			// Update teleporter 1
			item.customData = Integer.toString(tele2.id);
			item.update();

			// Add tele2 to the extra item
			extraItems.add(tele2);
		}
		else if(item.is(EItemBehaviour.POSTIT))
		{
			item.customData = "20";
			item.requireDbUpdate();
		}
		else if(item.is(EItemBehaviour.ROOMDIMMER))
		{

		}
	}
}
