package org.openhabbo.modules;

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.openhabbo.OpenHABBO;

public final class ModuleManager
{
	private final static String JAR_DIR = "/modules";

	private File jarDirectory;
	private Map<String, Module> modules;

	public ModuleManager()
	{
		this.jarDirectory = new File(new File("").getAbsolutePath() + JAR_DIR);
		this.modules = new HashMap<String, Module>();
	}

	public void reloadModules()
	{
		// Unload any loaded modules
		this.unloadModules();
		
		// Loading...
		OpenHABBO.getInstance().getLogger().info("ModuleMgr", "Loading modules...");
		
		// Discover + load all the modules, both .class-files and .jars
		int loaded = this.loadModulesFromDirs() + this.loadModulesFromJars();

		// Anything loaded?
		OpenHABBO.getInstance().getLogger().info("ModuleMgr", String.format("Loaded %d modules.", loaded));
	}

	private int loadModulesFromDirs()
	{
		int loaded = 0;
		for(File file : new File("").getAbsoluteFile().listFiles())
		{
			if(file.isDirectory() && file.getName().startsWith("mod_"))
			{
				if(this.loadModule(file.getName()))
				{
					loaded++;
				}
			}
		}
		return loaded;
	}

	private int loadModulesFromJars()
	{
		// Directory exists?
		int loaded = 0;
		if(!this.jarDirectory.exists() && !this.jarDirectory.mkdir())
		{
			OpenHABBO.getInstance().getLogger().warning("ModuleMgr", String.format("Directory does not exist, no .jar modules loaded.", JAR_DIR));
		}
		else
		{
			// Search the jar directory for mod_%.jar's
			for(String file : this.jarDirectory.list())
			{
				if(file.startsWith("mod_") && file.endsWith(".jar"))
				{
					String name = file.split("\\.", 2)[0];
					if(this.loadModule(name))
					{
						loaded++;
					}
				}
			}
		}
		return loaded;
	}

	@SuppressWarnings("unchecked")
	private boolean loadModule(String name)
	{
		// Already loaded?
		if(this.modules.containsKey(name))
		{
			return false;
		}
		
		// In the 'dont load me'-list?
		String skipList = OpenHABBO.getInstance().getProperties().get("modules.ignorelist", "");
		if(skipList.contains(name))
		{
			OpenHABBO.getInstance().getLogger().debug("ModuleMgr", String.format("Skipping %s, module is on ignorelist.", name));
			return false;
		}

		// The loader first tries to find it in the classpath, but will fall back to loading from the .jar when needed
		try
		{
			File file = new File(this.jarDirectory + "/" + name + ".jar");
			ClassLoader loader = new URLClassLoader(new URL[] { file.toURI().toURL() });
			Class<? extends Module> moduleClass = (Class<? extends Module>) loader.loadClass(name + "." + name);
			Module module = moduleClass.newInstance();
			return this.loadModule(module);
		}
		catch(Exception ex)
		{
			// We are _not_ interested in anything that happens here, this module is bad. Just bad
			return false;
		}
	}

	private boolean loadModule(Module module)
	{
		// Attempt load
		boolean loaded = false;
		try
		{
			module.reloadConfig();
			loaded = module.onLoad();
		}
		catch(Exception ex)
		{
			OpenHABBO.getInstance().getLogger().error("ModuleMgr", String.format("Error loading %s.", module.name), ex);
		}

		// Loaded successfully, no errors and return value 'true' at onLoad()?
		if(loaded)
		{
			this.modules.put(module.name, module);
			OpenHABBO.getInstance().getLogger().debug("ModuleMgr", String.format("Loaded %s.", module.name));
		}
		return loaded;
	}

	public void unloadModules()
	{
		int unloaded = 0;
		synchronized(this.modules)
		{
			for(Module module : this.modules.values())
			{
				this.unloadModule(module);
				unloaded++;
			}
			this.modules.clear();
		}
		if(unloaded > 0)
		{
			OpenHABBO.getInstance().getLogger().info("ModuleMgr", String.format("Unloaded %d modules.", unloaded));
		}
	}

	private void unloadModule(Module module)
	{
		try
		{
			module.onUnload();
		}
		catch(Exception ex)
		{
			OpenHABBO.getInstance().getLogger().error("ModuleMgr", String.format("Error unloading %s.", module.name), ex);
		}
		finally
		{
			OpenHABBO.getInstance().getLogger().debug("ModuleMgr", String.format("Unloaded %s.", module.name));
		}
	}

	public Map<String, String> getLoadedModules()
	{
		// Gather list and generate status string for every module
		Map<String, String> loaded = new HashMap<String, String>();
		for(Entry<String, Module> module : this.modules.entrySet())
		{
			loaded.put(module.getKey(), module.getValue().getStatusString());
		}

		// TODO: sort alphabetically (keys)
		return loaded;
	}

	public int count()
	{
		return this.modules.size();
	}
}
