package org.openhabbo.modules;

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.openhabbo.OpenHABBO;

public final class ModuleManager
{
	private final static String DIR_NAME = "/modules";

	private File directory;
	private Map<String, Module> modules;

	public ModuleManager()
	{
		this.directory = new File(new File("").getAbsolutePath() + DIR_NAME);
		this.modules = new HashMap<String, Module>();
	}

	public void loadModules()
	{
		// Directory exists?
		if(!this.directory.exists() && !this.directory.mkdir())
		{
			OpenHABBO.getInstance().getLogger().warning("ModuleMgr", String.format("Directory does not exist, no modules loaded.", DIR_NAME));
		}
		else
		{
			OpenHABBO.getInstance().getLogger().info("ModuleMgr", String.format("Loading .jar modules from \"%s\"...", DIR_NAME));

			// Search the directory for mod_.jar's
			int loaded = 0;
			for(String file : this.directory.list())
			{
				if(file.startsWith("mod_") && file.endsWith(".jar"))
				{
					String name = file.split("\\.", 2)[0];
					if(this.loadModule(name))
					{
						loaded++;
					}
				}
			}
		}
	}

	public void unloadModules()
	{
		OpenHABBO.getInstance().getLogger().info("ModuleMgr", "Unloading modules...");
		int unloaded = 0;
		synchronized(this.modules)
		{
			for(Module module : this.modules.values())
			{
				this.unloadModule(module);
				unloaded++;
			}
			this.modules.clear();
		}
		OpenHABBO.getInstance().getLogger().info("ModuleMgr", String.format("Unloaded %d modules.", unloaded));
	}

	public void reloadModules()
	{
		OpenHABBO.getInstance().getLogger().info("ModuleMgr", "Reload of all modules issued.");
		this.unloadModules();
		this.loadModules();
	}

	@SuppressWarnings("unchecked")
	private boolean loadModule(String name)
	{
		// The loader first tries to find it in the classpath, but will fall back to loading from the .jar when needed
		try
		{
			File file = new File(this.directory + "/" + name + ".jar");
			ClassLoader loader = new URLClassLoader(new URL[] { file.toURI().toURL() });
			Class<? extends Module> moduleClass = (Class<? extends Module>) loader.loadClass(name + "." + name);
			Module module = moduleClass.newInstance();
			return this.loadModule(module);
		}
		catch(Exception ex)
		{
			// We are _not_ interested in anything that happens here, this module is bad. Just bad
			return false;
		}
	}

	private boolean loadModule(Module module)
	{
		// Attempt load
		boolean loaded = false;
		try
		{
			loaded = module.onLoad();
		}
		catch(Exception ex)
		{
			OpenHABBO.getInstance().getLogger().error("ModuleMgr", String.format("Error loading %s.", module.name), ex);
		}
		
		// Loaded successfully, no errors and return value 'true' at onLoad()?
		if(loaded)
		{
			OpenHABBO.getInstance().getLogger().info("ModuleMgr", String.format("Loaded %s.", module.name));
		}
		return loaded;
	}

	private void unloadModule(Module module)
	{
		try
		{
			module.onUnload();
		}
		catch(Exception ex)
		{
			OpenHABBO.getInstance().getLogger().error("ModuleMgr", String.format("Error unloading %s.", module.name), ex);
		}
		finally
		{
			OpenHABBO.getInstance().getLogger().debug("ModuleMgr", String.format("Unloaded %s.", module.name));
		}
	}

	public Map<String, String> getLoadedModules()
	{
		// Gather list and generate status string for every module
		Map<String, String> loaded = new HashMap<String, String>();
		for(Entry<String, Module> module : this.modules.entrySet())
		{
			loaded.put(module.getKey(), module.getValue().getStatusString());
		}

		// TODO: sort alphabetically (keys)
		return loaded;
	}

	public int count()
	{
		return this.modules.size();
	}
}
