package org.openhabbo.net;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;

import org.openhabbo.OpenHABBO;
import org.openhabbo.net.messages.ClientMessage;
import org.openhabbo.util.Base64Encoding;
import org.openhabbo.util.CrossdomainPolicy;

public class NetConnection
{
	private final static int MESSAGEBUFFER_SIZE = 1024;

	public interface ConnectionHandler
	{
		public void onNetNewMessage(ClientMessage msg);

		public void onNetClose(String reason);
	}

	private SocketChannel m_channel;
	private SelectionKey m_key;
	private boolean m_closed;
	private NetWorker m_worker;
	private NetConnection.ConnectionHandler m_handler;
	private int m_msgFullLength;
	private ByteBuffer m_readBuffer;
	private ByteBuffer m_sendBuffer;

	public NetConnection(SocketChannel channel)
	{
		m_channel = channel;
	}

	public void closeSocket()
	{
		if(!m_closed)
		{
			m_closed = true;
			// m_msgBuffer = null;
			m_readBuffer = null;
			m_sendBuffer = null;
			if(m_worker != null)
			{
				m_worker.unregister(this);
			}
			try
			{
				m_channel.close();
			}
			catch(IOException ex)
			{
				OpenHABBO.getInstance().getLogger().error("NetConnection", "Failed to close socket.", ex);
			}
		}
	}

	public void close(String reason)
	{
		this.closeSocket();
		if(m_handler != null)
		{
			m_handler.onNetClose(reason);
		}
	}

	public boolean assignWorker(NetWorker worker, Selector selector)
	{
		try
		{
			m_channel.configureBlocking(false);
			m_key = m_channel.register(selector, SelectionKey.OP_READ | SelectionKey.OP_WRITE, this);
			m_readBuffer = ByteBuffer.allocate(NetConnection.MESSAGEBUFFER_SIZE);
			m_sendBuffer = ByteBuffer.allocate(NetConnection.MESSAGEBUFFER_SIZE * 4);
		}
		catch(IOException ex)
		{
			return false;
		}
		m_worker = worker;
		return true;
	}

	public void send(byte[] data)
	{
		if(m_key.isValid())
		{
			m_sendBuffer.put(data);
			m_key.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
		}
	}

	public void onWrite() throws IOException
	{
		m_key.interestOps(SelectionKey.OP_READ);
		m_sendBuffer.flip();
		m_channel.write(m_sendBuffer);
		if(!m_sendBuffer.hasRemaining())
		{
			m_sendBuffer.clear();
		}
		else
		{
			m_sendBuffer.flip();
			m_key.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
		}
	}

	public boolean onRead() throws IOException
	{
		// Read all available data into the buffer
		int read = m_channel.read(m_readBuffer);
		if(read <= 0)
		{
			return false;
		}

		// Flip to readable mode
		m_readBuffer.flip();

		// Process the entire buffer
		while(m_readBuffer.hasRemaining())
		{
			// Waiting for new message?
			if(m_msgFullLength == 0)
			{
				byte[] header = new byte[3];
				m_readBuffer.get(header);
				if(header[0] == 64)
				{
					m_msgFullLength = Base64Encoding.decodeInt(header);
					if(m_msgFullLength <= 0 || m_msgFullLength > NetConnection.MESSAGEBUFFER_SIZE)
					{
						return false;
					}
				}
				else
				{
					OpenHABBO.getInstance().getLogger().debug(m_handler, "requesting policy file");
					this.send(CrossdomainPolicy.getXml().getBytes());
					return true;
				}
			}

			// Message complete?
			if(m_readBuffer.remaining() >= m_msgFullLength)
			{
				// Copy all msgdata into a new array
				byte[] bytes = new byte[m_msgFullLength];
				m_readBuffer.get(bytes);
				m_msgFullLength = 0;

				// Message received!
				ClientMessage msg = new ClientMessage(bytes, bytes.length);
				m_handler.onNetNewMessage(msg);
			}
		}

		// Prepare the buffer for the next data
		if(m_msgFullLength == 0)
		{
			m_readBuffer.clear();
		}
		else
		{
			m_readBuffer.flip();
		}
		return true;
	}

	public String getIpAddress()
	{
		return m_channel.socket().getInetAddress().getHostAddress();
	}

	public boolean closed()
	{
		return m_closed;
	}

	public NetConnection.ConnectionHandler getHandler()
	{
		return m_handler;
	}

	public void setHandler(NetConnection.ConnectionHandler handler)
	{
		m_handler = handler;
	}
}
