package org.openhabbo.rooms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomWorker;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public final class RoomManager
{
	public final static int MAX_LOAD_PER_WORKER = 30;

	private Map<String, RoomModel> models = new HashMap<String, RoomModel>();
	private Map<Integer, Room> cache = new HashMap<Integer, Room>();
	private Map<Integer, RoomInstance> instances;
	private List<RoomWorker> workers;

	public RoomManager(int expectedInstances, int initialWorkers)
	{
		this.instances = new HashMap<Integer, RoomInstance>(expectedInstances);
		this.workers = new ArrayList<RoomWorker>(initialWorkers);
		OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Optimized for %d concurrent room instances, spawning %d initial workers...", expectedInstances, initialWorkers));
		for(int x = 0; x < initialWorkers; x++)
		{
			this.spawnNewWorker();
		}
	}

	public void onShutdown()
	{
		synchronized(this.workers)
		{
			for(RoomWorker worker : this.workers)
			{
				worker.stop();
			}
			this.workers.clear();
		}

		synchronized(this.instances)
		{
			if(this.instances.size() > 0)
			{
				OpenHABBO.getInstance().getLogger().info("RoomMgr", "Unloading all room instances...");
				for(RoomInstance instance : this.instances.values())
				{
					instance.onUnload();
				}
				this.instances.clear();
				OpenHABBO.getInstance().getLogger().info("SessionMgr", "All room instances unloaded.");
			}
		}
	}

	public void reloadModels()
	{
		synchronized(this.models)
		{
			this.models.clear();

			// Load all model data and parse them
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM roommodels");
			while(result.row())
			{
				RoomModel model = RoomModel.parse(result);
				model.parseMap();
				this.models.put(model.name, model);
			}
			result.release();

			OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Reloaded %d room models.", this.models()));
		}
	}

	public RoomModel getModel(String name)
	{
		return this.models.get(name);
	}

	public int models()
	{
		return this.models.size();
	}

	public void clearCache()
	{
		synchronized(this.cache)
		{
			this.cache.clear();
		}
	}

	public Room getRoomInfo(int roomId)
	{
		// Lookup room instances
		RoomInstance instance = this.getRoomInstance(roomId, false);
		if(instance != null)
		{
			return instance.getInfo();
		}

		// Lookup database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r LEFT JOIN users u ON u.id = r.owner_id WHERE r.id =" + roomId);
		Room info = result.row() ? Room.parse(result) : null;
		result.release();
		return info;
	}

	public List<Room> getRoomsForUser(int userId)
	{
		ArrayList<Room> rooms = new ArrayList<Room>(5);

		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id=" + userId);
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}

	public RoomInstance getRoomInstance(int roomId, boolean allowCreate)
	{
		// Look for existing instance
		RoomInstance instance = this.instances.get(roomId);
		if(instance != null)
		{
			return instance;
		}

		// Allowed to create a new instance?
		if(allowCreate)
		{
			return this.createRoomInstance(roomId);
		}
		else
		{
			return null;
		}
	}

	private int spawnNewWorker()
	{
		int index = this.workers.size();
		this.workers.add(new RoomWorker(index, this));
		this.workers.get(index).start();
		OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Spawned RoomWorker #%d.", index));
		return index;
	}

	private RoomWorker getBestWorker()
	{
		int lowest = 999, bestIndex = -1;
		synchronized(this.workers)
		{
			for(int index = 0; index < this.workers.size(); index++)
			{
				int load = this.workers.get(index).load();
				if((load < RoomManager.MAX_LOAD_PER_WORKER) && load < lowest)
				{
					lowest = load;
					bestIndex = index;
				}
			}
			if(bestIndex == -1)
			{
				bestIndex = this.spawnNewWorker();
			}
			return this.workers.get(bestIndex);
		}
	}

	private RoomInstance createRoomInstance(int roomId)
	{
		// Get room info
		Room info = this.getRoomInfo(roomId);
		if(info == null)
		{
			return null;
		}

		// Get room model
		RoomModel model = this.getModel(info.modelName);
		if(model == null)
		{
			OpenHABBO.getInstance().getLogger().warning("RoomMgr", String.format("Room #%d refers to the missing room model \"%s\". Could not create room instance.", info.id, info.modelName));
			return null;
		}

		// Create instance and load data like furniture
		RoomInstance instance = new RoomInstance(info, model);
		if(!instance.onLoad())
		{
			return null;
		}

		// Assign to worker
		RoomWorker worker = this.getBestWorker();
		worker.add(instance);
		instance.setWorkerIndex(worker.index());

		// Add to instance map
		synchronized(this.instances)
		{
			this.instances.put(instance.getInfo().id, instance);
		}

		OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Created instance for room #%d.", info.id));
		return instance;
	}

	public void unloadRoomInstance(int roomId)
	{
		// Attempt to remove from the instance map
		RoomInstance instance;
		synchronized(this.instances)
		{
			instance = this.instances.remove(roomId);
		}

		// Instance removed? Process further unloading
		if(instance != null)
		{
			// Un-assign from worker
			synchronized(this.workers)
			{
				RoomWorker worker = this.workers.get(instance.getWorkerIndex());
				if(worker != null)
				{
					worker.remove(roomId);
				}
			}

			// Unload/save data
			instance.onUnload();

			// Unloaded successfully!
			OpenHABBO.getInstance().getLogger().info("RoomMgr", String.format("Unloaded instance of room #%d.", roomId));
		}
	}

	public Room getRoomInfoFromInstance(int roomId)
	{
		RoomInstance instance = this.getRoomInstance(roomId, false);
		if(instance != null)
		{
			return instance.getInfo();
		}
		else
		{
			return null;
		}
	}

	public int instances()
	{
		return this.instances.size();
	}

	public int workers()
	{
		return this.workers.size();
	}

	public int createGuestRoom(int userId, String name, String modelName)
	{
		return OpenHABBO.getInstance().getDb().executeReturnKey("INSERT INTO rooms(owner_id,name,model_name) VALUES (" + userId + ",'" + SQLHelper.escape(name) + "','" + SQLHelper.escape(modelName) + "')");
	}

	public void deleteGuestRoom(int roomId)
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM rooms WHERE id =" + roomId);
		// OpenHABBO.getInstance().getDb().execute("DELETE FROM room_userbans)

		// onDeleteGuestRoom(roomId)
	}

	public Map<Integer, Long> getBansForRoom(int roomId)
	{
		// Initialize an empty map with init cap of 3
		Map<Integer, Long> bans = new HashMap<Integer, Long>(3);

		// Query the user IDs + expiretimes for bans in this rooms that haven't expired yet, we don't bother retrieving them
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT user_id,expires FROM room_userbans WHERE room_id=" + roomId + " AND expires > " + System.currentTimeMillis());
		while(result.row())
		{
			bans.put(result.getInt("user_id"), result.getLong("expires"));
		}
		result.release();

		// Return the bans
		return bans;
	}

	public void addRoomBan(int roomId, int userId, long expireTime)
	{
		OpenHABBO.getInstance().getDb().execute("INSERT INTO room_userbans(room_id,user_id,expires) VALUES(" + roomId + "," + userId + "," + expireTime + ")");
	}

	public void removeRoomBan(int roomId, int userId)
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM room_userbans WHERE room_id=" + roomId + " AND user_id=" + userId);
	}

	public Map<Integer, String> getRightsForRoom(int roomId)
	{
		// Initialize an empty map with init cap of 3
		Map<Integer, String> users = new HashMap<Integer, String>(3);

		// Query the user IDs + usernames
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT user_id,username FROM room_rights JOIN users ON id=user_id WHERE room_id=" + roomId);
		while(result.row())
		{
			users.put(result.getInt("user_id"), result.getString("username"));
		}
		result.release();

		// Return the users
		return users;
	}

	public void addRoomRights(int roomId, int userId)
	{
		OpenHABBO.getInstance().getDb().execute("INSERT INTO room_rights(room_id,user_id) VALUES(" + roomId + "," + userId + ")");
	}

	public void removeRoomRights(int roomId, int userId)
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId + " AND user_id=" + userId);
	}
}
