package org.openhabbo.rooms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public class RoomNavigator
{
	private long m_lastRefresh;
	private int m_maxAge;

	private Object m_lock = new Object();
	private Map<Integer, RoomCategory> m_categories = new HashMap<Integer, RoomCategory>();
	private List<NavigatorFrontpageItem> m_frontpage = new ArrayList<NavigatorFrontpageItem>();
	private List<Room> m_busiestGuestRooms = new ArrayList<Room>();
	private List<Room> m_topRatedGuestRooms = new ArrayList<Room>();

	public RoomNavigator(int maxAgeSeconds)
	{
		m_maxAge = (maxAgeSeconds * 1000);
	}

	public void reload()
	{
		synchronized(m_lock)
		{
			// Load categories
			m_categories.clear();
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM navigator_categories ORDER BY ordernum ASC");
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					RoomCategory category = RoomCategory.parse(result);
					m_categories.put(category.id, category);
				}
			}

			// Load frontpage items
			m_frontpage.clear();
			result.query("SELECT id,name,bannertype,type,inner_id,image,image_src,enabled FROM navigator_frontpage ORDER BY ordernum ASC");
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					NavigatorFrontpageItem fpi = new NavigatorFrontpageItem();
					fpi.id = result.getInt("id");
					fpi.name = result.getString("name");
					fpi.bannerType = result.getInt("bannertype");
					fpi.image = result.getString("image");
					fpi.imageIsExternal = result.getString("image_src").equals("external");

					// Get inner item
					String innerType = result.getString("type");
					int innerId = result.getInt("inner_id");
					if(innerType.equals("room"))
					{
						fpi.room = OpenHABBO.getInstance().getRoomMgr().getRoomInfo(innerId);
					}
					else
					{
						fpi.category = this.getCategory(innerId);
					}

					// Inner item exists? (don't enable invalid frontpage items)
					if(fpi.room != null || fpi.category != null)
					{
						m_frontpage.add(fpi);
					}
					else
					{
						OpenHABBO.getInstance().getLogger().warning("RoomNavigator", String.format("Frontpage item #%d (\"%s\") refers to the missing %s #%d. Item ignored.", fpi.id, fpi.name, innerType, innerId));
					}
				}
			}
			result.release();

			// Yay!
			OpenHABBO.getInstance().getLogger().info("RoomNavigator", String.format("Reloaded %d guestroom categories and %d frontpage items.", this.categories(), this.frontpageItems()));
		}
	}

	public boolean requiresRefresh()
	{
		return((System.currentTimeMillis() - m_lastRefresh) > m_maxAge);
	}

	public void refresh()
	{
		synchronized(m_lock)
		{
			// Reload category rooms
			for(RoomCategory category : m_categories.values())
			{
				category.rooms.clear();
				DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE category_id=" + category.id + " ORDER BY users_now DESC,rating DESC LIMIT 40");
				while(result.row())
				{
					Room room = Room.parse(result);
					category.rooms.put(room.id, room);
				}
				result.release();
			}

			// Refresh toprated rooms
			m_topRatedGuestRooms.clear();
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id > 0 ORDER BY rating DESC LIMIT 40");
			while(result.row())
			{
				m_topRatedGuestRooms.add(Room.parse(result));
			}
			result.release();

			// Refresh frontpage items
			for(NavigatorFrontpageItem fpi : m_frontpage)
			{
				if(fpi.room != null)
				{
					Room newInfo = OpenHABBO.getInstance().getRoomMgr().getRoomInfo(fpi.room.id);
					if(newInfo != null)
					{
						fpi.room = newInfo;
					}
				}
			}

			// This is for debug only
			OpenHABBO.getInstance().getLogger().debug("RoomNavigator", "Refreshed successfully.");
		}
	}

	public RoomCategory getCategory(int categoryId)
	{
		return m_categories.get(categoryId);
	}

	public int categories()
	{
		return m_categories.size();
	}

	public Collection<RoomCategory> getCategories()
	{
		return m_categories.values();
	}

	public int frontpageItems()
	{
		return m_frontpage.size();
	}

	public List<NavigatorFrontpageItem> getFrontpageItems()
	{
		return m_frontpage;
	}

	public List<Room> getBusiestGuestRooms()
	{
		return m_busiestGuestRooms;
	}

	public List<Room> getTopRatedRooms()
	{
		return m_topRatedGuestRooms;
	}

	public List<Room> getGuestRoomSearchResult(String criteria, boolean byTags)
	{
		ArrayList<Room> rooms = new ArrayList<Room>();

		criteria = SQLHelper.escape(criteria);
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE u.username='" + criteria + "' OR r.name LIKE '" + criteria + "%' OR r.tags LIKE '%" + criteria + "%' LIMIT 30;");
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}
}
