package org.openhabbo.rooms.instances;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.rooms.RoomManager;

public class RoomWorker implements Runnable
{
	private int m_index;
	private RoomManager m_mgr;

	private Thread m_thread;
	private boolean m_running;
	private List<Integer> m_toRemove;
	private Map<Integer, RoomInstance> m_instances;

	public RoomWorker(int index, RoomManager mgr)
	{
		m_index = index;
		m_mgr = mgr;
		m_toRemove = new ArrayList<Integer>();
		m_instances = new HashMap<Integer, RoomInstance>(RoomManager.MAX_LOAD_PER_WORKER);
	}

	public void start()
	{
		if(!m_running)
		{
			m_running = true;
			m_thread = new Thread(this, this.toString());
			m_thread.start();
		}
	}

	public void stop()
	{
		m_running = false;
		m_thread.interrupt();
		synchronized(m_instances)
		{
			m_instances.clear();
		}
	}

	public void add(RoomInstance instance)
	{
		synchronized(m_instances)
		{
			m_instances.put(instance.getInfo().id, instance);
		}
	}

	public void remove(int roomId)
	{
		synchronized(m_instances)
		{
			m_toRemove.add(roomId);
		}
	}

	@Override
	public void run()
	{
		while(m_running)
		{
			// Measure start time, pulse all instances, measure end time
			long startTime = System.nanoTime();
			synchronized(m_instances)
			{
				if(m_instances.size() > 0)
				{
					for(RoomInstance instance : m_instances.values())
					{
						try
						{
							instance.pulse();
							if(instance.requiresUnload())
							{
								m_mgr.unloadRoomInstance(instance.getInfo().id);
							}
						}
						catch(Exception ex)
						{
							OpenHABBO.getInstance().getLogger().error(this, String.format("Error pulsing room instance #%d.", instance.getInfo().id), ex);
						}
					}
				}
				if(m_toRemove.size() > 0)
				{
					for(Integer roomId : m_toRemove)
					{
						m_instances.remove(roomId);
					}
					m_toRemove.clear();
				}
			}
			long endTime = System.nanoTime();

			// Now sleep the essential amount of ms
			double sleepTime = (endTime - startTime);
			sleepTime /= 1000000;
			sleepTime = (RoomInstance.PULSE_INTERVAL - sleepTime);
			if(sleepTime > 0.0)
			{
				try
				{
					Thread.sleep((long) sleepTime);
				}
				catch(InterruptedException ex)
				{

				}
			}
		}
	}

	public int index()
	{
		return m_index;
	}

	public RoomManager getMgr()
	{
		return m_mgr;
	}

	public int load()
	{
		return m_instances.size();
	}

	public boolean running()
	{
		return m_running;
	}

	@Override
	public String toString()
	{
		return "RoomWorker #" + m_index;
	}
}
