package org.openhabbo.sessions;

import java.util.List;

import mod_achievement.AchievementManager;

import org.openhabbo.OpenHABBO;
import org.openhabbo.items.ItemInventory;
import org.openhabbo.modules.ModuleResource;
import org.openhabbo.net.NetConnection;
import org.openhabbo.net.messages.ClientMessage;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.registries.SessionMsgRegistry;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomUser;
import org.openhabbo.support.ModerationBan;
import org.openhabbo.users.User;
import org.openhabbo.users.effects.AvatarEffectInventory;

public final class Session extends ModuleResource implements NetConnection.ConnectionHandler
{
	// Events
	public final static int ON_LOGIN = 4000;
	public final static int ON_LOGOUT = 4001;
	public final static int ON_ROOMENTER = 4002;
	public final static int ON_ROOMEXIT = 4003;
	
	// Core
	public final long id;
	private NetConnection conn;

	// Message handling
	private long lastMsgTime;
	private ServerMessage msg;
	private ServerMessage responseObject;

	// Session variables
	private User user;
	private int accessLogId;
	//private InstantMessenger messenger;
	private ItemInventory itemInventory;
	private AvatarEffectInventory effectInventory;
	private RoomInstance roomInstance;
	private RoomUser roomUser;

	public int authenticatedRoomId;
	public int authenticatedTeleporterId;

	public Session(long sessionId, NetConnection conn)
	{
		this.id = sessionId;
		this.conn = conn;
		this.msg = new ServerMessage();
		this.responseObject = new ServerMessage();

		this.onCreate();
	}

	private void onCreate()
	{

	}

	public void onEnd()
	{
		if(!this.conn.closed())
		{
			this.conn.closeSocket();
		}
		if(this.authenticated())
		{
			this.onLogout();
			OpenHABBO.getInstance().getSessionMgr().removeAuthedSession(this);
		}
	}

	public void authenticate(String ticket)
	{
		// User exists for this auth ticket?
		User user = OpenHABBO.getInstance().getUserMgr().getUserByAuthTicket(ticket);
		if(user == null)
		{
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, "invalid auth ticket");
			return;
		}

		// User is banned?
		List<ModerationBan> bans = OpenHABBO.getInstance().getSupportCenter().getBansForUserId(user.id, false);
		if(bans.size() > 0)
		{
			// Send the first ban and exit auth routine
			this.sendBan(bans.get(0));
			return;
		}

		// Insufficient permissions?
		if(user.role == null || !user.role.hasRight("fuse_login"))
		{
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
			return;
		}

		// Add to session list
		this.user = user;
		OpenHABBO.getInstance().getSessionMgr().addAuthedSession(this);

		// Send user right list
		this.sendUserRights();

		// Login OK!
		this.msg.init(EServerMessage.AuthenticationOK);
		this.send(this.msg);

		// Run other logic
		this.onLogin();
	}

	private void onLogin()
	{
		// Reload achievements
		this.user.achievements = AchievementManager.getAchievementsForUser(this.user.id);

		// Reload item/pets inventory
		this.itemInventory = new ItemInventory(this.user.id);
		this.itemInventory.reload();

		// Reload effect inventory
		this.effectInventory = new AvatarEffectInventory(this.user.id);
		this.effectInventory.reload();
		
		// Execute other on-login tasks
		this.fireEvent(Session.ON_LOGIN);
	}

	private void onLogout()
	{
		// Stop trading, leave the room, etc etc
		this.leaveRoom();

		// Clear inventories to speed up garbagecollection
		this.itemInventory.clear();
		this.effectInventory.clear();

		// Run other stuff
		this.fireEvent(Session.ON_LOGOUT);
	}

	public void tryRoom(int roomId, String password)
	{
		// Leave any room instance the user might be in now
		this.leaveRoom();

		// Get/create room instance
		RoomInstance room = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}

		// Room is full?
		if(room.getInfo().usersNow >= room.getInfo().usersMax)
		{
			// Can't override user limit?
			if(!this.getUser().role.hasRight("fuse_enter_full_rooms"))
			{
				// "Room is full"
				this.msg.init(EServerMessage.RoomEntryError);
				this.msg.appendInt(1);
				this.send(this.msg);

				// Kick to hotel view
				this.msg.init(EServerMessage.CloseConnection);
				this.send(this.msg);

				return;
			}
		}

		// The room owner can ALWAYS enter the room, never has to check for user limit, bans, state, etc
		if(this.getUser().id != room.getInfo().ownerId)
		{
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().role.hasRight("fuse_enter_full_rooms"))
				{
					// "Room is full"
					this.msg.init(EServerMessage.RoomEntryError);
					this.msg.appendInt(1);
					this.send(this.msg);

					// Kick to hotel view
					this.msg.init(EServerMessage.CloseConnection);
					this.send(this.msg);

					return;
				}
			}

			// Execute access checks for room state?
			if(!this.getUser().role.hasRight("fuse_enter_any_room"))
			{
				// Password room?
				if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equalsIgnoreCase(password))
				{
					this.sendGenericError(-100002);
					this.kickFromRoom(null);
					return;
				}
				// Doorbell room?
				else if(room.getInfo().state == Room.STATE_DOORBELL)
				{

				}

				// User is banned?
				if(room.userIsBanned(this.getUser().id))
				{
					// "You are banned from this room" message
					this.msg.init(EServerMessage.RoomEntryError);
					this.msg.appendInt(4);
					this.send(this.msg);

					// Kick to hotel viw
					this.msg.init(EServerMessage.CloseConnection);
					this.send(this.msg);

					return;
				}
			}
		}

		// Send room stats tracker url (?)
		this.msg.init(166);
		this.msg.append("/client/");
		if(room.getInfo().isPublicRoom())
		{
			this.msg.append("public/");
			this.msg.append(room.getInfo().modelName);
			this.msg.append("/0");
		}
		else
		{
			this.msg.append("private/");
			this.msg.append(Integer.toString(room.getInfo().id));
			this.msg.append("/id");
		}
		this.send(this.msg);

		// Send room model
		this.msg.init(EServerMessage.RoomReady);
		this.msg.appendStr(room.getModel().name);
		this.msg.appendInt(room.getInfo().id);
		this.send(this.msg);

		// Forward to room instance
		this.goToRoom(roomId);
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomInstance instance = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}

		// If entering a guestroom, then send more data, client doesn't ask for it
		if(instance.getInfo().isGuestRoom())
		{
			// Send wallpaper/floor/landscape if not default data
			if(!instance.getInfo().wallpaper.equals("0.0"))
			{
				this.msg.init(EServerMessage.RoomProperty);
				this.msg.appendStr("wallpaper");
				this.msg.appendStr(instance.getInfo().wallpaper);
				this.send(this.msg);
			}
			if(!instance.getInfo().floor.equals("0.0"))
			{
				this.msg.init(EServerMessage.RoomProperty);
				this.msg.appendStr("floor");
				this.msg.appendStr(instance.getInfo().floor);
				this.send(this.msg);
			}
			if(!instance.getInfo().landscape.equals("0.0"))
			{
				this.msg.init(EServerMessage.RoomProperty);
				this.msg.appendStr("landscape");
				this.msg.appendStr(instance.getInfo().landscape);
				this.send(this.msg);
			}

			// Send rating
			this.msg.init(EServerMessage.RoomRating);
			this.msg.appendInt(-1);//instance.getInfo().rating);
			this.send(this.msg);

			// Send ongoing room event
			this.msg.init(EServerMessage.RoomEvent);
			this.msg.appendStr("-1");
			this.send(this.msg);
		}

		// This is the current room instance now
		this.roomInstance = instance;
	}

	public boolean enterRoom()
	{
		this.roomUser = this.roomInstance.addSession(this);
		if(this.roomUser != null)
		{
			this.fireEvent(Session.ON_ROOMENTER);
			return true;
		}
		else
		{
			this.kickFromRoom("Invalid room entry!");
			return false;
		}
	}

	private void leaveRoom()
	{
		// Reset room ID if not teleporting
		if(this.authenticatedTeleporterId == 0)
		{
			this.authenticatedRoomId = 0;
		}

		// Leave room instance
		if(this.roomInstance != null)
		{
			this.fireEvent(Session.ON_ROOMEXIT);
			this.roomInstance.removeSession(this);
			this.roomInstance = null;
			this.roomUser = null;
		}
	}

	public void kickFromRoom(String text)
	{
		// SHE GAVE ME THE PLEASE-FACE
		this.leaveRoom();

		// I GAVE HER THE LEAVE-FACE
		this.msg.init(EServerMessage.CloseConnection);
		this.send(this.msg);

		// BODY LIKE A BOTTLE, PRETTY LITTLE MODEL MODEL MODEEL
		if(text != null)
		{
			this.sendNotif(text);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Fucking heartbeat! ("D{")
		if(msg.msgId == 315) return;

		// Log that this session received a message!
		this.lastMsgTime = System.currentTimeMillis();
		OpenHABBO.getInstance().getLogger().debug(this, "<-- " + msg.toString());

		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable purposes
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			OpenHABBO.getInstance().getNetMgr().blacklistIp(this.getConnection().getIpAddress(), "MSG_EOF");
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, "MSG_EOF detected in incoming message");
			return;
		}

		// These are the only responding handlers for SendSessionParams and SsoLogin
		if(!this.authenticated() && (msg.msgId != 206 && msg.msgId != 415))
		{
			OpenHABBO.getInstance().getLogger().warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.msgId, msg.header()));
		}
		else
		{
			// Lookup message handler and execute handling logic
			SessionMessageHandler handler = SessionMsgRegistry.lookup(msg.msgId);
			if(handler != null)
			{
				System.out.println(handler.getClass().getSimpleName());
				try
				{
					handler.handle(this, msg, this.responseObject);
				}
				catch(Exception ex)
				{
					OpenHABBO.getInstance().getLogger().error(this, String.format("Error handling msg #%d (\"%s\").", msg.msgId, msg.header()), ex);
				}
				this.responseObject.reset();
			}
			else
			{
				OpenHABBO.getInstance().getLogger().warning(this, String.format("SessionMsgRegistry does not have an entry for msg #%d. (\"%s\")", msg.msgId, msg.header()));
			}
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		OpenHABBO.getInstance().getSessionMgr().endSession(this.id, reason);
	}

	public void send(ServerMessage msg)
	{
		OpenHABBO.getInstance().getLogger().debug(this, "--> " + msg.toString());
		this.conn.send(msg.getResult());
	}

	public void sendGenericError(int code)
	{

	}

	public void sendBan(ModerationBan ban)
	{
		this.msg.init(EServerMessage.UserBanned);
		this.msg.appendStr(ban.message);
		// TODO: "Date the ban will be lifted: %s" etc
		this.send(this.msg);
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null);
	}

	public void sendNotif(String text, String url)
	{
		this.msg.init(EServerMessage.Notif);
		this.msg.appendStr(text);
		this.msg.appendStr(url);
		this.send(this.msg);
	}
	
	public void sendSystemNotif(String text)
	{
		this.msg.init(810);
		this.msg.appendBool(true);
		this.msg.appendStr(text);
		this.send(this.msg);
	}

	public void sendUserRights()
	{
		if(this.user != null)
		{
			this.msg.init(EServerMessage.UserRights);
			this.msg.appendInt(this.user.role.rights.length);
			for(String right : this.user.role.rights)
			{
				this.msg.appendStr(right);
			}
			this.send(this.msg);
		}
	}

	public void sendCreditBalance()
	{
		this.msg.init(EServerMessage.CreditBalance);
		this.msg.appendStr(this.user.credits + ".0");
		this.send(this.msg);
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		this.msg.init(EServerMessage.PixelsReceivedNotification);
		this.msg.appendInt(this.user.pixels);
		this.msg.appendInt(added);
		this.send(this.msg);
	}

	public void sendSubscriptionData(String subscriptionId)
	{
		this.msg.init(EServerMessage.ScrSendUserInfo);
		this.msg.appendStr(subscriptionId);
		this.msg.appendInt(0); // days left
		this.msg.appendBool(true); // show expire date?
		this.msg.appendInt(0); // months left
		this.send(this.msg);
	}

	public void sendRoomPrivileges()
	{
		if(this.roomUser != null)
		{
			if(this.roomUser.hasRoomRights)
			{
				this.msg.init(EServerMessage.YouAreController);
				this.send(this.msg);
			}
			if(this.roomUser.hasRoomOwner)
			{
				this.msg.init(EServerMessage.YouAreOwner);
				this.send(this.msg);
			}
		}
	}

	public NetConnection getConnection()
	{
		return this.conn;
	}

	public long getLastMsgTime()
	{
		return this.lastMsgTime;
	}

	public boolean authenticated()
	{
		return(this.user != null);
	}

	public User getUser()
	{
		return this.user;
	}

	public int getAccessLogId()
	{
		return this.accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.accessLogId == 0)
		{
			this.accessLogId = accessLogId;
		}
	}

	//public InstantMessenger getInstantMessenger()
	//{
	//	return this.messenger;
	//}

	public ItemInventory getItemInventory()
	{
		return this.itemInventory;
	}

	public AvatarEffectInventory getEffectInventory()
	{
		return this.effectInventory;
	}

	public boolean inRoom()
	{
		return(this.roomInstance != null);
	}

	public RoomInstance getRoomInstance()
	{
		return this.roomInstance;
	}

	public RoomUser getRoomUser()
	{
		return this.roomUser;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
