package org.openhabbo.sessions;

import java.util.HashMap;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.net.NetConnection;
import org.openhabbo.users.User;

public class SessionManager
{
	private long m_sessionCounter;
	private Object m_lock;
	private Map<Long, Session> m_sessions;
	private Map<Integer, Session> m_sessionsByUserId;
	private Map<String, Session> m_sessionsByUsername;

	public SessionManager(int expected)
	{
		m_lock = new Object();
		m_sessions = new HashMap<Long, Session>(expected);
		m_sessionsByUserId = new HashMap<Integer, Session>(expected);
		m_sessionsByUsername = new HashMap<String, Session>(expected);
		OpenHABBO.getInstance().getLogger().info("SessionMgr", String.format("Optimized for %d concurrent sessions.", expected));
	}

	public void onShutdown()
	{
		synchronized(m_lock)
		{
			if(m_sessions.size() > 0)
			{
				OpenHABBO.getInstance().getLogger().info("SessionMgr", "Ending all sessions...");
				for(Session session : m_sessions.values())
				{
					session.onEnd();
				}
				m_sessions.clear();
				OpenHABBO.getInstance().getLogger().info("SessionMgr", "All sessions ended.");
			}
		}
	}

	public void startSession(NetConnection conn)
	{
		// Create Session object, assign it a unique ID and assign it as ConnectionHandler for the network connection
		Session session = new Session(++m_sessionCounter, conn);
		conn.setHandler(session);
		synchronized(m_lock)
		{
			m_sessions.put(session.id, session);
		}

		// New session started!
		OpenHABBO.getInstance().getLogger().info("SessionMgr", String.format("Session #%d joined from %s.", session.id, session.getConnection().getIpAddress()));
	}

	public void endSession(long sessionId, String reason)
	{
		// Remove session from map
		Session session;
		synchronized(m_lock)
		{
			session = m_sessions.remove(sessionId);
		}

		// Actually a session mapped for this ID? End it
		if(session != null)
		{
			session.onEnd();
			OpenHABBO.getInstance().getLogger().info("SessionMgr", String.format("Session #%d ended. (reason: %s)", sessionId, reason));
		}
	}

	public void addAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		synchronized(m_lock)
		{
			// Already an authed session for this user ID? End it, don't tolerate concurrent logins
			Session concurrent = m_sessionsByUserId.get(user.id);
			if(concurrent != null)
			{
				this.endSession(concurrent.id, "concurrent user login");
			}

			// Add the session to the session user maps
			m_sessionsByUserId.put(user.id, session);
			m_sessionsByUsername.put(user.username.toLowerCase(), session);
		}

		// Write the accesslog
		int accessLogId = OpenHABBO.getInstance().getAccessCtrl().logUserLogin(user.id, session.getConnection().getIpAddress());
		session.setAccessLogId(accessLogId);

		// User logged in!
		OpenHABBO.getInstance().getLogger().info("SessionMgr", String.format("User #%d (%s, role: %s) logged in on session #%d.", user.id, user.username, user.role.name, session.id));
	}

	public void removeAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		// Remove from session-by-user map
		synchronized(m_lock)
		{
			if(m_sessionsByUserId.remove(user.id) != null)
			{
				m_sessionsByUsername.remove(user.username.toLowerCase());
			}
		}

		// Write the accesslog
		OpenHABBO.getInstance().getAccessCtrl().logUserLogout(session.getAccessLogId());

		// User logged out!
		OpenHABBO.getInstance().getLogger().info("SessionMgr", String.format("User #%d (%s) logged out.", user.id, user.username));

	}

	public Session getSession(long sessionId)
	{
		return m_sessions.get(sessionId);
	}

	public Session getSessionByUserId(int userId)
	{
		return m_sessionsByUserId.get(userId);
	}

	public Session getSessionByUsername(String username)
	{
		return m_sessionsByUsername.get(username.toLowerCase());
	}

	public void sendNotifToAll(String text)
	{
		this.sendNotifToAll(text, null);
	}

	public void sendNotifToAll(String text, String url)
	{
		synchronized(m_lock)
		{
			for(Session session : m_sessionsByUserId.values())
			{
				session.sendNotif(text, url);
			}
		}
	}

	public int sessions()
	{
		return m_sessions.size();
	}

	public int authedSessions()
	{
		return m_sessionsByUserId.size();
	}
}
