package org.openhabbo.storage;

import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

public class DatabaseQueryQueue implements Runnable
{
	private SQLDatabase db;
	private DatabaseConnection conn;
	private Statement st;
	private List<String> queries;

	private boolean alive;
	private Thread thread;

	public DatabaseQueryQueue(SQLDatabase db)
	{
		this.db = db;
		this.queries = new ArrayList<String>(50);
	}

	public boolean open()
	{
		// Not already open?
		if(this.isOpen())
		{
			return false;
		}

		// Get database connection
		DatabaseConnection conn = this.db.getConnMgr().getConnection();
		if(conn == null)
		{
			this.db.getLogger().warning("DatabaseQueryQueue", "Could not open because no database connection was available.");
			return false;
		}

		// Attempt creating the statement that will be used to process queue's with
		try
		{
			this.st = conn.getInnerConnection().createStatement();
			this.conn = conn;
		}
		catch(SQLException ex)
		{
			this.db.getConnMgr().releaseConnection(conn);
			this.db.getLogger().error("DatabaseQueryQueue", "Error creating statement!", ex);
			return false;
		}

		// Create thread
		this.thread = new Thread(this, "DatabaseQueryQueue");
		this.thread.setPriority(Thread.MIN_PRIORITY);

		// Start thread
		this.alive = true;
		this.thread.start();

		// Open for business
		return true;
	}

	public void close()
	{
		// Actually open?
		if(!this.isOpen())
		{
			return;
		}

		// Abort the thread
		this.alive = false;
		this.thread.interrupt();

		// Flush the queue
		this.process();

		// Close the statement
		try
		{
			this.st.close();
		}
		catch(SQLException ex)
		{
			this.db.getLogger().error("DatabaseQueryQueue", "Error closing statement.", ex);
		}

		// Release the connection
		this.db.getConnMgr().releaseConnection(this.conn);
	}

	public void enqueue(String sql)
	{
		synchronized(this.queries)
		{
			this.queries.add(sql);
		}
	}

	private void process()
	{
		// Don't bother with empty queues
		if(this.queries.size() == 0)
		{
			return;
		}

		// Execute all the queries
		synchronized(this.queries)
		{
			for(String query : this.queries)
			{
				try
				{
					this.st.executeUpdate(query);
				}
				catch(SQLException ex)
				{
					this.db.getLogger().error("DatabaseQueryQueue", String.format("Error executing \"%s\", error message: %s.", query, ex.getMessage()), null);
				}
			}

			// Clear the queue
			this.queries.clear();
		}
	}

	@Override
	public void run()
	{
		while(this.alive)
		{
			// Process queue
			this.process();

			// Process the queue every second
			try
			{
				Thread.sleep(1000);
			}
			catch(InterruptedException e)
			{
				return;
			}
		}
	}

	public boolean isOpen()
	{
		return(this.st != null);
	}

	public SQLDatabase getDb()
	{
		return this.db;
	}
}
