package org.openhabbo.core;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Date;
import java.text.SimpleDateFormat;

public final class Logger
{
	public enum ELogLevel
	{
		Debug, Info, Warning, Error
	}

	private final static Logger instance = new Logger();

	public final static Logger getInstance()
	{
		return Logger.instance;
	}

	private final SimpleDateFormat timeformat;
	private ELogLevel minImportance;
	private BufferedWriter writer;

	public Logger()
	{
		this.timeformat = new SimpleDateFormat("dd-MM-yyyy @ HH:mm:ss");
		this.minImportance = ELogLevel.Debug;
	}

	public boolean setOutputFile(String fileName)
	{
		// Append .log extension to file if not already there
		if(!fileName.endsWith(".log"))
		{
			fileName = fileName + ".log";
		}

		// Attempt to create filewriter
		try
		{
			this.writer = new BufferedWriter(new FileWriter(fileName, true));
			return true;
		}
		catch(IOException ex)
		{
			System.out.println("ERROR creating new FileWriter for filename \"" + fileName + "\"");
			return false;
		}
	}

	public void setMinImportance(ELogLevel level)
	{
		this.minImportance = level;
	}

	private void writeLog(ELogLevel level, String line)
	{
		if(level.ordinal() < this.minImportance.ordinal())
		{
			return;
		}

		// Log to console
		System.out.println(line);

		// Log to file?
		if(this.writer != null)
		{
			if(level != ELogLevel.Debug)
			{
				try
				{
					synchronized(this.writer)
					{
						this.writer.write(line);
						this.writer.write("\r\n");
						this.writer.flush();
					}
				}
				catch(IOException ex)
				{
					System.out.println("Error logging to file!");
				}
			}
		}
	}

	public void debug(Object system, String s)
	{
		this.writeLog(ELogLevel.Debug, this.timeformat.format(new Date()) + " - [" + system + "][#debug] " + s);
	}

	public void info(Object system, String s)
	{
		this.writeLog(ELogLevel.Info, this.timeformat.format(new Date()) + " - [" + system + "] " + s);
	}

	public void warning(Object system, String s)
	{
		this.writeLog(ELogLevel.Warning, this.timeformat.format(new Date()) + " - [" + system + "][#warning] " + s);
	}

	public void error(Object system, String s, Exception ex)
	{
		this.writeLog(ELogLevel.Error, this.timeformat.format(new Date()) + " - [" + system + "][#error] " + s + ", stacktrace: \r\n" + Logger.convertStackTraceToString(ex));
	}

	public static String convertStackTraceToString(Exception ex)
	{
		StringWriter sw = new StringWriter();
		if(ex != null)
		{
			ex.printStackTrace(new PrintWriter(sw));
		}
		return sw.toString();
	}
}
