package org.openhabbo.net_old;

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

/*
 * This is how the networking layers work at accepting new connections:
 * NetListener is a ServerSocketChannel listening for connections, once it receives a connection request, it will check if the remote host is blacklisted. If so, connection is dropped. If not blacklisted, it will forward the NetConnection to the NetManager. (NetManager.onAcceptedByListener)
 * NetManager will find the most suitable NetWorker for the newly accepted NetConnection, or spawn a new NetWorker if needed. It will then enqueue the NetConnection with the NetWorker so it's interests can be registered on the NetWorker thread, rather than the NetListener thread.
 * NetWorker will wakeup from it's blocking select() call and attempt to register any pending NetConnections, getting it's selector interested in events from the NetConnection's channel. If the connection has been closed in the meantime, it will be disposed and registration fails. If registration succeeds, it will notify the NetManager that the NetConnection has been accepted by the NetWorker. (NetManager.onAcceptedByProcessor)
 * NetManager will now ask the SessionManager to start a new Session for the newly accepted NetConnection.
 */
public final class NetManager
{
	private NetListener listener;
	private List<String> blacklist;
	private NetWorker[] workers;

	public NetManager(int port, int numWorkers)
	{
		this.listener = new NetListener(port, this);
		this.blacklist = new ArrayList<String>();
		this.workers = new NetWorker[numWorkers];
		for(int index = 0; index < numWorkers; index++)
		{
			this.workers[index] = new NetWorker(index, this);
			this.workers[index].start();
			Logger.getInstance().info("NetMgr", String.format("Spawned NetWorker #%d.", index));
		}
	}

	public void stopWorkers()
	{
		for(int index = 0; index < this.workers.length; index++)
		{
			this.workers[index].stop();
			this.workers[index] = null;
		}
		Logger.getInstance().info("NetMgr", "Stopped all workers.");
	}

	private NetWorker getLowestLoadedWorker()
	{
		int lowest = 0, bestIndex = -1;
		for(int index = 0; index < this.workers.length; index++)
		{
			int load = this.workers[index].load();
			if(bestIndex == -1 || load < lowest)
			{
				lowest = load;
				bestIndex = index;
			}
		}
		return this.workers[bestIndex];
	}

	public void onAcceptedByListener(NetConnection conn)
	{
		// Check for blacklist status
		if(this.ipIsBlacklisted(conn.getIpAddress()))
		{
			Logger.getInstance().warning("NetMgr", String.format("Denying connection request from blacklisted IP %s.", conn.getIpAddress()));
			conn.closeSocket();
			return;
		}

		// Get the best worker
		NetWorker worker = this.getLowestLoadedWorker();
		if(!worker.alive())
		{
			worker.start();
		}
		
		// Enqueue on worker
		worker.enqueueNewConnection(conn);
	}

	public void onAcceptedByWorker(NetConnection conn)
	{
		SessionManager.getInstance().startSession(conn);
	}

	public boolean ipIsBlacklisted(String ip)
	{
		return this.blacklist.contains(ip);
	}

	public void reloadBlacklist()
	{
		synchronized(this.blacklist)
		{
			this.blacklist.clear();

			// Reload all the blacklisted IPs
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT ip FROM system_ipblacklist");
			while(result.row())
			{
				this.blacklist.add(result.getString("ip"));
			}
			result.release();

			Logger.getInstance().info("NetMgr", String.format("Reloaded %d blacklisted IP(s).", this.blacklist.size()));
		}
	}

	public void blacklistIp(String ip, String info)
	{
		// Already blacklisted?
		if(this.ipIsBlacklisted(ip))
		{
			return;
		}

		// Add to internal blacklist
		synchronized(this.blacklist)
		{
			this.blacklist.add(ip);
		}

		// Insert new IP to blacklist table
		OpenHABBO.getInstance().getDb().execute("INSERT INTO system_ipblacklist(ip,timestamp,info) VALUES ('" + SQLHelper.escape(ip) + "','" + SQLHelper.getDateTimeStr() + "','" + SQLHelper.escape(info) + "')");

		// Log it
		Logger.getInstance().info("NetMgr", String.format("Blacklisted IP address %s (\"%s\").", ip, info));
	}

	public NetListener getListener()
	{
		return this.listener;
	}

	/* Begin singleton */
	private static NetManager instance;

	public static NetManager getInstance()
	{
		return instance;
	}

	public static void createInstance(int port, int numWorkers)
	{
		instance = new NetManager(port, numWorkers);
	}
	/* End singleton */
}
