package org.openhabbo.rooms.instances;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.openhabbo.communication.EServerMessage;
import org.openhabbo.communication.ServerMessage;
import org.openhabbo.core.DynamicObject;
import org.openhabbo.events.Event;
import org.openhabbo.items.EItemFlags;
import org.openhabbo.items.Item;
import org.openhabbo.items.ItemManager;
import org.openhabbo.items.impl.TeleporterItem;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.RoomModel;
import org.openhabbo.rooms.pathfinding.JoehPathfinderStaticImpl;
import org.openhabbo.rooms.pathfinding.PathfinderNode;
import org.openhabbo.rooms.pathfinding.RotationCalculator;
import org.openhabbo.sessions.Session;
import org.openhabbo.util.SecurityUtil;

public final class RoomInstance extends DynamicObject
{
	public final static int PULSE_INTERVAL = 470;

	// Events
	public final static Event onLoad = new Event("RoomInstance.OnLoad");
	public final static Event onUnload = new Event("RoomInstance.OnUnload");
	public final static Event onSessionEnter = new Event("RoomInstance.OnSessionEnter");
	public final static Event onSessionExit = new Event("RoomInstance.OnSessionExit");
	public final static Event onPulseBegin = new Event("RoomInstance.OnPulseBegin");
	public final static Event onPulseEnd = new Event("RoomInstance.OnPulseEnd");

	// Core room info
	private final Room info;
	private final RoomModel model;
	private final RoomMap map;

	// Instance info
	private int workerIndex;
	private int idleTimer;
	private boolean unloading;

	// Content
	private final ConcurrentHashMap<Integer, Long> bans;
	private final ConcurrentHashMap<Integer, String> rights;
	private final ConcurrentHashMap<Integer, Item> items;

	// Avatars
	private int avatarIdCounter;
	private final ConcurrentHashMap<String, RoomAvatar> avatars;
	private final ConcurrentHashMap<Integer, RoomUser> users;

	public RoomInstance(Room info, RoomModel model)
	{
		// Core room info
		this.info = info;
		this.model = model;
		this.map = model.createMap();

		// Worker
		this.workerIndex = -1;
		this.idleTimer = 0;
		this.unloading = false;

		// Content
		this.bans = new ConcurrentHashMap<Integer, Long>(0);
		this.rights = new ConcurrentHashMap<Integer, String>(0);
		this.items = new ConcurrentHashMap<Integer, Item>();

		// Avatars
		this.avatarIdCounter = 0;
		this.avatars = new ConcurrentHashMap<String, RoomAvatar>();
		this.users = new ConcurrentHashMap<Integer, RoomUser>();
	}

	public void load() throws Exception
	{
		// Reload bans & rights
		this.bans.clear();
		this.rights.clear();
		this.rights.putAll(RoomManager.getInstance().getRightsForRoom(this.info.id));

		// Reload items
		this.items.clear();
		for(Item item : ItemManager.getInstance().getItemsForRoom(this.info.id))
		{
			item.setRoom(this);
			this.items.put(item.id, item);
		}

		// Generate map
		this.rebuildFloorMap();
		
		// Reload other stuff
		RoomInstance.onLoad.fire(this);
	}

	public void unload()
	{
		// Already unloading?
		if(this.unloading)
		{
			return;
		}

		// Flag that this room instance is unloading
		this.unloading = true;

		// Unload stuff
		RoomInstance.onUnload.fire(this);

		// Remove all users
		for(RoomUser user : this.users.values())
		{
			user.getSession().kickFromRoom();
		}
		this.users.clear();
		this.avatars.clear();
		this.rights.clear();
		this.bans.clear();
		if(this.info.usersNow > 0)
		{
			this.info.usersNow = 0;
			this.info.saveUserCount();
		}

		// Save & clear items
		for(Item item : this.items.values())
		{
			if(item.requiresDbUpdate())
			{
				item.update();
			}
		}
		this.items.clear();
	}

	public void spawnAvatar(RoomAvatar avatar)
	{
		// Allocate avatar ID
		avatar.avatarId = this.avatarIdCounter++;
		avatar.setRoom(this);

		// Block map?
		if(!this.info.allowWalkthrough)
		{
			RoomTile tile = this.map.getTile(avatar.x, avatar.y);
			if(tile != null)
			{
				tile.hasAvatar = true;
			}
		}
		
		// Add to collection
		this.avatars.put(avatar.getAvatarKey(), avatar);

		// Broadcast entry
		ServerMessage msg = new ServerMessage(EServerMessage.Users);
		msg.appendInt(1);
		msg.appendObject(avatar);
		this.broadcast(msg);

		// Apply tile to avatar
		this.triggerItemStepOn(avatar);
	}

	public void removeAvatar(RoomAvatar avatar)
	{
		// Invalidate for any references floating around
		avatar.invalidate();
		avatar.setRoom(null);

		// Unblock map
		this.map.getTile(avatar.x, avatar.y).hasAvatar = false;

		// Remove from collection
		this.avatars.remove(avatar.getAvatarKey());

		// Broadcast leaving
		ServerMessage msg = new ServerMessage(EServerMessage.UserRemove);
		msg.appendStr(avatar.avatarId);
		this.broadcast(msg);
	}

	public RoomAvatar getAvatarByAvatarKey(String avatarKey)
	{
		return this.avatars.get(avatarKey);
	}

	public RoomAvatar getAvatarByAvatarId(int avatarId)
	{
		for(RoomAvatar avatar : this.avatars.values())
		{
			if(avatar.avatarId == avatarId)
			{
				return avatar;
			}
		}
		return null;
	}

	public RoomUser getUserByUserId(int userId)
	{
		return this.users.get(userId);
	}

	public RoomUser getUserByUsername(String username)
	{
		for(RoomUser user : this.users.values())
		{
			if(user.getInfo().username.equalsIgnoreCase(username))
			{
				return user;
			}
		}
		return null;
	}

	public RoomAvatar getAvatarOnTile(int x, int y)
	{
		for(RoomAvatar avatar : this.avatars.values())
		{
			if(avatar.x == x && avatar.y == y)
			{
				return avatar;
			}
		}
		return null;
	}

	public RoomUser addSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.unloading)
		{
			return null;
		}
		
		// Already in room?
		if(this.users.containsKey(session.getUser().id))
		{
			return null;
		}

		// Create roomuser with unique virtual ID
		RoomUser user = new RoomUser(session);

		// Position user at spawn position
		if(session.authenticatedTeleporterId == 0)
		{
			user.x = this.model.doorX;
			user.y = this.model.doorY;
			user.z = this.model.doorZ;
			user.headRotation = user.bodyRotation = this.model.doorRotation;
		}
		else
		{
			// Teleporter exists?
			Item tele = this.getItem(session.authenticatedTeleporterId);
			if(tele == null)
			{
				return null;
			}
			else
			{
				// Arrive in teleporter
				user.x = tele.x;
				user.y = tele.y;
				user.z = tele.z;
				user.headRotation = user.bodyRotation = tele.rotation;
				((TeleporterItem)tele).arriveUser(user);
			}
		}

		// Determine privileges if in guestroom
		if(this.info.isGuestRoom())
		{
			boolean isAnyRoomOwner = user.getInfo().rights.contains("fuse_is_any_room_owner");
			user.hasRoomOwner = (isAnyRoomOwner || (user.getInfo().id == this.info.ownerId));
			user.hasRoomRights = (user.hasRoomOwner || this.rights.containsKey(user.getInfo().id));
			if(user.hasRoomRights)
			{
				user.addStatus("flatctrl", user.hasRoomOwner ? "useradmin" : null);
			}
		}

		// Spawn avatar
		this.users.put(user.getInfo().id, user);
		this.spawnAvatar(user);

		// Update current user count
		synchronized(this.info)
		{
			this.info.usersNow = this.users.size();
			this.info.saveUserCount();
		}

		// Done!
		return user;
	}

	public void removeSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.unloading)
		{
			return;
		}

		// Remove user from collection
		RoomUser user = this.users.remove(session.getUser().id);

		// Actually user removed?
		if(user != null)
		{
			// Remove avatar
			this.removeAvatar(user);

			// Update current user count
			synchronized(this.info)
			{
				this.info.usersNow = this.users.size();
				this.info.saveUserCount();
			}
		}
	}

	public boolean ringDoorbell(Session ringing)
	{
		boolean heard = false;
		ServerMessage msg = null;
		for(RoomUser user : this.users.values())
		{
			if(user.hasRoomRights)
			{
				// Notify of doorbelling player
				if(msg == null)
				{
					msg = new ServerMessage(EServerMessage.Doorbell);
					msg.appendStr(ringing.getUser().username);
				}
				user.getSession().send(msg);
				heard = true;
			}
		}

		// True if atleast one user heard the doorbell and is allowed to answer it
		return heard;
	}

	public void broadcast(ServerMessage msg)
	{
		if(!this.users.isEmpty())
		{
			//Logger.getInstance().debug(this, "--> " + msg.toString());
			for(RoomUser user : this.users.values())
			{
				user.getSession().send(msg);
			}
		}
	}

	public void addUserBan(int userId, int seconds)
	{
		long expireTime = System.currentTimeMillis() + (seconds * 1000);
		this.bans.put(userId, expireTime);
	}

	public void removeUserBan(int userId)
	{
		this.bans.remove(userId);
	}

	public boolean userIsBanned(int userId)
	{
		Long expireTime = this.bans.get(userId);
		if(expireTime != null)
		{
			if(System.currentTimeMillis() < expireTime)
			{
				return true;
			}
			else
			{
				this.removeUserBan(userId);
			}
		}
		return false;
	}

	public boolean addUserRights(int userId, String username)
	{
		if(this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.put(userId, username);
			RoomManager.getInstance().addRoomRights(this.info.id, userId);
			return true;
		}
	}

	public boolean removeUserRights(int userId)
	{
		if(!this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.remove(userId);
			RoomManager.getInstance().removeRoomRights(this.info.id, userId);
			return true;
		}
	}

	public boolean userHasRights(int userId)
	{
		return this.rights.containsKey(userId);
	}

	public void rebuildFloorMap()
	{
		// Reset map
		this.map.reset();

		// Apply floor items to map
		for(Item item : this.items.values())
		{
			if(item.is(EItemFlags.FLOORITEM))
			{
				this.map.applyFloorItem(item);
			}
		}

		// Restore door tile
		RoomTile door = this.map.getDoor();
		door.state = RoomTile.OPEN;
		door.height = this.map.getModel().doorZ;
	}

	public Item getItem(int itemId)
	{
		return this.items.get(itemId);
	}

	public List<Item> getItems(char behaviorFlag)
	{
		List<Item> items = new ArrayList<Item>();
		for(Item item : this.items.values())
		{
			if(item.is(behaviorFlag))
			{
				items.add(item);
			}
		}
		return items;
	}

	public List<Item> getFloorItemsOnTile(int x, int y)
	{
		List<Item> items = new ArrayList<Item>(0);
		for(Item item : this.items.values())
		{
			if(item.is(EItemFlags.FLOORITEM))
			{
				for(RoomTile tile : this.map.getAffectedTiles(item))
				{
					if(tile.x == x && tile.y == y)
					{
						items.add(item);
					}
				}
			}
		}
		return items;
	}
	
	public List<Item> getFloorItemsOnTileSimple(int x, int y)
	{
		List<Item> items = new ArrayList<Item>(0);
		for(Item item : this.items.values())
		{
			if(item.is(EItemFlags.FLOORITEM))
			{
				if(item.x == x && item.y == item.y)
				{
					items.add(item);
				}
			}
		}
		return items;
	}


	public boolean placeFloorItem(Item item, int x, int y, byte rotation, Session session)
	{
		return this.moveFloorItem(item, x, y, rotation, true, session);
	}

	public boolean moveFloorItem(Item item, int x, int y, byte rotation, Session session)
	{
		return this.moveFloorItem(item, x, y, rotation, false, session);
	}

	private final boolean moveFloorItem(Item item, int newX, int newY, byte newRotation, boolean isNew, Session session)
	{
		// Valid item? (if new, should not be in room), (if not new, should be in room)
		if(item == null || ((isNew && this.items.containsKey(item.id) || !isNew && !this.items.containsKey(item.id))))
		{
			return false;
		}

		// Validate new rotation, this can really trouble client and server!
		if(newRotation != 0 && newRotation != 2 && newRotation != 4 && newRotation != 6)
		{
			return false;
		}
		
		// Get the tiles this object would reside on if this item was placed here, check them for 'freeness'
		List<RoomTile> newTiles = this.map.getAffectedTiles(newX, newY, item.definition.length, item.definition.width, newRotation);
		for(RoomTile tile : newTiles)
		{
			// Bad tile?
			if(tile == RoomTile.OUT_OF_RANGE || this.map.getModel().getMap()[tile.x][tile.y] == 'x')
			{
				return false;
			}

			// Is there an avatar on this tile?
			if(tile.hasAvatar)
			{
				// Conflicting rotation?
				if(newRotation == item.rotation && (tile.x != item.x || tile.y != item.y))
				{
					return false;
				}
			}
		}

		// Initial item height is specified by root tile
		float newZ = this.getMap().getTile(newX, newY).height;
		
		// Trying to stack on itself? (by rotating an item that is already in the space)
		if(!isNew)
		{
			if(item.x == newX && item.y == newY && item.z != newZ)
			{
				newZ = item.z;
			}
		}
		
		// Check top items under the item being moved for stacking
		for(RoomTile tile : newTiles)
		{
			// Get top item
			Item top = tile.topItem;
			if(top != null && top != item)
			{
				// Trying to stack an unstackable item?
				if(item.is(EItemFlags.ALLOW_1X1_STACK_ONLY))
				{
					return false;
				}
				
				// Prohibit stacking other than 1x1?
				if(top.is(EItemFlags.ALLOW_1X1_STACK_ONLY))
				{
					// This item is not 1x1?
					if(!(item.definition.length == 1 && item.definition.width == 1))
					{
						return false;
					}
				}
				
				// Can stack on top?
				if(top.is(EItemFlags.CAN_STACK_ON_TOP))
				{
					newZ = (top.z + top.definition.height);
				}
				else
				{
					return false;
				}
			}
		}

		// Max stacking height
		final int maxItemZ = 15;
		if(newZ > maxItemZ)
		{
			newZ = maxItemZ;
		}

		// Handle other stuff
		if(!item.onPlace(session, this))
		{
			return false;
		}

		// Backup the old position
		int oldX = item.x;
		int oldY = item.y;
		byte oldRotation = item.rotation;

		// Set object on new position
		item.x = newX;
		item.y = newY;
		item.z = newZ;
		item.rotation = newRotation;

		// If item is new to room...
		if(isNew)
		{
			// Associate the item with this room
			item.setRoom(this);
			this.items.put(item.id, item);

			// Important change, update straight away
			item.roomId = this.info.id;
			item.ownerId = this.info.ownerId;
			item.update();
		}
		else
		{
			// Just a db update on instance unload will do!
			item.requireDbUpdate();
		}

		// Broadcast placement
		ServerMessage msg = new ServerMessage(isNew ? EServerMessage.ObjectAdd : EServerMessage.ObjectUpdate);
		msg.appendObject(item);
		this.broadcast(msg);

		// Generate the new map
		this.rebuildFloorMap();

		// Refresh avatars on previous tiles?
		if(!isNew)
		{
			for(RoomTile tile : this.map.getAffectedTiles(oldX, oldY, item.definition.length, item.definition.width, oldRotation))
			{
				this.refreshAvatarsOnTile(tile.x, tile.y);
			}
		}

		// Refresh avatars on new tiles
		for(RoomTile tile : this.map.getAffectedTiles(newX, newY, item.definition.length, item.definition.width, newRotation))
		{
			this.refreshAvatarsOnTile(tile.x, tile.y);
		}

		// Move OK!
		return true;
	}


	public boolean placeWallItem(Item item, String position, Session session)
	{
		return this.moveWallItem(item, position, session, true);
	}
	
	public boolean moveWallItem(Item item, String newPosition, Session session)
	{
		return this.moveWallItem(item, newPosition, session, false);
	}
	
	private boolean moveWallItem(Item item, String newPosition, Session session, boolean isNew)
	{
		// Valid item?
		if(item == null || (isNew && this.items.containsKey(item.id)))
		{
			return false;
		}
		
		// Valid position?
		newPosition = SecurityUtil.checkWallPosition(newPosition);
		if(newPosition == null)
		{
			return false;
		}

		// Trigger the interactor
		if(!item.onPlace(session, this))
		{
			return false;
		}

		// Update the item with new info
		item.setRoom(this);
		item.roomId = this.info.id;
		item.ownerId = this.info.ownerId;
		item.wallPosition = newPosition;
		item.update();

		// Add to the room instance
		this.items.put(item.id, item);

		// Broadcast update
		ServerMessage msg = new ServerMessage(isNew ? EServerMessage.ItemAdd : EServerMessage.ItemUpdate);
		msg.appendObject(item);
		this.broadcast(msg);

		// Move OK!
		return true;
	}
	
	public final Item pickupItem(int itemId, Session session)
	{
		return this.pickupItem(itemId, session, true, false);
	}

	public final Item pickupAndDeleteItem(int itemId)
	{
		return this.pickupItem(itemId, null, false, true);
	}

	private Item pickupItem(int itemId, Session session, boolean execPickupLogic, boolean delete)
	{
		// Find the item, run pre-pickup checks
		Item item = this.getItem(itemId);
		if(item == null || (execPickupLogic && !item.onPickup(session)))
		{
			return null;
		}

		// Remove from collection
		this.items.remove(item.id);

		// Notify clients of removal
		ServerMessage msg = new ServerMessage();
		if(item.is(EItemFlags.FLOORITEM))
		{
			msg.init(EServerMessage.ObjectRemove);
			msg.appendStr(item.id);
			msg.appendStr("");
			msg.appendBool(false);
		}
		else if(item.is(EItemFlags.WALLITEM))
		{
			msg.init(EServerMessage.ItemRemove);
			msg.appendStr(item.id);
			msg.appendBool(false);
		}
		this.broadcast(msg);

		// Update floormap if flooritem
		if(item.is(EItemFlags.FLOORITEM))
		{
			// Re-generate floormap
			this.rebuildFloorMap();

			// Refresh avatars on tile
			for(RoomTile tile : this.map.getAffectedTiles(item))
			{
				this.refreshAvatarsOnTile(tile.x, tile.y);
			}
		}

		// Clear item
		item.setRoom(null);
		if(delete)
		{
			item.delete();
			return null;
		}
		else
		{
			item.roomId = 0;
			item.ownerId = 0;
			item.x = 0;
			item.y = 0;
			item.z = 0;
			item.rotation = 0;
			item.wallPosition = null;
			return item;
		}
	}
	

	public void refreshAvatarsOnTile(int x, int y)
	{
		for(RoomAvatar avatar : this.avatars.values())
		{
			if(avatar.x == x && avatar.y == avatar.y)
			{
				this.refreshAvatar(avatar);
			}
		}
	}

	private void refreshAvatar(RoomAvatar avatar)
	{
		// Get tile
		RoomTile tile = this.map.getTile(avatar.x, avatar.y);
		if(tile == null)
		{
			return;
		}
		
		// Update height
		avatar.z = tile.height;
		
		// Check for items
		avatar.removeNonMoveStatuses();
		if(tile.hasSteppableItem())
		{
			tile.topItem.onAvatarStepOn(avatar);
		}
		
		// Immediate refresh (no waiting on roominstance pulse)
		ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
		msg.appendInt(1);
		avatar.appendStatusString(msg);
		this.broadcast(msg);

		// Live updated!
		avatar.requiresUpdate = false;
	}
	
	private void triggerItemStepOn(RoomAvatar avatar)
	{
		RoomTile tile = this.map.getTile(avatar.x, avatar.y);
		if(tile.hasSteppableItem())
		{
			tile.topItem.onAvatarStepOn(avatar);
		}
		
		// TODO: check effect
	}
	
	private void triggerItemStepOff(RoomAvatar avatar)
	{
		avatar.removeNonMoveStatuses();
		
		RoomTile tile = this.map.getTile(avatar.x, avatar.y);
		if(tile.hasSteppableItem())
		{
			tile.topItem.onAvatarStepOff(avatar);
		}
		
		// TODO: check effect
	}

	public void updateAvatarPosition(RoomAvatar avatar, int x, int y, boolean instantUpdate)
	{
		// Tiles exist?
		RoomTile from = this.map.getTile(avatar.x, avatar.y);
		RoomTile to = this.map.getTile(x, y);
		if(to == from)
		{
			return;
		}
		
		// Update position
		avatar.x = to.x;
		avatar.y = to.y;
		avatar.z = to.getTopHeight();
		
		// Deal with avatar map?
		if(!this.info.allowWalkthrough)
		{
			from.hasAvatar = false;
			to.hasAvatar = true;
		}
		
		// Effects?
		if(avatar instanceof RoomUser)
		{
			RoomUser user = (RoomUser)avatar;
			
			// Get current effect
			user.tmpEffectId = -1;
			int old = user.getCurrentEffect();
			
			// Step/step!
			from.stepOff(avatar);
			to.stepOn(avatar);
			
			// Effect changed?
			if(user.tmpEffectId != old)
			{
				user.setEffect(user.tmpEffectId, false);
			}
			user.tmpEffectId = -1;
		}
		
		// Update 
		if(!instantUpdate)
		{
			avatar.requiresUpdate = true;
		}
		else
		{
			// Force update
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(1);
			avatar.appendStatusString(msg);
			this.broadcast(msg);
			
			// Done!
			avatar.requiresUpdate = false;
		}
	}
	
	public void pulse()
	{
		// Begin pulse?
		if(RoomInstance.onPulseBegin.fire(this))
		{
			// Pulse the items, process timed items, etc
			this.pulseItems();

			// Pulse the avatars, kick idle users, etc
			this.pulseAvatars();

			// End pulse
			RoomInstance.onPulseEnd.fire(this);
		}
	}

	private void pulseItems()
	{
		for(Item item : this.items.values())
		{
			if(item.requiresPulse())
			{
				item.pulse();
			}
		}
	}

	private void pulseAvatars()
	{
		// These values are dealt with at the end of the pulse
		int updates = 0;
		boolean hasUsers = false;
		RoomUser userInDoor = null;

		// Process all avatars
		for(RoomAvatar avatar : this.avatars.values())
		{
			// Avatar is user?
			if(!hasUsers && avatar instanceof RoomUser)
			{
				hasUsers = true;
			}

			// Any updates required?
			avatar.pulse();
			if(avatar.requiresUpdate)
			{
				// Atleast one update to be broadcasted
				updates++;

				// Avatar in need of movement:phase1?
				if(avatar.hasGoal())
				{
					// Destination reached?
					if(avatar.x == avatar.goalX && avatar.y == avatar.goalY)
					{
						// Trip complete!
						avatar.clearMovement();

						// Stopped moving!
						this.triggerItemStepOn(avatar);
						if(avatar.x == this.model.doorX && avatar.y == this.model.doorY)
						{
							if(avatar instanceof RoomUser)
							{
								if(userInDoor == null)
								{
									userInDoor = (RoomUser) avatar;
								}
							}
						}
					}
					else
					{
						// Determine next tile in path
						PathfinderNode next = null;
						if(avatar.overrideNextTile)
						{
							// Override used!
							avatar.overrideNextTile = false;

							// Substitute the next step
							next = new PathfinderNode();
							next.x = avatar.goalX;
							next.y = avatar.goalY;
						}
						else
						{
							// Start new path?
							if(avatar.startNewPath)
							{
								avatar.startNewPath = false;
								this.startAvatarMovement(avatar, avatar.goalX, avatar.goalY, avatar.overrideNextTile);
							}

							// Pop the next tile from the path
							if(!avatar.path.isEmpty())
							{
								next = avatar.path.remove(0);
								if(next == null || !this.map.tileIsPassable(next.x, next.y, avatar.pathfindingAllowedOn, true))
								{
									next = null;
								}
								else
								{
									// Walking diagonal (hacky wacky isometrics!)
									final boolean CROSS_DIAGONAL = true;
									if(!avatar.path.isEmpty() && CROSS_DIAGONAL)
									{
										// Peek + test next node (already popped one, so this is like path[1])
										PathfinderNode node = avatar.path.get(0);
										if(this.map.tileIsPassable(node.x, node.y, avatar.pathfindingAllowedOn))
										{
											// Can't skip more than one tile!
											if(!(Math.abs(avatar.x - node.x) > 1 || Math.abs(avatar.y - node.y) > 1))
											{
												byte xAdd1 = 0;
												byte xAdd2 = 0;
												if(node.x > avatar.x)
												{
													xAdd1 = -1;
													xAdd2 = 1;
												}
												else
												{
													xAdd1 = 1;
													xAdd2 = -1;
												}

												// Valid shortcut?
												if(this.map.tileIsPassable(node.x + xAdd1, node.y, avatar.pathfindingAllowedOn) && this.map.tileIsPassable(avatar.x + xAdd2, avatar.y, avatar.pathfindingAllowedOn))
												{
													next = avatar.path.remove(0);
												}
											}
										}
									}
								}
							}
						}

						// Is there a next tile available?
						if(next == null)
						{
							// Retry?
							if(++avatar.movementTries < 3)
							{
								avatar.startMovingTo(avatar.goalX, avatar.goalY);
								continue;
							}
							else
							{
								avatar.clearMovement();
							}

							// Stopped moving.
							this.triggerItemStepOn(avatar);
						}
						else
						{
							// Moved a tile
							avatar.movementTries = 0;

							// Redirect?
							RoomTile currentTile = this.map.getTile(avatar.x, avatar.y);
							RoomTile nextTile = this.map.getTile(next.x, next.y);
							if(nextTile.redirect != null)
							{
								nextTile = nextTile.redirect;
							}

							// Set next coordinate for the next phase
							avatar.nextX = next.x;
							avatar.nextY = next.y;
							avatar.nextZ = nextTile.height;
							avatar.requiresUpdate = true;
							
							// Unblock & block tiles
							currentTile.hasAvatar = false;
							nextTile.hasAvatar = !this.info.allowWalkthrough;
						}
					}
				}
			}
		}

		// Any updates? This includes simple status updates, but also movement:phase2
		if(updates > 0)
		{
			// Build the status update
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(updates);
			for(RoomAvatar avatar : this.avatars.values())
			{
				// This avatar was one of the batch that required an update?
				if(avatar.requiresUpdate)
				{
					// Simple status update?
					if(avatar.nextX == -1)
					{
						avatar.appendStatusString(msg);
						avatar.requiresUpdate = false;
					}
					else
					{
						// Step off current item
						this.triggerItemStepOff(avatar);
						
						// Rotate avatars body for moving direction
						byte rotation = RotationCalculator.calcAvatarMoveDirection(avatar.x, avatar.y, avatar.nextX, avatar.nextY);

						// Apply rotation
						avatar.headRotation = rotation;
						avatar.bodyRotation = rotation;

						// Add the movement status for the status update
						avatar.statuses.put("mv", avatar.nextX + "," + avatar.nextY + "," + Float.toString(avatar.nextZ));
						avatar.appendStatusString(msg);
						avatar.statuses.remove("mv");
						
						// Update position serverside
						this.updateAvatarPosition(avatar, avatar.nextX, avatar.nextY, false);
						avatar.nextX = -1;
					}
				}
			}

			// Broadcast the status update
			this.broadcast(msg);
		}

		// Update idle timer, kick user in door
		if(hasUsers)
		{
			this.idleTimer = 0;
			if(userInDoor != null)
			{
				userInDoor.getSession().kickFromRoom();
			}
		}
		else
		{
			this.idleTimer++;
		}
	}

	private void startAvatarMovement(RoomAvatar avatar, int goalX, int goalY, boolean overrideNext)
	{
		// Override?
		if(overrideNext)
		{
			avatar.goalX = goalX;
			avatar.goalY = goalY;
			avatar.overrideNextTile = true;
		}
		else
		{
			// Get goal tile
			RoomTile goal = this.map.getTile(goalX, goalY);
			if(goal == null)
			{
				return;
			}
			
			// Redirect?
			if(goal.redirect != null)
			{
				goal = goal.redirect;
			}

			// Open last step tile for pathfinder if the goal is a last step tile
			byte backup = -1;
			if(goal.is(RoomTile.LASTSTEP))
			{
				// No avatar here? (quick check)
				if(!goal.hasAvatar)
				{
					backup = goal.state;
					goal.state = RoomTile.OPEN;
				}
			}

			// Find path using the beloved Joe 'Joeh' Hegarty 2D A* pathfinder
			boolean noDescLimit = this.info.isGuestRoom();
			List<PathfinderNode> path = JoehPathfinderStaticImpl.findPath(avatar.x, avatar.y, goalX, goalY, this.map, 1.5f, noDescLimit ? 100.0f : 2.0f, avatar.pathfindingAllowedOn);

			// Restore map
			if(backup != -1)
			{
				goal.state = backup;
			}

			// Is there a path found?
			if(!path.isEmpty())
			{
				// Remove current tile node
				path.remove(0);

				// Set this path and destination for avatar
				avatar.path = path;
				avatar.goalX = goalX;
				avatar.goalY = goalY;
			}
		}
	}
	
	public final Room getInfo()
	{
		return this.info;
	}

	public final RoomModel getModel()
	{
		return this.model;
	}

	public final RoomMap getMap()
	{
		return this.map;
	}

	public final int getWorkerIndex()
	{
		return this.workerIndex;
	}

	public void setWorkerIndex(int index)
	{
		this.workerIndex = index;
	}

	public boolean requiresUnload()
	{
		final int maxIdleSeconds = 30;
		return(this.idleTimer > ((maxIdleSeconds * 1000) / RoomInstance.PULSE_INTERVAL));
	}

	public Collection<RoomAvatar> getAllAvatars()
	{
		return this.avatars.values();
	}

	public Collection<RoomUser> getAllUsers()
	{
		return this.users.values();
	}

	public Set<Entry<Integer, String>> getUsersWithRights()
	{
		return this.rights.entrySet();
	}

	public Collection<Item> getAllItems()
	{
		return this.items.values();
	}

	@Override
	public final String toString()
	{
		return "RoomInstance #" + this.info.id;
	}
}