package org.openhabbo.rooms.instances;

import org.openhabbo.items.Item;
import org.openhabbo.util.FlagsUtil;

public final class RoomTile
{
	public final static byte OPEN = (1 << 0);
	public final static byte CLOSED = (1 << 1);
	public final static byte LASTSTEP = (1 << 2);
	public final static byte ITEM = (1 << 3);
	public final static byte WATER = (1 << 4);
	
	public final int x;
	public final int y;
	public byte state;
	public float height;
	public boolean hasAvatar;
	public Item topItem;
	public RoomTile redirect;
	
	public RoomTile(int x, int y)
	{
		this.x = x;
		this.y = y;
	}
	
	public final boolean is(byte state)
	{
		return FlagsUtil.isSet(this.state, state);
	}
	
	public boolean isOpen()
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}
		
		// Not open?
		if(!this.is(RoomTile.OPEN))
		{
			return false;
		}
		
		// Open!
		return true;
	}
	

	public boolean isPassable(byte allowedStates, boolean lastStep)
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}

		// Normal tile?
		if(!this.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if(!FlagsUtil.isSet(allowedStates, this.state))
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		
		// Blocked?
		if(this.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (this.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
	
	public void clear(char defaultState)
	{
		this.topItem = null;
		this.redirect = null;
		if(Character.isDigit(defaultState))
		{
			this.state = RoomTile.OPEN;
			this.height = Integer.parseInt(Character.toString(defaultState));
		}
		else
		{
			this.state = RoomTile.CLOSED;
			this.height = 0.0f;
		}
	}
	
	public float getTopHeight()
	{
		if(this.topItem != null)
		{
			return this.topItem.z + this.topItem.definition.height;
		}
		else
		{
			return this.height;
		}
	}
	
	public void stepOn(RoomAvatar avatar)
	{
		if(this.hasSteppableItem())
		{
			this.topItem.onAvatarStepOn(avatar);
		}
	}
	
	public void stepOff(RoomAvatar avatar)
	{
		if(this.hasSteppableItem())
		{
			this.topItem.onAvatarStepOff(avatar);
		}
	}
	
	public boolean hasSteppableItem()
	{
		return (this.is(RoomTile.ITEM) && this.topItem != null);
	}
}
