package mod_rollers;

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.items.Item;
import org.openhabbo.rooms.instances.RoomAvatar;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomTile;
import org.openhabbo.communication.EServerMessage;
import org.openhabbo.communication.ServerMessage;

public final class RollerEngine
{
	private final static int INTERVAL_TIME = (RoomInstance.PULSE_INTERVAL * 4);

	private long lastTime;
	private final RoomInstance room;

	private final ServerMessage msg;
	private final List<Object> moved;
	
	public RollerEngine(RoomInstance room)
	{
		this.lastTime = System.currentTimeMillis();
		this.room = room;

		// These are re-used all the time
		this.msg = new ServerMessage();
		this.moved = new ArrayList<Object>();
	}

	public void update()
	{
		// Time to roll the rollers?
		if((System.currentTimeMillis() - this.lastTime) < RollerEngine.INTERVAL_TIME)
		{
			return;
		}

		// Search through all the items in the room
		boolean rebuildMap = false;
		for(Item tmpItem : this.room.getAllItems())
		{
			// Is this a roller?
			if(tmpItem instanceof RollerItem)
			{
				// Cast to RollerItem
				RollerItem roller = (RollerItem) tmpItem;

				// Determine next tile
				RoomTile next = roller.getTileInFront();
				boolean nextHasRoller = false;
				boolean nextHasSpecialStance = false;
				if(next != null && !next.hasAvatar)
				{
					// Check the items on the next tile
					List<Item> onNext = this.room.getFloorItemsOnTile(next.x, next.y);
					for(Item item : onNext)
					{
						if(item instanceof RollerItem)
						{
							nextHasRoller = true;
						}
					}
					
					// Move the items on this roller
					for(Item item : this.room.getFloorItemsOnTileSimple(roller.x, roller.y))
					{
						// Not already moved?
						if(item != roller && !moved.contains(item))
						{
							// Determine next height
							float nextZ = (nextHasRoller ? item.z : item.z - roller.definition.height);
							
							// Broadcast update
							this.msg.init(EServerMessage.SlideObjectBundle);
							msg.appendInt(item.x);
							msg.appendInt(item.y);
							msg.appendInt(next.x);
							msg.appendInt(next.y);
							msg.appendInt(1);
							msg.appendInt(item.id);
							msg.appendStr(item.z);
							msg.appendStr(nextZ);
							msg.appendInt(roller.id);
							room.broadcast(msg);
							
							// Update item
							item.x = next.x;
							item.y = next.y;
							item.z = nextZ;
							item.requireDbUpdate();
							
							// Add to list so it doesn't get moved twice in the same cycle
							moved.add(item);
							
							// Map requires updating now!
							rebuildMap = true;
						}
					}
					
					// Valid avatar on the roller?
					if(next.isOpen())
					{
						RoomAvatar avatar = room.getAvatarOnTile(roller.x, roller.y);
						if(avatar != null && !avatar.hasGoal() && !moved.contains(avatar))
						{
							// Determine next height
							float nextZ = (nextHasRoller ? avatar.z : avatar.z - roller.definition.height);
							
							// Broadcast update
							this.msg.init(EServerMessage.SlideObjectBundle);
							msg.appendInt(avatar.x);
							msg.appendInt(avatar.y);
							msg.appendInt(next.x);
							msg.appendInt(next.y);
							msg.appendInt(0);
							msg.appendInt(roller.id);
							msg.appendInt(2);
							msg.appendInt(avatar.avatarId);
							msg.appendStr(avatar.z);
							msg.appendStr(nextZ);
							room.broadcast(msg);
	
							// Update map
							roller.getTile().hasAvatar = false;
							if(!room.getInfo().allowWalkthrough)
							{
								next.hasAvatar = true;
							}
							
							// Update avatar
							avatar.x = next.x;
							avatar.y = next.y;
							avatar.z = nextZ;
							
							// Add to list so it doesn't get moved twice in the same cycle
							moved.add(avatar);
						}
					}
				}
			}
		}

		// Atleast one item moved?
		if(rebuildMap)
		{
			this.room.rebuildFloorMap();
		}

		// Done!
		this.msg.reset();
		this.moved.clear();
		this.lastTime = System.currentTimeMillis();
	}
}