package mod_rollers;

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.communication.EServerMessage;
import org.openhabbo.communication.ServerMessage;
import org.openhabbo.items.Item;
import org.openhabbo.rooms.instances.RoomAvatar;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomTile;

public final class RollerEngine
{
	private final static int INTERVAL_TIME = (RoomInstance.PULSE_INTERVAL * 4);

	private long lastTime;
	private final RoomInstance room;

	private final ServerMessage msg;
	private final List<Object> moved;
	
	public RollerEngine(RoomInstance room)
	{
		this.lastTime = System.currentTimeMillis();
		this.room = room;

		// These are re-used all the time
		this.msg = new ServerMessage();
		this.moved = new ArrayList<Object>();
	}

	public void update()
	{
		// Time to roll the rollers?
		if((System.currentTimeMillis() - this.lastTime) < RollerEngine.INTERVAL_TIME)
		{
			return;
		}

		// Search through all the items in the room
		boolean rebuildMap = false;
		for(Item tmpItem : this.room.getAllItems())
		{
			// Is this a roller?
			if(tmpItem instanceof RollerItem)
			{
				// Cast to RollerItem
				RollerItem roller = (RollerItem) tmpItem;

				// Determine next tile
				RoomTile next = roller.getTileInFront();
				if(room.getMap().tileExists(next) && !room.getMap().avatars[next.x][next.y])
				{
					// Valid items on the roller?
					List<Item> onNext = this.room.getFloorItemsOnTile(next.x, next.y);
					for(Item item : this.room.getFloorItemsOnTile(roller.x, roller.y))
					{
						// Skip the roller itself + already moved items
						if(item != roller && !moved.contains(item))
						{
							// Determine new height
							float height = item.z - roller.definition.height;
							for(Item i : onNext)
							{
								if(i instanceof RollerItem)
								{
									height += i.definition.height;
									break;
								}
							}
							
							// Broadcast update
							this.msg.init(EServerMessage.SlideObjectBundle);
							msg.appendInt(item.x);
							msg.appendInt(item.y);
							msg.appendInt(next.x);
							msg.appendInt(next.y);
							msg.appendInt(1);
							msg.appendInt(item.id);
							msg.appendStr(item.z);
							msg.appendStr(height);
							msg.appendInt(roller.id);
							this.room.broadcast(msg);
							
							// Update item
							item.x = next.x;
							item.y = next.y;
							item.z = height;
							item.requireDbUpdate();
							
							// Add to list so it doesn't get moved twice in the same cycle
							moved.add(item);
							
							// Map requires updating now!
							rebuildMap = true;
						}
					}
					
					// Valid avatar on the roller?
					if(this.room.getMap().tileIsOpen(next.x, next.y))
					{
						RoomAvatar avatar = room.getAvatarOnTile(roller.x, roller.y);
						if(avatar != null && !avatar.hasGoal() && !moved.contains(avatar))
						{
							// Broadcast update
							this.msg.init(EServerMessage.SlideObjectBundle);
							msg.appendInt(avatar.x);
							msg.appendInt(avatar.y);
							msg.appendInt(next.x);
							msg.appendInt(next.y);
							msg.appendInt(0);
							msg.appendInt(roller.id);
							msg.appendInt(2);
							msg.appendInt(avatar.avatarId);
							msg.appendStr(avatar.z);
							msg.appendStr(room.getMap().height[next.x][next.y]);
							this.room.broadcast(msg);
	
							// Update map
							room.getMap().avatars[avatar.x][avatar.y] = false;
							room.getMap().avatars[next.x][next.y] = true;
							
							// Update avatar
							avatar.x = next.x;
							avatar.y = next.y;
							avatar.z = room.getMap().height[next.x][next.y];
							
							// Add to list so it doesn't get moved twice in the same cycle
							moved.add(avatar);
						}
					}
				}
			}
		}

		// Atleast one item moved?
		if(rebuildMap)
		{
			this.room.rebuildFloorMap();
		}

		// Done!
		this.msg.reset();
		this.moved.clear();
		this.lastTime = System.currentTimeMillis();
	}
}