package mod_wired;

import java.util.ArrayList;
import java.util.List;

import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import org.openhabbo.communication.ServerMessage;
import org.openhabbo.core.Logger;
import org.openhabbo.items.Item;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomUser;

@SuppressWarnings({ "unchecked" })
public abstract class WiredItem extends Item
{
	private JSONObject settings;
	
	private int selectionType;
	private List<Integer> selectedIds;
	private String text;
	private List<Integer> params;
	
	private boolean flashToggle;
	
	@Override
	public final void onRoomAdd(RoomInstance room)
	{
		this.flashToggle = false;
		this.parseData();
	}
	
	@Override
	public final void onRoomRemove(RoomInstance room)
	{
		this.data = "";
		this.selectionType = 0;
		this.selectedIds = null;
		this.text = null;
		this.params = null;
	}
	
	public final void setSetting(String key, Object value)
	{
		if(key == null)
		{
			this.settings.remove(key);
		}
		else
		{
			this.settings.put(key, value);
		}
	}

	public final <T> T getSetting(String key)
	{
		if(this.settings != null)
		{
			Object value = this.settings.get(key);
			if(value != null)
			{
				return (T) value;
			}
		}
		return (T)null;
	}
	
	private void parseData()
	{
		this.selectedIds = new ArrayList<Integer>(5);
		this.params = new ArrayList<Integer>(this.getAmountOfParams());
		
		// Any available data?
		if(this.data != null && !this.data.isEmpty())
		{
			try
			{
				// Try parsing the data
				JSONObject data = (JSONObject) new JSONParser().parse(this.data);
				this.selectionType = ((Long) data.get("sel")).intValue();
				for(Long id : (List<Long>)data.get("ids"))
				{
					this.selectedIds.add(id.intValue());
				}
				this.text = (String) data.get("txt");
				for(Long param : (List<Long>)data.get("prm"))
				{
					this.params.add(param.intValue());
				}
				
				// All good!
				return;
			}
			catch (Exception e)
			{
				// Meh, clear all data and revert!
			}
		}
		
		// Parsing failed: set default settings
		this.selectionType = 0;
		this.text = "";
		for(int x = 0; x < this.getAmountOfParams(); x++)
		{
			this.params.add(0);
		}
		this.updateData();
	}
	
	private void updateData()
	{
		JSONObject data = new JSONObject();
		data.put("sel", this.selectionType);
		data.put("ids", this.selectedIds);
		data.put("txt", this.text);
		data.put("prm", this.params);
		this.data = data.toJSONString();
		System.out.println(this.data);
		this.requireDbUpdate();
	}
	
	public void saveSettings()
	{
		this.updateData();
	}
	
	
	protected abstract int getInterface();
	
	protected abstract int getAmountOfParams();
	
	public final List<Integer> getParams()
	{
		return this.params;
	}
	
	public final int getParam(int param)
	{
		if(param >= 0 && param < this.getAmountOfParams())
		{
			return this.params.get(param);
		}
		else
		{
			return 0;
		}
	}
	
	public final void setParam(int param, int value)
	{
		if(param >= 0 && param < this.getAmountOfParams())
		{
			this.params.set(param, value);
		}
	}
	
	public final String getText()
	{
		return this.text;
	}
	
	public final void setText(String text)
	{
		this.text = text;
	}
	
	public final int getSelectionType()
	{
		return this.selectionType;
	}
	
	public final void setSelectionType(int selectionType)
	{
		this.selectionType = selectionType;
	}
	
	public final void clearSelectedItems()
	{
		this.selectedIds.clear();
	}
	
	public final void selectItem(Item item)
	{
		// TODO: check if item supported and not reached max yet
		if(item != null)
		{
			this.selectedIds.add(item.id);
		}
	}
	
	public final Item getSelectedItem()
	{
		List<Item> selected = this.getSelectedItems();
		if(selected.size() >= 1)
		{
			return selected.get(0);
		}
		else
		{
			return null;
		}
	}

	public final List<Item> getSelectedItems()
	{
		List<Item> items = new ArrayList<Item>(1);
		for(int itemId : this.selectedIds)
		{
			Item item = this.getRoom().getItem(itemId);
			if(item != null)
			{
				items.add(item);
			}
		}
		return items;
	}
	
	public final List<Item> getStackedItems()
	{
		return this.getRoom().getFloorItemsOnTile(this.x, this.y);
	}
	
	@Override
	public final void toggle(RoomUser user, int instruction)
	{
		if(user != null && user.hasRoomOwner)
		{
			user.getSession().send(this.serializeDialog());
		}
	}
	
	public final void flash()
	{
		this.setTempState(this.flashToggle ? "0" : "1");
		this.flashToggle = !this.flashToggle;
	}
	
	protected abstract ServerMessage serializeDialog();
}
