package com.fjordev.popbird.internal

import scala.collection._
import scala.concurrent.duration._
import scala.concurrent.ExecutionContext.Implicits.global

import akka.actor.{Actor, ActorLogging}

import twitter4j.{Query, TwitterException}

import com.fjordev.popbird.Account
import AccountActor._

/**
 * continuously drives all the business logic inside an {@link Account}.
 * @param account The account to run.
 * @param clientFactory The client factory for the { @link Application} that this account was authorized with.
 */
class AccountActor(account: Account, clientFactory: TwitterClientFactory) extends Actor with ActorLogging
{
  val twitter = clientFactory.newClient(account)

  val followQueue = mutable.Queue[Long]()

  // Loaded!
  log.info(s"Hello world, my token is ${account.token}")

  // Buzz...
  self ! RefreshProfile

  def receive =
  {
    case RefreshProfile =>
    {
      log.info("Refreshing profile...")
      val user = twitter.verifyCredentials()

      account.screenName = user.getScreenName()
      account.name = user.getName()
      account.profileImageURL = user.getProfileImageURL()
      account.save()

      log.info(s"Refreshed, name=${account.screenName}, img=${account.profileImageURL}")
    }

    case Follow(userId) =>
    {
      val user = twitter.createFriendship(userId)

      log.info(s"Now following @${user.getScreenName}")

      scheduler.scheduleOnce(AccountActor.CheckTime, self, CheckFollowedBack(userId))

      self ! FollowNext
    }

    case Unfollow(userId) =>
    {
      val user = twitter.destroyFriendship(userId)

      log.info(s"No longer following @${user.getScreenName}")
    }

    case CheckFollowedBack(userId) =>
    {
      val status = twitter.showFriendship(account.userId, userId)
      if (status.isSourceFollowedByTarget)
      {
        log.info(s"followed back by @${status.getTargetUserScreenName}, success!")
      }
      else
      {
        log.info(s"Not followed back by @${status.getTargetUserScreenName}...")

        self ! Unfollow(userId)
      }
    }

    case SearchUsers =>
    {
      // Build query, get max results
      val query = new Query(account.keywords.mkString(" OR ")).lang(account.searchLanguage).count(AccountActor.SearchMaxResults)

      // Execute search
      val result = twitter.search(query)
      val tweets = result.getTweets()
      log.info(s"Searched for '${query.getQuery}', got ${tweets.size} results in ${result.getCompletedIn} seconds")

      // Process results
      val timeFormat = new java.text.SimpleDateFormat("dd-MM-yyyy HH:mm:ss")
      for (tweet <- JavaConversions.collectionAsScalaIterable(tweets))
      {
        val user = tweet.getUser()
        log.info(s"At ${timeFormat.format(tweet.getCreatedAt)}, @${user.getScreenName} (${user.getName}) wrote: " +
                 s"'${tweet.getText}'")

        followQueue += user.getId
      }

      // Go!
      self ! FollowNext
    }

    case FollowNext =>
    {
      log.info(s"${followQueue.size} users waiting to be followed")

      followQueue.dequeueFirst(_ > 0) match
      {
        case None => self ! SearchUsers
        case Some(userId) => scheduler.scheduleOnce(AccountActor.FollowDelay, self, Follow(userId))
      }
    }
  }

  def scheduler = context.system.scheduler

  def twitter(block: () => Unit)
  {
    try
    {
      block
    }
    catch
      {
        case tex: TwitterException =>
        {
          tex.printStackTrace()
          if (tex.isCausedByNetworkIssue)
          {
            // do it again
          }
        }
      }
  }
}

import com.typesafe.config.ConfigFactory

object AccountActor
{
  // TODO: use typesafe config
  private val config = ConfigFactory.load()

  /**
   * The delay between two follow events.
   */
  def FollowDelay = 40 seconds

  /**
   * Perform a 'followed back' check after this amount of time.
   */
  def CheckTime = 60 seconds

  /**
   * The maximum amount of results to request and process when performing a tweet search.
   */
  def SearchMaxResults = 100

  /**
   * Reauthenticate and update account data.
   */
  case object RefreshProfile

  /**
   * Schedules a {@link Follow} action for the next user in the target cache.
   */
  case object FollowNext

  /**
   * Perform a tweet search, filter the users and add them to the target cache.
   */
  case object SearchUsers

  /**
   * Attempt to follow a specific user. Schedule a {@link CheckFollowedBack} action on success.
   * @param userId The Twitter user ID of the user to follow.
   */
  case class Follow(userId: Long)

  /**
   * Attempt to unfollow a specific user.
   * @param userId The Twitter user ID of the user to follow.
   */
  case class Unfollow(userId: Long)

  /**
   * Perform a 'followed back by' check by a specific user. {@link Unfollow} the user if he/she has not (yet)
   * back-followed the account.
   * @param userId The Twitter user ID of the user to check.
   */
  case class CheckFollowedBack(userId: Long)

}