package habbod;

import habbod.communication.RequestRegistry;
import habbod.ServerInstance;
import habbod.core.Logger;
import habbod.core.PropertiesBox;
import habbod.core.ShutdownHook;
import habbod.core.StatsMonitor;
import habbod.core.Logger.ELogLevel;
import habbod.game.GameManager;
import habbod.game.chatcommands.ChatCmdRegistry;
import habbod.net.NetServer;
import habbod.remote.RemoteCmdRegistry;
import habbod.remote.RemoteServer;
import habbod.sessions.SessionManager;
import habbod.storage.SQLDatabase;

/**
 * A new Habbo Hotel (www.habbo.com) game server emulator written by Nils 'Nillus' W.
 */
public class ServerInstance
{
	private boolean alive = false;
	private long startTime = 0;
	
	private String propsFile = null;
	private PropertiesBox props = null;
	private SQLDatabase db = null;
	private StatsMonitor monitor = null;
	
	private NetServer gameServer;
	private SessionManager sessionMgr;
	private RemoteServer remoteServer;
	
	private GameManager gameMgr;
	
	public ServerInstance(String propsFile)
	{
		this.alive = false;
		this.startTime = 0;
		this.propsFile = propsFile;
		this.props = new PropertiesBox();
		this.db = null;
		this.monitor = new StatsMonitor(this);
		
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1101111            habbod           1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111  COPYRIGHT (C) 2012 Nils W. 1111011");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println();
	}

	public boolean reloadProperties()
	{
		boolean succes = this.props.load(this.propsFile);
		if(succes)
		{
			Logger.getInstance().info("habbod", String.format("Loaded %d properties.", this.props.size()));
		}
		else
		{
			Logger.getInstance().warning("habbod", "Failed loading system properties.");
		}
		return succes;
	}

	public boolean start()
	{
		// Set log importancy
		Logger.getInstance().setMinImportance(ELogLevel.Info);
		
		// Log startup time
		long beginTime = System.currentTimeMillis();
		Logger.getInstance().info("habbod", "Starting...");

		// Load system properties
		if(!this.reloadProperties())
		{
			return false;
		}

		// Prepare SQLDatabase
		this.db = new SQLDatabase(this.props);
		if(this.db.test())
		{
			Logger.getInstance().info("SQLDatabase", "Connection successful.");
			if(!this.db.getQueryQueue().open())
			{
				return false;
			}
			else
			{
				this.db.getConnMgr().startMonitor();
			}
		}
		else
		{
			Logger.getInstance().warning("SQLDatabase", "Connection failed.");
			return false;
		}
		
		// Load request handlers
		RequestRegistry.reload();
		
		// Load remotecommands
		RemoteCmdRegistry.reload();
		
		// Load chatcommands
		ChatCmdRegistry.reload();
		
		// Load session manager
		this.sessionMgr = new SessionManager();
		
		// Load game manager
		this.gameMgr = new GameManager(this);
		this.gameMgr.start();
		
		// Setup remoteserver for RCON'ish issueing of commands
		int remotePortNumber = ServerInstance.get().getProperties().getInt("remote.portnumber", 1338);
		this.remoteServer = new RemoteServer(remotePortNumber);
		this.remoteServer.addAllowedIp("0:0:0:0:0:0:0:1");
		this.remoteServer.addAllowedIp("0:0:0:0:0:0:0:1%0");
		this.remoteServer.addAllowedIp("127.0.0.1");
		if(!this.remoteServer.start())
		{
			return false;
		}
		
		// Setup network manager for game clients
		int portNumber = this.props.getInt("net.portnumber", 1337);
		this.gameServer = new NetServer();
		this.gameServer.reloadBlacklist();
		if(!this.gameServer.start(portNumber))
		{
			return false;
		}
		
		// Launch stats monitor
		this.monitor.resetStatistics();
		this.monitor.reloadPeakOnlineUsers();
		this.monitor.setUpdateInterval(this.props.getInt("stats.update.seconds", StatsMonitor.UPDATE_INTERVAL_SECONDS));
		this.monitor.start();
		
		// Attach shutdown hook
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));
		
		// Startup okay!
		this.alive = true;
		this.startTime = System.currentTimeMillis();
		Logger.getInstance().info("habbod", String.format("Started in %d ms.", (this.startTime - beginTime)));
		return true;
	}

	public void stop(String reason)
	{
		// Only start if not already shutdown'd yet
		if(!this.alive)
		{
			return;
		}
		this.alive = false;
		Logger.getInstance().info("habbod", String.format("Entering shutdown sequence... (reason: %s)", reason));
		
		// Stop the remote listener
		this.remoteServer.stop();
		
		// Stop the game listener
		this.gameServer.stop();

		// Shutdown game
		this.gameServer.stop();
		this.gameMgr.stop();
		
		// Close any leftover connections
		this.gameServer.close();
		
		// Stop monitor and reset statistics
		this.monitor.interrupt();
		this.monitor.resetStatistics();

		// Close SQLDatabase
		this.db.getQueryQueue().close();
		this.db.getConnMgr().closeAll();
		this.db.getConnMgr().stopMonitor();

		// Sequence complete!
		Logger.getInstance().info("habbod", "Shutdown sequence completed.");
	}

	public String versionString()
	{
		return "habbod v0.9.5-dev 'Palladium'";
	}

	public boolean alive()
	{
		return this.alive;
	}

	public long getStartTime()
	{
		return this.startTime;
	}

	public PropertiesBox getProperties()
	{
		return this.props;
	}

	public SQLDatabase getDb()
	{
		return this.db;
	}
	
	public StatsMonitor getStatsMonitor()
	{
		return this.monitor;
	}
	
	public long getMemoryUsage()
	{
		Runtime runtime = Runtime.getRuntime();
		return (runtime.totalMemory() - runtime.freeMemory()) / 1024;
	}
	
	public NetServer getGameServer()
	{
		return this.gameServer;
	}
	
	public SessionManager getSessionMgr()
	{
		return this.sessionMgr;
	}
	
	public RemoteServer getRemoteServer()
	{
		return this.remoteServer;
	}
	
	public GameManager getGameMgr()
	{
		return this.gameMgr;
	}
	
	/* Singleton */
	private static ServerInstance instance;

	public final static ServerInstance get()
	{
		return instance;
	}

	public static ServerInstance create(String propsFile)
	{
		if(instance == null)
		{
			instance = new ServerInstance(propsFile);
			Logger.getInstance().info("habbod", instance.versionString());
		}
		
		return instance;
	}
}
