package habbod.game.catalog;

import habbod.ServerInstance;
import habbod.core.Logger;
import habbod.game.items.InventoryItem;
import habbod.game.items.ItemBase;
import habbod.game.items.ItemDefinition;
import habbod.game.items.ItemManager;
import habbod.game.items.pets.PetManager;
import habbod.game.items.pets.PetUtil;
import habbod.game.users.User;
import habbod.game.util.GameUtil;
import habbod.sessions.Session;
import habbod.sessions.SessionManager;
import habbod.storage.DatabaseResult;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public final class Catalog
{
	private Map<Integer, CatalogPage> pages;
	private Map<Integer, CatalogProduct> products;

	public Catalog()
	{
		this.pages = Collections.emptyMap();
		this.products = Collections.emptyMap();
	}

	public void reload()
	{
		// Create new collections
		Map<Integer, CatalogPage> tmpPages = new HashMap<Integer, CatalogPage>();
		Map<Integer, CatalogProduct> tmpProducts = new HashMap<Integer, CatalogProduct>();

		// Load all the pages
		DatabaseResult result = ServerInstance.get().getDb().query("SELECT * FROM catalog_pages ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogPage page = new CatalogPage();
			page.id = result.getInt("id");
			page.parentId = result.getInt("parent_id");
			page.name = result.getString("name");
			page.requiredRight = result.getString("required_right");
			if(page.requiredRight == null || page.requiredRight.isEmpty())
			{
				page.requiredRight = "default";
			}
			page.isEnabled = result.getBoolean("enabled");
			page.isCategory = result.getBoolean("is_category");
			page.iconColor = result.getInt("icon_color");
			page.iconImage = result.getInt("icon_image");
			page.layout = result.getString("layout");
			page.headlineImage = result.getString("headline_image");
			page.teaserImage = result.getString("teaser_image");
			page.specialImage = result.getString("special_image");
			page.text1 = result.getString("text1");
			page.text2 = result.getString("text2");
			page.details = result.getString("details");
			page.teaserText = result.getString("teaser_text");
			tmpPages.put(page.id, page);
		}
		result.release();

		// Link the pages in parent/child fashion
		for(CatalogPage page : tmpPages.values())
		{
			if(page.parentId > 0)
			{
				CatalogPage parent = tmpPages.get(page.parentId);
				if(parent == null)
				{
					Logger.getInstance().warning("Catalog", String.format("Page #%d is a child of page #%d, which is missing. Page not attached to parent.", page.id, page.parentId));
				}
				else
				{
					parent.pages.add(page);
				}
			}
		}

		// Load the products, grab the item definitions and attach them to the pages
		result = ServerInstance.get().getDb().query("SELECT * FROM catalog_products ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogProduct product = new CatalogProduct();
			product.id = result.getInt("id");
			product.pageId = result.getInt("page_id");
			product.code = result.getString("code");
			product.itemDefinitionId = result.getInt("definition_id");
			product.itemAmount = result.getInt("amount");
			product.costCredits = result.getInt("cost_credits");
			product.costPixels = result.getInt("cost_pixels");
			product.costQuestPoints = result.getInt("cost_questpoints");
			product.itemDefinition = ItemManager.getInstance().getDefinition(product.itemDefinitionId);
			if(product.itemDefinition == null)
			{
				if(product.itemDefinitionId > 0)
				{
					Logger.getInstance().warning("Catalog", String.format("Product #%d ('%s') refers to the missing item definition #%d. Product ignored.", product.id, product.code, product.itemDefinitionId));
				}
			}
			else
			{
				if(product.pageId > 0)
				{
					CatalogPage page = tmpPages.get(product.pageId);
					if(page == null)
					{
						Logger.getInstance().warning("Catalog", String.format("Product #%d ('%s') is supposed to appear at page #%d, which is missing. Product not attached to page.", product.id, product.code, product.pageId));
					}
					else
					{
						product.page = page;
						page.products.add(product);
					}
				}
				tmpProducts.put(product.id, product);
			}
		}
		result.release();

		// Set instance fields
		this.pages = Collections.unmodifiableMap(tmpPages);
		this.products = Collections.unmodifiableMap(tmpProducts);

		// Done!
		Logger.getInstance().info("Catalog", String.format("Reloaded %d pages and %d products.", this.pages.size(), this.products.size()));
	}

	public CatalogPage getPage(int pageId)
	{
		return this.pages.get(pageId);
	}

	public List<CatalogPage> getVisibleIndexPagesForUser(User user)
	{
		List<CatalogPage> visible = new ArrayList<CatalogPage>();
		for(CatalogPage page : this.pages.values())
		{
			if(page.parentId == -1 && page.isVisibleTo(user))
			{
				visible.add(page);
			}
		}
		return visible;
	}

	public Collection<CatalogPage> getAllPages()
	{
		return this.pages.values();
	}

	public CatalogProduct getProduct(int productId)
	{
		return this.products.get(productId);
	}

	public Collection<CatalogProduct> getAllProducts()
	{
		return this.products.values();
	}

	public void billUser(Session session, CatalogProduct product)
	{
		if(product.isPurchasableBy(session.getUser()))
		{
			// Payday!
			if(product.costCredits > 0)
			{
				session.getUser().credits -= product.costCredits;
				session.sendCreditBalance();
			}
			if(product.costPixels > 0)
			{
				session.getUser().pixels -= product.costPixels;
				session.sendPixelBalance();
			}
			session.getUser().saveCreditsAndPixels();

			// Write transaction log
			/*
			if(product.costCredits > 0)
			{
				UserManager.getInstance().logCreditTransaction(session.getUser().id, -product.costCredits, "catalog product (#" + product.id + ")");
			}*/
		}
	}
	
	public List<InventoryItem> purchaseProduct(CatalogProduct product, User user, String purchaseData, boolean processPurchaseData)
	{
		// Valid args?
		if(product == null || user == null || product.itemDefinition == null)
		{
			return null;
		}

		// Process purchase data
		if(processPurchaseData)
		{
			try
			{
				purchaseData = this.processDataToState(product, user, purchaseData);
			}
			catch(Exception ex)
			{
				purchaseData = ItemBase.ERROR_DATA;
			}
			if(purchaseData != null && purchaseData.equals(ItemBase.ERROR_DATA))
			{
				Logger.getInstance().warning("Catalog", String.format("Purchase of product #%d by user #%d failed, bad purchasedata.", product.id, user.id));
				return null;
			}
		}
		
		// Create the shipping items
		List<InventoryItem> shipping = new ArrayList<InventoryItem>(1);
		
		// Take care of AvatarEffects
		if(product.code.equals("AvatarEffect"))
		{
			Session session = SessionManager.getInstance().getSessionByUserId(user.id);
			if(session != null)
			{
				session.getEffectInventory().addEffect(1337, 3600);
			}
			return shipping;
		}
		
		// Take care of pets
		else if(product.code.equals("Pet"))
		{
			// Parse pet data
			String[] parts = purchaseData.split("\n", 3);
			String name = parts[0];
			int breed = Integer.parseInt(parts[1]);
			String color = parts[2];
			int type = PetUtil.productNameToPetType(product.itemDefinition.sprite);
			
			// Create the pet for this user, will appear in inventory
			PetManager.getInstance().createPet(user.id, name, type, breed, color);
			
			// TODO: add free food item to shipping (sample!)
			return shipping;
		}
		
		// Start creating the items
		for(int x = 0; x < product.itemAmount; x++)
		{
			// Create the items
			InventoryItem item = new InventoryItem();
			item.definition = product.itemDefinition;
			item.state = "0";
			if(item.create(user.id))
			{
				shipping.add(item);
				
				// Set post.it default color
				if(item.definition.equals("PostIt"))
				{
					item.data = "FFFF33";
				}
				
				// Teleporters get a second, linking item
				else if(item.definition.equals("Teleporter"))
				{
					// Create the second item
					InventoryItem item2 = new InventoryItem();
					item2.definition = item.definition;
					item2.data = Integer.toString(item.id);
					if(item2.create(user.id))
					{
						shipping.add(item2);
						
						// Update the first item with a reference to the second item
						item.data = Integer.toString(item2.id);
						item.update(user.id);
					}
				}
			}
		}    

		// These are the items that are shipped
		return shipping;
	}
	

	public boolean purchaseAsGift(CatalogProduct product, User user, String purchaseData, int receiverId, String giftNote)
	{
		// Run pre-purchase logic, filter & apply purchase data
		try
		{
			purchaseData = this.processDataToState(product, user, purchaseData);
		}
		catch(Exception ex)
		{
			purchaseData = ItemBase.ERROR_DATA;
		}
		if(purchaseData != null && purchaseData.equals(ItemBase.ERROR_DATA))
		{
			Logger.getInstance().warning("mod_gift", String.format("Gift-purchase of product #%d by user #%d failed, bad purchase data.", product.id, user.id));
			return false;
		}

		// Get all presentbox definitions, shuffle them
		List<ItemDefinition> availableBoxes = Collections.emptyList();//ItemManager.getInstance().getDefinitionsByClass(PresentBoxItem.class);
		
		// Any definitions?
		if(availableBoxes.isEmpty())
		{
			Logger.getInstance().warning("gift", "No item definitions with class PresentBoxItem available! Gifts cannot be purchased.");
			return false;
		}
		
		// Create presentbox, state=giftnote, data=purchasedata
		InventoryItem item = GameUtil.getRandomElement(availableBoxes).newItem();
		item.state = "!" + giftNote;
		item.data = (product.id + GameUtil.DATA_DELIMITER + purchaseData);
		ItemManager.getInstance().storeItem(item, receiverId);

		// Is the receiver online?
		Session receiver = SessionManager.getInstance().getSessionByUserId(receiverId);
		if(receiver != null)
		{
			// Add the item to the inventory component
			receiver.getItemInventory().addItem(item, false);
			receiver.getItemInventory().refresh();

			// Send out a notification (TODO: localize this?)
			receiver.sendNotif("You have received a gift!");
		}

		// Purchase OK!
		return true;
	}
	
	private String processDataToState(CatalogProduct product, User user, String purchaseData)
	{
		// Parse decoration data
		if(product.itemDefinition.equals("Decoration"))
		{
			if(product.itemDefinition.data.equals("landscape"))
			{
				return Float.toString(Float.parseFloat(purchaseData));
			}
			else
			{
				return Integer.toString(Integer.parseInt(purchaseData));
			}
		}
		
		// Inscription for prizetrophies
		else if(product.itemDefinition.is('T'))
		{
			return(user.username + (char) 9 + GameUtil.getCurrentDate() + (char) 9 + (purchaseData));
		}
		
		
		// Nothing special, discard state
		return "";
	}

	/* Begin singleton */
	private final static Catalog instance = new Catalog();

	public final static Catalog getInstance()
	{
		return instance;
	}
	/* End singleton */
}
