package habbod.game.rooms.instances;

import habbod.communication.EServerMessage;
import habbod.communication.EServerMessageAlternate;
import habbod.communication.ServerMessage;
import habbod.communication.messages.outgoing.room.engine.RoomUserEffectMessageComposer;
import habbod.game.users.User;
import habbod.game.users.effects.EffectInventory;
import habbod.game.util.GameUtil;
import habbod.sessions.Session;

public final class RoomUser extends RoomHuman
{
	private final static int maxIdleAwakeTime = GameUtil.secondsToRoomPulses(5 * 60); // 5 minutes of idling before sleeping
	private final static int maxIdleSleepTime = (maxIdleAwakeTime + GameUtil.secondsToRoomPulses(10 * 60)); // 5 minutes idling -> sleep, 10 more minutes sleeping -> kick

	private Session session;
	public boolean hasRoomOwner;
	public boolean hasRoomRights;
	
	public int danceId;
	private int roomEffectId;
	private int ownEffectId;
	public int carryItemId;
	public int carryItemTimer;
	
	private int warpPhase = 0;
	private RoomTile warpDestination;
	
	private long chatLastMessageTime;
	private int chatFastMessageCount;
	
	public RoomUser(Session session)
	{
		this.session = session;
		this.pathfindingAllowedOn = (RoomTile.OPEN | RoomTile.WATER);
		
		// Session has effect?
		EffectInventory effects = session.getEffectInventory();
		if(effects != null && effects.getCurrentEffect() > 0)
		{
			session.getRoomUser().setEffect(effects.getCurrentEffect(), true);
		}
	}
	
	public Session getSession()
	{
		return this.session;
	}
	
	public User getInfo()
	{
		return this.session.getUser();
	}
	
	@Override
	protected void onChatFromAvatar(RoomAvatar from, String text, boolean isShout)
	{
		// Not on ignore list?
		if(true)
		{
			this.lookTo(from.getTile(), true);
		}
	}
	
	@Override
	public void chat(String text, boolean isShout)
	{
		// Chat timeout in effect?
		long nowTime = System.currentTimeMillis();
		if(this.session.chatTimeoutExpireTime > 0)
		{
			// Calculate remaining seconds
			int chatTimeoutSeconds = (int) ((this.session.chatTimeoutExpireTime - nowTime) / 1000);
			if(chatTimeoutSeconds > 0)
			{
				// Notify client
				ServerMessage msg = new ServerMessage(EServerMessageAlternate.ROOM_MUTED);
				msg.appendInt(chatTimeoutSeconds);
				this.getSession().send(msg);
				
				// Block chat
				return;
			}
		}
		
		// Increment count if there is less than 1.5 seconds between chat messages
		long timeBetween = nowTime - this.chatLastMessageTime;
		if(timeBetween > 1500)
		{
			this.chatFastMessageCount = 0;
		}
		else
		{
			this.chatFastMessageCount++;
		}
		
		// Forward chat
		super.chat(text, isShout);
		this.chatLastMessageTime = nowTime;
		this.session.chatTimeoutExpireTime = 0;
		
		// Sent 3 or more fast messages in a row?
		if(this.chatFastMessageCount >= 3)
		{
			// Give 30 seconds timeout?
			if(this.getSession().getUser().role.rights.contains("enable_floodtimeout"))
			{
				this.session.chatTimeoutExpireTime = nowTime + (30 * 1000);
			}
		}
	}
	
	public void pulse()
	{
		// Call RoomAvatar pulse() logic
		super.pulse();
		
		// Deal with carry item
		if(this.carryItemId > 0)
		{
			this.carryItemTimer--;
			if(this.carryItemTimer <= 0)
			{
				this.stopCarryItem();
			}
		}
		
		// Not sleeping?
		if(!this.isAsleep)
		{
			// Needs to go to sleep?
			if(this.idleTime > maxIdleAwakeTime)
			{
				// Set the sleep flag so new incoming users see it too
				this.isAsleep = true;

				// Notify the room that this avatar went asleep!
				ServerMessage msg = new ServerMessage(EServerMessage.Sleep);
				msg.appendInt(this.avatarId);
				msg.appendBool(true);
				this.getRoom().broadcast(msg);
			}
		}
		else
		{
			// Sleeping long enough for kicking?
			if(this.idleTime > maxIdleSleepTime)
			{
				this.getSession().kickFromRoom();
			}
		}
		
		// Deal with warping
		if(this.warpPhase > 0)
		{
			if(this.warpPhase == 4)
			{
				// Enable sparkles
				this.getRoom().broadcast(RoomUserEffectMessageComposer.compose(this, 21));
			}
			else if(this.warpPhase == 3)
			{
				// WARP
			}
			else if(this.warpPhase == 2)
			{
				// Disable sparkles
				this.getRoom().broadcast(RoomUserEffectMessageComposer.compose(this, 0));
			}
			this.warpPhase--;
		}
	}
	
	@Override
	public void unidle()
	{
		super.unidle();
		
		// User was sleeping?
		if(this.isAsleep)
		{
			// Not sleeping anymore
			this.isAsleep = false;
			
			// Notify room
			ServerMessage msg = ServerMessage.create(EServerMessage.Sleep);
			msg.appendInt(this.avatarId);
			msg.appendBool(false);
			this.getRoom().broadcast(msg);
		}
	}
	
	public void carryItem(int itemId)
	{
		// Set the ID + timer
		this.carryItemId = itemId;
		if(itemId > 0)
		{
			this.carryItemTimer = 240;
		}
		else
		{
			this.carryItemTimer = 0;
		}

		// Notify the roominstance
		ServerMessage msg = new ServerMessage(EServerMessage.CarryObject);
		msg.appendInt(this.avatarId);
		msg.appendInt(this.carryItemId);
		this.getRoom().broadcast(msg);
	}
	
	public void stopCarryItem()
	{
		this.carryItem(0);
	}
	
	public void setEffect(int effectId, boolean isOwn)
	{
		// Set the effect
		if(isOwn)
		{
			this.ownEffectId = effectId;
		}
		else
		{
			this.roomEffectId = effectId;
		}
		
		// Notify the room
		ServerMessage msg = new ServerMessage(485);
		msg.appendInt(this.avatarId);
		msg.appendInt(this.getCurrentEffect());
		msg.appendBool(false);
		this.getRoom().broadcast(msg);
	}
	
	public int getCurrentEffect()
	{
		if(this.roomEffectId > 0)
		{
			return this.roomEffectId;
		}
		else
		{
			return this.ownEffectId;
		}
	}
	
	public void scheduleWarp(RoomTile dest)
	{
		// Enable effect
		
		// Set timer etc
		this.warpPhase = 4;
		this.warpDestination = dest;
	}
	
	@Override
	public void serialize(ServerMessage msg)
	{
		User user = this.getInfo();
		msg.appendInt(user.id);
		msg.appendStr(user.username);
		msg.appendStr(user.motto);
		msg.appendStr(user.figure);
		msg.appendInt(this.avatarId);
		msg.appendInt(this.tile.x);
		msg.appendInt(this.tile.y);
		msg.appendStr(this.tile.height);
		msg.appendInt(2);
		msg.appendInt(1);
		msg.appendStr(user.gender);
		msg.appendInt(-1);
		msg.appendInt(-1);
		msg.appendInt(-1);
		msg.appendStr("");
		msg.appendInt(0);
	}

	@Override
	public String getAvatarKey()
	{
		return "u" + this.session.getUser().id;
	}
}
