package habbod.game.rooms.items.floor;

import habbod.game.items.Item;
import habbod.game.items.ItemManager;
import habbod.game.items.UsableItem;
import habbod.game.rooms.instances.RoomTile;
import habbod.game.rooms.instances.RoomUser;
import habbod.game.util.GameUtil;

import java.util.List;


import static habbod.core.Localization._I;

public final class TeleporterItem extends UsableItem
{
	private boolean isWaiting = false;
	
	@Override
	public void onPurchase(List<Item> shipping)
	{
		// Store this teleporter, get unique ID
		ItemManager.getInstance().storeItem(this);
		
		// Create new teleporter, link to this teleporter ID
		Item other = this.definition.newInstance();
		other.ownerId = this.ownerId;
		other.data = Integer.toString(this.id);
		ItemManager.getInstance().storeItem(other);

		// Update this teleporter with ID of other teleporter
		this.data = Integer.toString(other.id);
		this.update();

		// Add both teleporters to shipping
		shipping.add(this);
		shipping.add(other);
	}

	@Override
	public boolean onToggle(RoomUser user, int instruction)
	{
		// Can go in teleporters?
		if(user == null || !user.canMove() || this.inUse())
		{
			return false;
		}

		// Infront of teleporter? Attempt to move to the front
		RoomTile frontTile = this.getTileInFront();
		if(!user.at(frontTile))
		{
			user.setGoal(frontTile);
			return false;
		}
		
		// Tadaa
		this.setUser(user);

		// Lock own movement, move into the teleporter
		user.lockMovement();
		user.overrideNextTile = true;
		user.setGoal(this.getTile());

		// Wait...
		this.setTimeout(3);
		
		return true;
	}

	@Override
	protected void onTimeout()
	{
		// User inside?
		if(this.inUse())
		{
			// Get user
			RoomUser user = this.getUser();
			
			// Get out?
			if(user.getSession().authenticatedTeleporterId == this.id)
			{
				user.getSession().authenticatedTeleporterId = 0;
				this.kickUserOut();
				return;
			}
			
			// Get tele info
			int otherId = GameUtil.parseInt(this.data);
			int otherRoomId = ItemManager.getInstance().getTeleporterRoomId(otherId);
			
			// Waiting for teleport start?
			if(this.isWaiting)
			{
				// Same room?
				if(otherRoomId == this.getRoom().getInfo().id)
				{
					// Warp to the other tele
					TeleporterItem other = (TeleporterItem)this.getRoom().getFloorItem(otherId);
					user.headRotation = user.bodyRotation = other.rotation;
					user.updatePosition(other.getTile(), false);
					
					// Open the door in the other tele
					(other).openDoorTemporarily();
					
					// Get out!
					user.unlockMovement();
					user.setGoal(other.getTileInFront());
				}
				else
				{
					// Teleporting!
					user.getSession().authenticatedTeleporterId = otherId;
					user.getSession().tryGetRoomAccess(otherRoomId, "");
				}
				
				// Waiting done! Done with this teleporter
				this.isWaiting = false;
				this.clearUser();
			}
			
			// User just entered teleporter?
			else if(user.at(this.getTile()))
			{
				// Linked teleporter in room?
				if(otherRoomId == 0)
				{
					user.getSession().sendNotif(_I("This teleporter is currently not linked."));
					this.kickUserOut();
				}
				else
				{
					// Flash this teleporter, wait 2 pulses before the actual teleport
					this.isWaiting = true;
					this.setTempState("2");
					this.setTimeout(3);
				}
			}
		}
		else
		{
			this.setState("0");
		}
	}

	private void openDoorTemporarily()
	{
		this.setTempState("1");
		this.setTimeout(2);
	}

	private void kickUserOut()
	{
		RoomUser user = this.getUser();
		if(user != null)
		{
			// Open the door...
			this.clearUser();
			this.openDoorTemporarily();

			// Start move to tile infront
			user.unlockMovement();
			user.setGoal(this.getTileInFront());
		}
	}
	
	public void arriveUser(RoomUser user)
	{
		this.setUser(user);
		this.setTimeout(2);
	}
}
