package habbod.game.rooms.items.wired;

import habbod.communication.ServerMessage;
import habbod.game.rooms.instances.RoomInstance;
import habbod.game.rooms.instances.RoomUser;
import habbod.game.rooms.items.RoomFloorItem;
import habbod.game.rooms.items.wired.addons.*;
import habbod.game.util.GameUtil;

import java.util.ArrayList;
import java.util.List;



public abstract class WiredTrigger extends WiredItem
{
	private long lastTime;
	protected abstract boolean suppliesUser();
	
	public final void execute(RoomUser user)
	{
		// Infinite loop prevention
		if((System.currentTimeMillis() - lastTime) < RoomInstance.PULSE_INTERVAL)
		{
			return;
		}
		this.lastTime = System.currentTimeMillis();
		
		// Fire!
		this.flash();
		
		// Get all items in the stack
		List<RoomFloorItem> items = this.getStackedItems();

		// Search for Addons +  Conditions
		boolean useRandomAction = false;
		boolean useUnusedAction = false;
		for(RoomFloorItem item : items)
		{
			// Is this a Condition?
			if(item instanceof WiredCondition)
			{
				// Evaluate the Condition
				WiredCondition condition = (WiredCondition)item;
				if(condition.evaluate(user) == true)
				{
					condition.flash();
				}
				else
				{
					return;
				}
			}
			else if(item instanceof WiredRandomActionAddon)
			{
				useRandomAction = true;
			}
			else if(item instanceof WiredUnusedActionAddon)
			{
				useUnusedAction = true;
			}
		}
		
		// Execute the proper actions
		if(useRandomAction)
		{
			// Pick a random action
			List<WiredAction> actions = new ArrayList<WiredAction>(2);
			for(RoomFloorItem item : items)
			{
				if(item instanceof WiredAction)
				{
					actions.add((WiredAction)item);
				}
			}
			
			/// ... and execute it!
			this.executeAction(GameUtil.getRandomElement(actions), user);
		}
		else
		{
			for(RoomFloorItem item : items)
			{
				if(item instanceof WiredAction)
				{
					this.executeAction((WiredAction)item, user);
				}
			}
		}
	}
	
	private void executeAction(WiredAction action, RoomUser user)
	{
		if(action != null && (user != null || !action.requiresUser()))
		{
			action.flash();
			if(action.getDelay() == 0)
			{
				action.execute(user);
			}
			else
			{
				action.execute(user, action.getDelay());
			}
		}
	}
	
	public List<WiredItem> getIncompatibleActions()
	{
		List<WiredItem> incompatible = new ArrayList<WiredItem>(0);
		
		// This trigger doesn't supply a user?
		if(!this.suppliesUser())
		{
			// Search all items in the stack
			for(RoomFloorItem item : this.getStackedItems())
			{
				// Item is action?
				if(item instanceof WiredAction)
				{
					// Action is incompatible?
					WiredAction action = (WiredAction)item;
					if(action.requiresUser())
					{
						incompatible.add(action);
					}
				}
			}
		}
		
		return incompatible;
	}
	
	@Override
	public ServerMessage serializeDialog()
	{
		// Aggregrate lists
		List<RoomFloorItem> selected = this.getSelectedItems();
		List<WiredItem> conflicts = this.getIncompatibleActions();
		
		// Serialize info
		ServerMessage msg = new ServerMessage(650);
		msg.appendBool(false); // Disable advanced selection options
		msg.appendInt(WiredItem.SELECT_LIMIT);
		msg.appendInt(selected.size());
		for(RoomFloorItem sel : selected)
		{
			msg.appendInt(sel.id);
		}
		msg.appendInt(this.definition.displayId);
		msg.appendInt(this.id);
		msg.appendStr(this.getText());
		msg.appendInt(this.getParams().length);
		for(int param : this.getParams())
		{
			msg.appendInt(param);
		}
		msg.appendInt(this.getSelectionType());
		msg.appendInt(this.getInterface());
		msg.appendInt(conflicts.size());
		for(WiredItem item : conflicts)
		{
			msg.appendInt(item.definition.displayId);
		}
		msg.appendStr(""); // Unknown
		
		return msg;
	}
}
