package net.nillus.habbo.game.items.impl;

import java.util.List;

import net.nillus.habbo.game.items.Item;
import net.nillus.habbo.game.rooms.instances.RoomInstance;
import net.nillus.habbo.game.util.GameUtil;
import net.nillus.habbo.sessions.Session;

public final class RoomDimmerItem extends Item
{
	private final static int PRESET_AMOUNT = 3;

	public boolean enabled;
	public int currentPreset;
	private Preset[] presets;

	@Override
	public void onPurchase(List<Item> shipping)
	{
		this.reset();
		super.onPurchase(shipping);
	}
	
	@Override
	protected void onRoomAdd(RoomInstance room)
	{
		room.setRoomDimmerItem(this);
	}

	@Override
	protected void onRoomRemove(RoomInstance room)
	{
		room.setRoomDimmerItem(null);
	}

	@Override
	public boolean onPlace(Session session, RoomInstance room)
	{
		if(room.getRoomDimmer() != null)
		{
			session.sendNotif("You can only place one roomdimmer in a room!");
			return false;
		}
		else
		{
			return true;
		}
	}

	public void saveSettings()
	{
		this.data = this.generateData();
		this.setState(this.generateState());
	}

	public void reset()
	{
		this.enabled = false;
		this.currentPreset = 1;
		this.presets = new Preset[RoomDimmerItem.PRESET_AMOUNT];
		for(int x = 0; x < this.presets.length; x++)
		{
			Preset preset = new Preset();
			preset.id = (x + 1);
			preset.color = Preset.DEFAULT_COLOR;
			preset.intensity = 255;
			preset.backgroundOnly = false;
			this.presets[x] = preset;
		}
		this.saveSettings();
	}

	public Preset getPreset(int presetId)
	{
		for(Preset preset : this.presets)
		{
			if(preset.id == presetId)
			{
				return preset;
			}
		}
		return null;
	}

	public Preset[] getPresets()
	{
		return this.presets;
	}

	private String generateState()
	{
		StringBuilder sb = new StringBuilder();
		Preset preset = this.getPreset(this.currentPreset);
		sb.append(this.enabled ? '2' : '1');
		sb.append(',');
		sb.append(this.currentPreset);
		sb.append(',');
		sb.append(preset.backgroundOnly ? '2' : '1');
		sb.append(',');
		sb.append(preset.color);
		sb.append(',');
		sb.append(preset.intensity);
		return sb.toString();
	}

	private String generateData()
	{
		StringBuilder sb = new StringBuilder();
		sb.append(this.enabled ? '1' : '0');
		sb.append(GameUtil.DATA_DELIMITER);
		sb.append(this.currentPreset);
		sb.append(GameUtil.DATA_DELIMITER);
		sb.append(this.presets.length);
		for(Preset preset : this.presets)
		{
			sb.append(GameUtil.DATA_DELIMITER);
			sb.append(preset.id);
			sb.append(GameUtil.DATA_DELIMITER);
			sb.append(preset.color);
			sb.append(GameUtil.DATA_DELIMITER);
			sb.append(preset.intensity);
			sb.append(GameUtil.DATA_DELIMITER);
			sb.append(preset.backgroundOnly ? '1' : '0');
		}
		return sb.toString();
	}

	private void reloadData()
	{
		String[] bits = this.data.split(String.valueOf(GameUtil.DATA_DELIMITER));
		int n = 0;

		// Construct presets and settings from data
		try
		{
			this.enabled = (bits[n++].equals("1"));
			this.currentPreset = Integer.parseInt(bits[n++]);
			this.presets = new Preset[Integer.parseInt(bits[n++])];
			for(int x = 0; x < this.presets.length; x++)
			{
				Preset preset = new Preset();
				preset.id = Integer.parseInt(bits[n++]);
				preset.color = bits[n++];
				preset.intensity = Integer.parseInt(bits[n++]);
				preset.backgroundOnly = (bits[n++].equals("1"));
				this.presets[x] = preset;
			}

			// Generate the state from the new presets
			this.state = this.generateState();
		}
		catch(Exception ex)
		{
			// M'kay this is bad
			this.reset();
		}
	}
	
	public final static class Preset
	{
		private final static String[] ALLOWED_COLORS = { "#000000", "#0053F7", "#EA4532", "#82F349", "#74F5F5", "#E759DE", "#F2F851" };
		public final static String DEFAULT_COLOR = ALLOWED_COLORS[0];

		public int id;
		public String color;
		public int intensity;
		public boolean backgroundOnly;

		public static boolean isValidColor(String color)
		{
			for(int x = 0; x < ALLOWED_COLORS.length; x++)
			{
				if(ALLOWED_COLORS[x].equals(color))
				{
					return true;
				}
			}
			return false;
		}
		
		public static boolean isValidIntensity(int intensity)
		{
			return (intensity >= 0 &&  intensity <= 255);
		}
	}

}
