package net.nillus.habbo.game.rooms.instances;

import java.util.List;

import net.nillus.habbo.game.items.Item;
import net.nillus.habbo.game.items.wired.WiredTrigger;
import net.nillus.habbo.util.FlagsUtil;

public final class RoomTile
{
	public final static byte OPEN = (1 << 0);
	public final static byte CLOSED = (1 << 1);
	public final static byte LASTSTEP = (1 << 2);
	public final static byte WATER = (1 << 3);
	public final static RoomTile OUT_OF_RANGE = new RoomTile(-1, -1); 
	
	public final int x;
	public final int y;
	public byte state;
	public float height;
	public boolean hasAvatar;
	public Item topItem;
	public RoomTile redirect;
	public int effectId;
	
	public WiredTrigger trigger;
	
	private List<Item> items;
	
	public RoomTile(int x, int y)
	{
		this.x = x;
		this.y = y;
	}
	
	public boolean is(byte state)
	{
		return FlagsUtil.isSet(this.state, state);
	}
	
	public boolean touches(RoomTile tile)
	{
		return(Math.abs(this.x - tile.x) <= 1 && Math.abs(this.y - tile.y) <= 1);
	}
	
	public boolean isOpen()
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}
		
		// Not open?
		if(!this.is(RoomTile.OPEN))
		{
			return false;
		}
		
		// Open!
		return true;
	}
	
	public boolean isPassable(byte allowedStates, boolean lastStep)
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}

		// Normal tile?
		if(!this.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if(!FlagsUtil.isSet(allowedStates, this.state))
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		
		// Blocked?
		if(this.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (this.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
	
	public void clear(char defaultState)
	{
		this.topItem = null;
		this.redirect = null;
		this.effectId = 0;
		if(Character.isDigit(defaultState))
		{
			this.state = RoomTile.OPEN;
			this.height = Integer.parseInt(Character.toString(defaultState));
		}
		else
		{
			this.state = RoomTile.CLOSED;
			this.height = 0.0f;
		}
		this.items.clear();
	}
	
	public float getTopHeight()
	{
		if(this.topItem != null)
		{
			return this.topItem.z + this.topItem.definition.height;
		}
		else
		{
			return this.height;
		}
	}
	
	public void stepOn(RoomAvatar avatar)
	{
		this.hasAvatar = true;
		
		// Position
		avatar.tile = this;
		
		// Item stepping
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOn(avatar);
		}
		
		// Effect handling
		if(avatar instanceof RoomUser)
		{
			RoomUser user = (RoomUser)avatar;
			if(this.effectId != user.getCurrentEffect())
			{
				user.setEffect(this.effectId, false);
			}
		}
	}
	
	public void stepOff(RoomAvatar avatar)
	{
		this.hasAvatar = false;
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOff(avatar);
		}
	}
	
	public void addItem(Item item)
	{
		if(item.z > this.height)
		{
			this.height = item.z;
		}
		this.items.add(item);
	}
	
	public List<Item> getItems()
	{
		return this.items;
	}
	
	public boolean equals(RoomTile tile)
	{
		return (tile != null && tile.x == this.x && tile.y == this.y);
	}
	
	@Override
	public String toString()
	{
		return ("[" + this.x + "," + this.y + "]");
	}
}
