package net.nillus.habbo.game.rooms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import net.nillus.habbo.habbod;
import net.nillus.habbo.core.Logger;
import net.nillus.habbo.events.Event;
import net.nillus.habbo.game.rooms.instances.RoomInstance;
import net.nillus.habbo.game.rooms.instances.RoomWorker;
import net.nillus.habbo.storage.DatabaseResult;
import net.nillus.habbo.storage.SQLHelper;

public final class RoomManager
{
	// Events
	public final static Event onGuestRoomCreate = new Event("RoomManager.OnGuestRoomCreate");
	public final static Event onGuestRoomDelete = new Event("RoomManager.OnGuestRoomDelete");
	
	private Map<String, RoomModel> models;
	private final ConcurrentHashMap<Integer, RoomInstance> instances;
	
	private final RoomWorker[] workers;

	private RoomManager(int numWorkers)
	{
		this.models = Collections.emptyMap();
		this.instances = new ConcurrentHashMap<Integer, RoomInstance>();
		
		this.workers = new RoomWorker[numWorkers];
		for(int index = 0; index < numWorkers; index++)
		{
			this.workers[index] = new RoomWorker(index, this);
			this.workers[index].start();
			Logger.getInstance().info("RoomMgr", String.format("Spawned RoomWorker #%d.", index));
		}
	}

	public void shutdown()
	{
		// Shutdown workers
		for(int index = 0; index < this.workers.length; index++)
		{
			this.workers[index].stop();
			this.workers[index] = null;
		}
		Logger.getInstance().info("RoomMgr", "Stopped all workers.");

		// Shutdown roominstances?
		if(!this.instances.isEmpty())
		{
			Logger.getInstance().info("RoomMgr", "Unloading all room instances...");
			for(RoomInstance instance : this.instances.values())
			{
				instance.unload();
			}
			this.instances.clear();
			Logger.getInstance().info("SessionMgr", "All room instances unloaded.");
		}
	}

	public void reloadModels()
	{
		// Create new collection
		Map<String, RoomModel> tmpModels = new HashMap<String, RoomModel>();
		
		// Load all model data
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT * FROM roommodels");
		while(result.row())
		{
			RoomModel model = new RoomModel();
			model.name = result.getString("name");
			model.doorX = result.getInt("door_x");
			model.doorY = result.getInt("door_y");
			model.doorZ = result.getFloat("door_z");
			model.doorRotation = result.getByte("door_rotation");
			model.hasSwimmingPool = result.getBoolean("has_pool");
			model.clubOnly = result.getBoolean("club_only");
			model.init(result.getString("heightmap"));
			tmpModels.put(model.name, model);
		}
		result.release();

		// Set instance field
		this.models = tmpModels;
		
		// Done!
		Logger.getInstance().info("RoomMgr", String.format("Reloaded %d room models.", this.models.size()));
	}

	public RoomModel getModel(String name)
	{
		return this.models.get(name);
	}

	public Collection<RoomModel> getAllRoomModels()
	{
		return this.models.values();
	}
	
	public RoomInstance getRoomInstance(int roomId, boolean allowCreate)
	{
		// Look for existing instance
		RoomInstance instance = this.instances.get(roomId);
		if(instance != null)
		{
			return instance;
		}

		// Allowed to create a new instance?
		if(allowCreate)
		{
			return this.createRoomInstance(roomId);
		}
		else
		{
			return null;
		}
	}

	private RoomWorker getLowestLoadedWorker()
	{
		int lowest = 0, bestIndex = -1;
		for(int index = 0; index < this.workers.length; index++)
		{
			int load = this.workers[index].load();
			if(bestIndex == -1 || load < lowest)
			{
				lowest = load;
				bestIndex = index;
			}
		}
		return this.workers[bestIndex];
	}

	private RoomInstance createRoomInstance(int roomId)
	{
		// Get room info
		Room info = this.getRoomInfo(roomId);
		if(info == null)
		{
			return null;
		}

		// Get room model
		RoomModel model = this.getModel(info.modelName);
		if(model == null)
		{
			Logger.getInstance().warning("RoomMgr", String.format("Room #%d refers to the missing room model \"%s\". Could not create room instance.", info.id, info.modelName));
			return null;
		}

		// Create instance and load data like furniture
		RoomInstance instance = new RoomInstance(info, model);
		try
		{
			instance.load();
		}
		catch(Exception ex)
		{
			Logger.getInstance().error("RoomMgr", String.format("Error loading instance of room #%d.", roomId), ex);
			return null;
		}

		// Assign to worker
		this.getLowestLoadedWorker().assign(instance);

		// Add to instance map
		this.instances.put(instance.getInfo().id, instance);

		// Instance created!
		Logger.getInstance().debug("RoomMgr", String.format("Created instance for room #%d.", info.id));
		return instance;
	}

	public void unloadRoomInstance(int roomId)
	{
		// Attempt to remove from the instance map
		RoomInstance instance = this.instances.remove(roomId);

		// Instance removed? Process further unloading
		if(instance != null)
		{
			// Un-assign from worker
			this.workers[instance.getWorkerIndex()].unassign(roomId);

			// Unload/save data
			instance.unload();

			// Unloaded successfully!
			Logger.getInstance().debug("RoomMgr", String.format("Unloaded instance of room #%d.", roomId));
		}
	}

	public Room getRoomInfo(int roomId)
	{
		// Lookup cache
		/*Room room = this.cache.get(roomId);
		if(room != null)
		{
			return room;
		}*/
		
		// Lookup room instances
		RoomInstance instance = this.getRoomInstance(roomId, false);
		if(instance != null)
		{
			return instance.getInfo();
		}

		// Lookup database
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r LEFT JOIN users u ON u.id=r.owner_id WHERE r.id=" + roomId);
		Room room = result.row() ? Room.parse(result) : null;
		result.release();
		return room;
	}
	
	public List<Room> getRoomsForUser(int userId)
	{
		ArrayList<Room> rooms = new ArrayList<Room>(5);

		DatabaseResult result = habbod.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id=" + userId);
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}

	public int createGuestRoom(int userId, String name, String modelName)
	{
		// Create the room
		int roomId = habbod.getInstance().getDb().executeReturnKey("INSERT INTO rooms(owner_id,name,model_name) VALUES (" + userId + ",'" + SQLHelper.escape(name) + "','" + SQLHelper.escape(modelName) + "')");
		if(roomId > 0)
		{
			// Create other stuff
			RoomManager.onGuestRoomCreate.fire(roomId);
		}
		return roomId;
	}

	public void deleteGuestRoom(int roomId)
	{
		// Delete room
		habbod.getInstance().getDb().execute("DELETE FROM rooms WHERE id=" + roomId);
		habbod.getInstance().getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId);
		
		// Delete content
		habbod.getInstance().getDb().execute("DELETE FROM items WHERE room_id=" + roomId);
		
		// Delete other stuff
		RoomManager.onGuestRoomCreate.fire(roomId);
		
		// Unload room instance (if loaded)
		this.unloadRoomInstance(roomId);
	}

	public Map<Integer, String> getRightsForRoom(int roomId)
	{
		// Initialize an empty map with init cap of 3
		Map<Integer, String> users = new HashMap<Integer, String>(3);

		// Query the user IDs + usernames
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT user_id,username FROM room_rights JOIN users ON id=user_id WHERE room_id=" + roomId);
		while(result.row())
		{
			users.put(result.getInt("user_id"), result.getString("username"));
		}
		result.release();

		// Return the users
		return users;
	}

	public void addRoomRights(int roomId, int userId)
	{
		habbod.getInstance().getDb().execute("INSERT INTO room_rights(room_id,user_id) VALUES(" + roomId + "," + userId + ")");
	}

	public void removeRoomRights(int roomId, int userId)
	{
		habbod.getInstance().getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId + " AND user_id=" + userId);
	}
	
	public void logRoomEntry(int userId, int roomId)
	{
		habbod.getInstance().getDb().execute("INSERT INTO user_roomvisits(user_id,room_id,entry) VALUES (" + userId + "," + roomId + ",'" + SQLHelper.getDateTimeStr() + "')");
	}
	
	public void logRoomExit(int userId, int roomId)
	{
		habbod.getInstance().getDb().execute("UPDATE user_roomvisits SET `exit` = '" + SQLHelper.getDateTimeStr() + "' WHERE user_id=" + userId + " AND room_id=" + roomId + " AND `exit` IS NULL LIMIT 1");
	}
	
	public List<Room> getRecentGuestRoomsForUser(int userId, int maxAmount)
	{
		List<Room> rooms = new ArrayList<Room>();

		List<Integer> exclude = new ArrayList<Integer>();
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT r.*,u.username AS owner,rv.entry FROM user_roomvisits rv JOIN rooms r ON r.id = rv.room_id JOIN users u ON u.id=r.owner_id WHERE rv.user_id=" + userId + " ORDER BY rv.entry DESC LIMIT " + maxAmount);
		while(result.row())
		{
			int roomId = result.getInt("id");
			if(!exclude.contains(roomId))
			{
				rooms.add(Room.parse(result));
				exclude.add(roomId);
			}
		}
		result.release();

		return rooms;
	}
	
	public List<RoomVisit> getLastRoomVisitsForUser(int userId, int maxAmount)
	{
		List<RoomVisit> visits = new ArrayList<RoomVisit>(maxAmount);
		
		DatabaseResult result =  habbod.getInstance().getDb().query("SELECT r.*,u.username AS owner,rv.entry,rv.exit FROM user_roomvisits rv JOIN rooms r ON r.id = rv.room_id LEFT JOIN users u ON u.id=r.owner_id WHERE rv.user_id=" + userId + " ORDER BY rv.entry DESC LIMIT " + maxAmount);
		while(result.row())
		{
			RoomVisit visit = new RoomVisit();
			visit.room = Room.parse(result);
			visit.entry = result.getTime("entry");
			visit.exit = result.getTime("exit");
			visits.add(visit);
		}
		result.release();
		
		return visits;
	}
	
	public List<Integer> getFavoriteRoomsForUser(int userId)
	{
		List<Integer> favorites = new ArrayList<Integer>(3);

		DatabaseResult result = habbod.getInstance().getDb().query("SELECT room_id FROM user_favoriterooms WHERE user_id=" + userId);
		while(result.row())
		{
			favorites.add(result.getInt("room_id"));
		}
		result.release();

		return favorites;
	}

	public void addFavoriteRoomForUser(int userId, int roomId)
	{
		habbod.getInstance().getDb().execute("INSERT INTO user_favoriterooms(user_id,room_id) VALUES (" + userId + "," + roomId + ")");
	}

	public void removeFavoriteRoomForUser(int userId, int roomId)
	{
		habbod.getInstance().getDb().execute("DELETE FROM user_favoriterooms WHERE user_id=" + userId + " AND room_id=" + roomId);
	}
	
	public Collection<RoomInstance> getAllRoomInstances()
	{
		return this.instances.values();
	}

	/* Begin singleton */
	private static RoomManager instance;

	public final static RoomManager getInstance()
	{
		return instance;
	}

	public static void createInstance(int numWorkers)
	{
		instance = new RoomManager(numWorkers);
	}
	/* End singleton */

}
