package net.nillus.habbo.sessions;

import java.util.ArrayList;
import java.util.List;

import net.nillus.habbo.communication.ClientMessage;
import net.nillus.habbo.communication.EServerMessage;
import net.nillus.habbo.communication.RequestHandler;
import net.nillus.habbo.communication.RequestRegistry;
import net.nillus.habbo.communication.ServerMessage;
import net.nillus.habbo.communication.messages.outgoing.handshake.AuthenticationOKMessageComposer;
import net.nillus.habbo.communication.messages.outgoing.handshake.UserRightsMessageComposer;
import net.nillus.habbo.communication.messages.outgoing.inventory.avatareffect.AvatarEffectsMessageComposer;
import net.nillus.habbo.communication.messages.outgoing.moderation.ModerationPresetsMessageComposer;
import net.nillus.habbo.communication.messages.outgoing.moderation.ModerationTicketMessageComposer;
import net.nillus.habbo.communication.messages.outgoing.navigator.FavoriteRoomsMessageComposer;
import net.nillus.habbo.core.DynamicObject;
import net.nillus.habbo.core.Logger;
import net.nillus.habbo.events.Event;
import net.nillus.habbo.game.items.ItemInventory;
import net.nillus.habbo.game.items.pets.PetInventory;
import net.nillus.habbo.game.items.trading.Trade;
import net.nillus.habbo.game.messenger.InstantMessenger;
import net.nillus.habbo.game.moderation.ModerationBan;
import net.nillus.habbo.game.moderation.ModerationCenter;
import net.nillus.habbo.game.moderation.ModerationPresetManager;
import net.nillus.habbo.game.moderation.ModerationTicket;
import net.nillus.habbo.game.rooms.Room;
import net.nillus.habbo.game.rooms.RoomManager;
import net.nillus.habbo.game.rooms.instances.RoomInstance;
import net.nillus.habbo.game.rooms.instances.RoomUser;
import net.nillus.habbo.game.rooms.navigator.FavoriteRoomList;
import net.nillus.habbo.game.rooms.navigator.RoomEvent;
import net.nillus.habbo.game.rooms.navigator.RoomEventManager;
import net.nillus.habbo.game.users.User;
import net.nillus.habbo.game.users.UserManager;
import net.nillus.habbo.game.users.achievements.AchievementManager;
import net.nillus.habbo.game.users.badges.BadgeInventory;
import net.nillus.habbo.game.users.effects.EffectInventory;
import net.nillus.habbo.net.NetConnectionHandler;

public final class Session extends DynamicObject implements NetConnectionHandler.CommunicationHandler
{
	// Events
	public final static Event onAuthenticate = new Event("Session.OnAuthenticate");
	public final static Event onLogin = new Event("Session.OnLogin");
	public final static Event onLogout = new Event("Session.OnLogout");
	public final static Event onRebuildRightsList = new Event("User.OnRebuildRightsList");
	public final static Event onRoomEnter = new Event("Session.OnRoomEnter");
	public final static Event onRoomExit = new Event("Session.OnRoomExit");
	public final static Event onChat = new Event("Session.OnChat");

	// Core
	public final long id;
	private NetConnectionHandler conn;

	// Message handling
	private long lastMsgTime;
	private ServerMessage msg;
	private ServerMessage responseObject;

	// Session variables
	private User user;
	private int accessLogId;
	private InstantMessenger messenger;
	private BadgeInventory badgeInventory;
	private ItemInventory itemInventory;
	private PetInventory petInventory;
	private EffectInventory effectInventory;
	private FavoriteRoomList favoriteRooms;
	
	// Room related variables
	private RoomInstance roomInstance;
	private RoomUser roomUser;
	private RoomEvent roomEvent;
	private Trade trade;
	
	public int waitingForDoorbell;
	public int authenticatedTeleporterId;

	public Session(long sessionId, NetConnectionHandler conn)
	{
		this.id = sessionId;
		this.conn = conn;
		this.msg = new ServerMessage();
		this.responseObject = new ServerMessage();
	}

	public void onEnd()
	{
		this.conn.close();
		if(this.authenticated())
		{
			this.onLogout();
			SessionManager.getInstance().removeAuthedSession(this);
		}
	}

	public void authenticate(String ticket)
	{
		// User exists for this auth ticket?
		User user = UserManager.getInstance().getUserByAuthTicket(ticket);
		if(user == null)
		{
			SessionManager.getInstance().endSession(this.id, "invalid auth ticket");
			return;
		}

		// Check for bans
		List<ModerationBan> bans = ModerationCenter.getInstance().getActiveBansForUserId(user.id);
		if(!bans.isEmpty())
		{
			// Send the first ban
			ModerationBan ban = bans.get(0);
			this.send(ModerationBan.createMsg(ban));
			
			// User is banned! :)
			SessionManager.getInstance().endSession(user.id, "user is banned");
			return;
		}
		
		// Run auth checks
		if(!Session.onAuthenticate.fire(this, user))
		{
			return;
		}

		// Insufficient permissions?
		if(user.role == null || !user.role.rights.contains("fuse_login"))
		{
			SessionManager.getInstance().endSession(this.id, String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
			return;
		}

		// Add to session list
		this.user = user;
		SessionManager.getInstance().addAuthedSession(this);

		// Send user right list
		this.rebuildAndSendUserRights();

		// Enable info feed
		this.msg.init(EServerMessage.InfoFeedEnable);
		this.msg.appendBool(true);
		this.msg.appendBool(false);
		this.send(this.msg);
		
		// Login OK!
		this.send(AuthenticationOKMessageComposer.compose());

		// Run other logic
		this.onLogin();
	}

	private void onLogin()
	{
		// Reload messenger
		this.messenger = new InstantMessenger(this);
		this.messenger.reload();
		
		// Reload badges
		this.badgeInventory = new BadgeInventory(this.user.id);
		this.badgeInventory.reload();
		
		// Reload item inventory
		this.itemInventory = new ItemInventory(this);
		this.itemInventory.reload();
		
		// Reload pet inventory
		this.petInventory = new PetInventory(this);
		this.petInventory.reload();
		
		// Reload tags
		this.user.tags = UserManager.getInstance().getTagsForUser(this.user.id);
		
		// Reload effects
		this.effectInventory = new EffectInventory(this);
		this.effectInventory.reload();
		
		// Reload favorite rooms
		this.favoriteRooms = new FavoriteRoomList(this);
		this.favoriteRooms.reload();
		
		// Reload achievements
		this.user.achievements = AchievementManager.getInstance().getAchievementsForUser(this.user.id);

		// Update last online time
		this.user.updateLastOnline(true);

		// Push favorite rooms (they aren't requested?)
		this.send(FavoriteRoomsMessageComposer.compose(this.favoriteRooms.getFavorites()));
		
		// Push effects (they aren't requested?)
		this.send(AvatarEffectsMessageComposer.compose(this.effectInventory.getEffects()));
		
		// User is moderator?
		if(this.user.rights.contains("fuse_mod"))
		{
			// Send preset moderation responses
			ModerationPresetManager presets = ModerationCenter.getInstance().getPresetManager();
			this.send(ModerationPresetsMessageComposer.compose(presets.getUserMessages(), presets.getRoomMessages()));
			
			// Send open tickets
			for(ModerationTicket ticket : ModerationCenter.getInstance().getAllTickets())
			{
				if(ticket.status == ModerationTicket.STATUS_OPEN || ticket.status == ModerationTicket.STATUS_PICKED)
				{
					this.send(ModerationTicketMessageComposer.compose(ticket));
				}
			}
		}
		
		// Execute other on-login tasks
		Session.onLogin.fire(this);
	}

	private void onLogout()
	{
		// Run hooks
		Session.onLogout.fire(this);

		// Stop trading, leave the room, etc etc
		this.leaveRoomSilent();
		
		// End the room event, host has gone offline!
		if(this.roomEvent != null)
		{
			RoomEventManager.getInstance().removeEvent(this.roomEvent);
		}

		// Go offline!
		this.messenger.offline();
		
		// Clear inventory to speed up garbagecollection
		this.itemInventory.clear();
		
		// Save badge inventory
		if(this.badgeInventory.requiresSave())
		{
			this.badgeInventory.save();
		}
		this.badgeInventory.clear();

		// Update last online time
		this.user.updateLastOnline(false);

		// Save misc things
		this.user.saveMisc();
	}

	public void tryGetRoomAccess(int roomId, String password)
	{
		// Leave any room instance the user might be in now
		this.leaveRoomSilent();

		// Get/create room instance
		RoomInstance room = RoomManager.getInstance().getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}
		
		// Connection to roomserver is open!
		this.msg.init(EServerMessage.OpenConnection);
		this.send(this.msg);
		
		// The room owner can ALWAYS enter the room, never has to check for user limit, bans, state, etc
		boolean skipAccessChecks = (this.getUser().id == room.getInfo().ownerId) || this.getUser().rights.contains("fuse_can_enter_any_room") || this.authenticatedTeleporterId > 0;
		if(!skipAccessChecks)
		{
			int errorCode = 0;
			
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().rights.contains("fuse_enter_full_rooms"))
				{
					errorCode = 1;
				}
			}

			// User is banned?
			else if(room.userIsBanned(this.getUser().id))
			{
				errorCode = 4;
			}
			
			// Any RoomEntryErrors?
			if(errorCode != 0)
			{
				this.msg.init(EServerMessage.RoomEntryError);
				this.msg.appendInt(errorCode);
				this.send(this.msg);
				this.kickFromRoom();
				return;
			}

			// Password room?
			if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equals(password))
			{
				this.sendGenericError(-100002);
				this.kickFromRoom();
				return;
			}

			// Doorbell room?
			else if(room.getInfo().state == Room.STATE_DOORBELL)
			{
				// Ring the doorbell, await response
				if(room.ringDoorbell(this))
				{
					// Cool, let's wait
					this.waitingForDoorbell = room.getInfo().id;

					// Waiting...
					this.msg.init(EServerMessage.Doorbell);
					this.msg.appendStr("");
					this.send(this.msg);
				}
				else
				{
					// NOBODY HEARD YOU, FOREVER ALONE
					this.msg.init(EServerMessage.FlatAccessDenied);
					this.send(this.msg);
					this.kickFromRoom();
				}
				return;
			}
		}
		
		// Guestroom accessible!
		if(room.getInfo().isGuestRoom())
		{
			this.msg.init(EServerMessage.FlatAccessible);
			this.send(this.msg);
		}
		
		// Send room stats tracker
		this.msg.init(EServerMessage.RoomStatsTracker);
		this.msg.append("/client/");
		if(room.getInfo().isPublicRoom())
		{
			this.msg.append("public/");
			this.msg.append(room.getInfo().modelName);
			this.msg.appendStr("/0");
		}
		else
		{
			this.msg.append("private/");
			this.msg.append(Integer.toString(room.getInfo().id));
			this.msg.appendStr("/id");
		}
		this.send(this.msg);

		// Forward to room instance
		this.goToRoom(roomId);
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomInstance instance = RoomManager.getInstance().getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}

		// Room is ready!
		this.msg.init(EServerMessage.RoomReady);
		this.msg.appendStr(instance.getModel().name);
		this.msg.appendInt(instance.getInfo().id);
		this.send(this.msg);
		
		// This is the current room instance now
		this.roomInstance = instance;
	}

	public boolean enterRoom()
	{
		this.roomUser = this.roomInstance.addSession(this);
		if(this.roomUser != null)
		{
			if(Session.onRoomEnter.fire(this))
			{
				RoomManager.getInstance().logRoomEntry(this.user.id, this.roomInstance.getInfo().id);
				
				this.messenger.updateStatus(false);
				
				return true;
			}
			else
			{
				// Silent kick (a hook specified that roomentry should not continue)
				this.kickFromRoom();
			}
		}
		else
		{
			// Hacky entry?
			this.kickFromRoom("Invalid room entry! Room probably was being unloaded while you were entering! Please try again.");
		}

		// Do not enter room
		return false;
	}

	private boolean leaveRoomSilent()
	{
		// Reset room ID
		this.waitingForDoorbell = 0;

		// Leave room instance
		if(this.roomInstance != null)
		{
			if(Session.onRoomExit.fire(this))
			{
				this.roomInstance.removeSession(this);
				this.roomInstance = null;
				this.roomUser = null;
				
				RoomManager.getInstance().logRoomExit(this.user.id, this.roomInstance.getInfo().id);
				
				// End trade
				if(this.trade != null)
				{
					trade.end(this);
				}
				this.trade = null;
				
				// Update messenger
				this.messenger.updateStatus(false);
			}
			else
			{
				// Did not leave room
				return false;
			}
		}

		// Left room!
		return true;
	}

	public void kickFromRoom()
	{
		this.kickFromRoom(null);
	}

	public void kickFromRoom(String text)
	{
		this.kickFromRoom(text, false);
	}
	
	public void kickFromRoom(String text, boolean isCaution)
	{
		// SHE GAVE ME THE PLEASE-FACE
		if(!this.leaveRoomSilent())
		{
			// Room exit cancelled
			return;
		}

		// I GAVE HER THE LEAVE-FACE
		this.msg.init(EServerMessage.CloseConnection);
		this.send(this.msg);

		// BODY LIKE A BOTTLE, PRETTY LITTLE MODEL MODEL MODEEL
		if(text != null)
		{
			this.sendNotif(text, isCaution);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Fucking heartbeat! ("D{")
		if(msg.type == 315) return;

		// Log that this session received a message!
		this.lastMsgTime = System.currentTimeMillis();
		
		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable purposes
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			SessionManager.getInstance().endSession(this.id, "MSG_EOF detected in incoming message", true);
			return;
		}

		// These are the only responding handlers for InitCrypto and SSOTicket
		if(!this.authenticated() && (msg.type != 206 && msg.type != 415))
		{
			Logger.getInstance().warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.type, msg.header()));
		}
		else
		{
			// Lookup message handler and execute handling logic
			RequestHandler handler = RequestRegistry.get().lookup(msg.type);
			if(handler != null)
			{
				Logger.getInstance().debug(this, "<-- (" + handler.getClass().getSimpleName().replace("Handler", "") + ") " + msg.toString());
				try
				{
					handler.handle(this, msg, this.responseObject);
				}
				catch(Exception ex)
				{
					Logger.getInstance().error(this, String.format("Error handling msg #%d (\"%s\").", msg.type, msg.header()), ex);
				}
				this.responseObject.reset();
			}
			else
			{
				Logger.getInstance().debug(this, "<-- " + msg.toString());
				Logger.getInstance().debug(this, String.format("SessionMsgRegistry does not have an entry for msg #%d. (\"%s\")", msg.type, msg.header()));
			}
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		SessionManager.getInstance().endSession(this.id, reason);
	}

	public void send(ServerMessage msg)
	{
		Logger.getInstance().debug(this, "--> " + msg.toString());
		this.conn.send(msg);
	}

	public void sendGenericError(int code)
	{
		this.msg.init(EServerMessage.GenericError);
		this.msg.appendInt(code);
		this.send(this.msg);
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null, false);
	}

	public void sendNotif(String text, boolean isCaution)
	{
		this.sendNotif(text, null, isCaution);
	}
	
	public void sendNotif(String text, String url, boolean isCaution)
	{
		this.msg.init(isCaution ? EServerMessage.Caution : EServerMessage.Notif);
		this.msg.appendStr(text);
		this.msg.appendStr(url);
		this.send(this.msg);
	}

	public void sendSystemNotif(String text)
	{
		this.msg.init(810);
		this.msg.appendBool(true);
		this.msg.appendStr(text);
		this.send(this.msg);
	}
	
	public void sendWhisperNotif(String text)
	{
		if(this.inRoom())
		{
			// Send to this user & receiver
			this.msg.init(EServerMessage.Whisper);
			this.msg.appendInt(this.getRoomUser().avatarId);
			this.msg.appendStr(text);
			this.msg.appendBool(false);
			this.send(this.msg);
		}
	}

	public void rebuildAndSendUserRights()
	{
		if(this.user != null)
		{
			// Default right list
			user.rights = new ArrayList<String>();
			user.rights.add("default");
			user.specialRightsLevel = 2;

			// Role rights?
			if(user.role != null)
			{
				user.rights.addAll(user.role.rights);
			}
			
			// User-specific rights?
			Session.onRebuildRightsList.fire(this);

			// Send the rights list
			this.send(UserRightsMessageComposer.compose(user));
		}
	}

	public void sendCreditBalance()
	{
		this.msg.init(EServerMessage.CreditBalance);
		this.msg.appendStr(this.user.credits + ".0");
		this.send(this.msg);
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		this.msg.init(EServerMessage.ActivityPointNotification);
		this.msg.appendInt(this.user.pixels);
		this.msg.appendInt(added);
		this.send(this.msg);
	}

	public void sendRoomPrivileges()
	{
		if(this.roomUser != null)
		{
			if(this.roomUser.hasRoomRights)
			{
				this.msg.init(EServerMessage.YouAreController);
				this.send(this.msg);
			}
			if(this.roomUser.hasRoomOwner)
			{
				this.msg.init(EServerMessage.YouAreOwner);
				this.send(this.msg);
			}
		}
	}

	public final NetConnectionHandler getConnection()
	{
		return this.conn;
	}

	public final long getLastMsgTime()
	{
		return this.lastMsgTime;
	}

	public final boolean authenticated()
	{
		return(this.user != null);
	}

	public final User getUser()
	{
		return this.user;
	}

	public final int getAccessLogId()
	{
		return this.accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.accessLogId == 0)
		{
			this.accessLogId = accessLogId;
		}
	}

	public InstantMessenger getMessenger()
	{
		return this.messenger;
	}
	
	public BadgeInventory getBadgeInventory()
	{
		return this.badgeInventory;
	}
	
	public final ItemInventory getItemInventory()
	{
		return this.itemInventory;
	}
	
	public PetInventory getPetInventory()
	{
		return this.petInventory;
	}
	
	public EffectInventory getEffectInventory()
	{
		return this.effectInventory;
	}
	
	public FavoriteRoomList getFavoriteRooms()
	{
		return this.favoriteRooms;
	}

	public final boolean inRoom()
	{
		return(this.roomInstance != null && this.roomUser != null);
	}

	public final RoomInstance getRoomInstance()
	{
		return this.roomInstance;
	}

	public final RoomUser getRoomUser()
	{
		return this.roomUser;
	}
	
	public final RoomEvent getRoomEvent()
	{
		return this.roomEvent;
	}
	
	public final void setRoomEvent(RoomEvent event)
	{
		this.roomEvent = event;
	}
	
	public boolean isTrading()
	{
		return (this.trade != null);
	}
	
	public Trade getTrade()
	{
		return this.trade;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
