package habbod2.storage;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import habbod2.System;

public class DatabaseConnectionPool extends System.Component implements Runnable
{
	private Database database;
	private List<DatabaseConnection> connections;
	private long connectionCounter;
	
	private Thread monitor;
	private boolean monitorAlive;
	
	public DatabaseConnectionPool(Database database)
	{
		super(database.getSystem());
		this.database = database;
		this.connections = new ArrayList<DatabaseConnection>(database.getEndpoint().maxConnections);
	}
	
	public DatabaseConnection getConnection()
	{
		synchronized(this.connections)
		{
			// Any available connections?
			Iterator<DatabaseConnection> iterator = this.connections.iterator();
			while(iterator.hasNext())
			{
				DatabaseConnection conn = iterator.next();
				if(conn.isAvailable())
				{
					try
					{
						// If it's not closed, then this is the connection to return!
						if(!conn.getInnerConnection().isClosed())
						{
							conn.available(false);
							return conn;
						}
						else
						{
							iterator.remove();
							this.getLogger().warning(this, String.format("Connection #%d was dead but still in the available connections list, removed.", conn.id));
						}
					}
					catch(SQLException ex)
					{
						// /care
					}
				}
			}

			// No open available connections. Allowed to create a new connection?
			DatabaseEndpoint ep = this.database.getEndpoint();
			if(ep.maxConnections <= 0 || this.connections.size() < ep.maxConnections)
			{
				try
				{
					// Create & wrap JDBC connection
					Connection innerConn = DriverManager.getConnection(ep.url, ep.user, ep.password);
					DatabaseConnection conn = new DatabaseConnection(this.connectionCounter++, innerConn, this);
					
					// Add it to the available connections
					conn.available(true);
					this.connections.add(conn);
					this.getLogger().debug(this, String.format("Opened connection to %s. (poolsize=%d)", ep.url, this.connections.size()));

					// Try again
					return this.getConnection();
				}
				catch(Exception ex)
				{
					this.getLogger().error(this, ex.getMessage());
					return null;
				}
			}

			// Can't create more connections at the moment? Wait for a returning connection, but no longer than x seconds
			try
			{
				final int maxWaitSeconds = 2;
				this.connections.wait(maxWaitSeconds * 1000);
			}
			catch(InterruptedException ex)
			{
				// Waitlock triggered!
				return this.getConnection();
			}

			// This is real bad, please have mercy as there's nothing we can do right now, database is dead
			this.getLogger().warning(this, "No available connections, not allowed to open a new connection and waiting for returning connections timed out.");
			return null;
		}
	}

	public void releaseConnection(DatabaseConnection conn)
	{
		if(conn != null)
		{
			// Update the connection and notify one arbitrary thread that is waiting
			synchronized(this.connections)
			{
				conn.available(true);
				this.connections.notify();
			}
		}
	}
	
	public int size()
	{
		return this.connections.size();
	}
	
	public void startMonitor()
	{
		if(!this.monitorAlive)
		{
			this.monitor = new Thread(this, this.toString());
			this.monitor.setPriority(Thread.MIN_PRIORITY);
			this.monitorAlive = true;
			this.monitor.start();
		}
	}
	
	public void stopMonitor()
	{
		if(this.monitorAlive)
		{
			this.monitorAlive = false;
			this.monitor.interrupt();
		}
	}

	@Override
	public void run()
	{
		// Configuration
		final int maxIdleSeconds = 60;
		final int cleanupInterval = 30;
		
		// Run for aslong as the monitor is alive
		while(this.monitorAlive)
		{
			// Sleep before cleaning up...
			try
			{
				Thread.sleep(cleanupInterval * 1000);
			}
			catch(InterruptedException ex)
			{
				break;
			}
			
			// Close + remove available connections that have been idle for too long
			int count = 0;
			synchronized(this.connections)
			{
				Iterator<DatabaseConnection> iterator = this.connections.iterator();
				while(iterator.hasNext())
				{
					DatabaseConnection conn = iterator.next();
					if(conn.isAvailable())
					{
						if(conn.getIdleSeconds() > maxIdleSeconds)
						{
							//this.close(conn);
							iterator.remove();
							count++;
						}
					}
				}
			}
			
			// Anything closed?
			if(count > 0)
			{
				this.getLogger().debug(this, String.format("Closed %d idle connection(s)!", count));
			}
		}
	}
}
