package habbod2.storage;

import habbod2.System;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

public class DatabaseUpdateQueue extends System.Component implements Runnable
{
	private Database database;
	private DatabaseConnection connection;
	
	private List<String> queue;
	private Statement statement;
	private long lastEnqueueTime;
	
	private boolean alive;
	private Thread thread;

	public DatabaseUpdateQueue(Database database)
	{
		super(database.getSystem());
		this.database = database;
		this.queue = new ArrayList<String>();
	}

	public boolean open()
	{
		// Not already open?
		if(this.isOpen())
		{
			return false;
		}

		// Get database connection
		DatabaseConnection conn = this.database.getPool().getConnection();
		if(conn == null)
		{
			this.getLogger().warning(this, "Could not open because no database connection was available.");
			return false;
		}

		// Attempt creating the statement that will be used to process queues
		try
		{
			this.statement = conn.getInnerConnection().createStatement();
			this.connection = conn;
		}
		catch(SQLException ex)
		{
			this.database.getPool().releaseConnection(conn);
			this.getLogger().error(this, "Error creating JDBC statement!", ex);
			return false;
		}

		// Create + start thread
		this.thread = new Thread(this, this.toString());
		this.alive = true;
		this.thread.start();

		// Open for business
		this.getLogger().info(this, "Successfully enabled.");
		return true;
	}

	public void close()
	{
		// Actually open?
		if(!this.isOpen())
		{
			return;
		}

		// Abort the thread
		this.alive = false;
		this.thread.interrupt();

		// Wait...
		this.getLogger().debug(this, "Waiting for queue to become idle...");
		while(this.getIdleTime() < 1000)
		{
			try
			{
				Thread.sleep(100);
			}
			catch(InterruptedException e)
			{
				
			}
		}
		this.process();
		
		// Close the statement
		try
		{
			this.statement.close();
		}
		catch(SQLException ex)
		{
			this.getLogger().error("DatabaseQueryQueue", "Error closing statement.", ex);
		}

		// Release the connection
		this.database.getPool().releaseConnection(this.connection);
	}

	public void enqueue(String sql)
	{
		synchronized(this.queue)
		{
			this.queue.add(sql);
			this.lastEnqueueTime = System.currentTimeMillis();
		}
	}

	private void process()
	{
		synchronized(this.queue)
		{
			if(this.queue.size() > 0)
			{
				for(String entry : this.queue)
				{
					try
					{
						this.statement.executeUpdate(entry);
					}
					catch(SQLException ex)
					{
						this.getLogger().error(this, String.format("Error executing \"%s\": %s", entry, ex.getMessage()));
					}
				}
				this.queue.clear();
			}
		}
	}

	@Override
	public void run()
	{
		while(this.alive)
		{
			// Process queue
			this.process();

			// Process the queue every second
			try
			{
				Thread.sleep(1000);
			}
			catch(InterruptedException e)
			{
				return;
			}
		}
	}

	public boolean isOpen()
	{
		return(this.statement != null);
	}
	
	public long getIdleTime()
	{
		return System.currentTimeMillis() - this.lastEnqueueTime;
	}
}
