package habbod2.game.catalog;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static habbod2.core.Locale._I;
import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;
import habbod2.game.GameComponent;
import habbod2.game.GameManager;
import habbod2.game.items.InventoryItem;
import habbod2.game.items.ItemBase;
import habbod2.game.items.ItemDefinition;
import habbod2.game.items.ItemManager;
import habbod2.game.items.pets.PetUtil;
import habbod2.game.users.User;
import habbod2.game.util.GameUtil;
import habbod2.game.rooms.items.floor.*;
import habbod2.game.rooms.items.wall.*;
import habbod2.game.rooms.items.misc.*;
import habbod2.game.sessions.Session;

public class Catalog extends GameComponent
{
	private Map<Integer, CatalogPage> pages;
	private Map<Integer, CatalogProduct> products;
	private List<ItemDefinition> presentBoxes;
	
	public Catalog(GameManager gameMgr)
	{
		super(gameMgr);
		this.pages = Collections.emptyMap();
		this.products = Collections.emptyMap();
		this.presentBoxes = Collections.emptyList();
	}

	public void reload()
	{
		// Create new collections
		Map<Integer, CatalogPage> tmpPages = new LinkedHashMap<Integer, CatalogPage>();
		Map<Integer, CatalogProduct> tmpProducts = new HashMap<Integer, CatalogProduct>();

		// Load all the pages
		DatabaseResult result = this.getDb().query("SELECT * FROM catalog_pages ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogPage page = new CatalogPage();
			page.id = result.getInt("id");
			page.parentId = result.getInt("parent_id");
			page.name = result.getString("name");
			page.requiredRight = result.getString("required_right");
			if(page.requiredRight != null && page.requiredRight.isEmpty())
			{
				page.requiredRight = null;
			}
			page.isEnabled = result.getBoolean("enabled");
			page.isCategory = result.getBoolean("is_category");
			page.iconColor = result.getInt("icon_color");
			page.iconImage = result.getInt("icon_image");
			page.layout = result.getString("layout");
			page.headlineImage = result.getString("headline_image");
			page.teaserImage = result.getString("teaser_image");
			page.specialImage = result.getString("special_image");
			page.text1 = result.getString("text1");
			page.text2 = result.getString("text2");
			page.details = result.getString("details");
			page.teaserText = result.getString("teaser_text");
			tmpPages.put(page.id, page);
		}

		// Link the pages in parent/child fashion
		for(CatalogPage page : tmpPages.values())
		{
			if(page.parentId > 0)
			{
				CatalogPage parent = tmpPages.get(page.parentId);
				if(parent == null)
				{
					this.getLogger().warning(this, String.format("Page #%d is a child of page #%d, which is missing. Page not attached to parent.", page.id, page.parentId));
				}
				else
				{
					parent.pages.add(page);
				}
			}
		}

		// Get item manager
		ItemManager itemMgr = this.getGameMgr().getItemMgr();
		
		// Load the products, grab the item definitions and attach them to the pages
		result.query("SELECT * FROM catalog_products ORDER BY ordernum ASC");
		while(result.row())
		{
			CatalogProduct product = new CatalogProduct();
			product.id = result.getInt("id");
			product.pageId = result.getInt("page_id");
			product.code = result.getString("code");
			product.itemDefinitionId = result.getInt("definition_id");
			product.itemAmount = result.getInt("amount");
			product.costCredits = result.getInt("cost_credits");
			product.costPixels = result.getInt("cost_pixels");
			product.costQuestPoints = result.getInt("cost_questpoints");
			product.item = itemMgr.getDefinition(product.itemDefinitionId);
			if(product.item == null)
			{
				// That's fine, but give a warning if there was supposed to be one!
				if(product.itemDefinitionId > 0)
				{
					this.getLogger().warning(this, String.format("Product #%d ('%s') refers to the missing item definition #%d. Product ignored.", product.id, product.code, product.itemDefinitionId));
				}
			}
			else
			{
				if(product.pageId > 0)
				{
					CatalogPage page = tmpPages.get(product.pageId);
					if(page == null)
					{
						this.getLogger().warning(this, String.format("Product #%d ('%s') is supposed to appear on page #%d, which is missing. Product not attached to page.", product.id, product.code, product.pageId));
					}
					else
					{
						product.page = page;
						page.products.add(product);
					}
				}
				tmpProducts.put(product.id, product);
			}
		}
		result.release();
		
		// Collect the presentboxes
		List<ItemDefinition> tmpBoxes = this.getGameMgr().getItemMgr().getDefinitionsByRoomClass(PresentBoxItem.class);

		// Set instance fields
		this.pages = Collections.unmodifiableMap(tmpPages);
		this.products = Collections.unmodifiableMap(tmpProducts);
		this.presentBoxes = Collections.unmodifiableList(tmpBoxes);

		// Done!
		this.getLogger().info(this, String.format("Reloaded %d pages and %d products.", this.pages.size(), this.products.size()));
	}

	public CatalogPage getPage(int pageId)
	{
		return this.pages.get(pageId);
	}

	public List<CatalogPage> getVisibleIndexPagesForUser(User user)
	{
		List<CatalogPage> visible = new ArrayList<CatalogPage>();
		for(CatalogPage page : this.pages.values())
		{
			if(page.parentId == -1 && page.isVisibleTo(user))
			{
				visible.add(page);
			}
		}
		return visible;
	}

	public Collection<CatalogPage> getAllPages()
	{
		return this.pages.values();
	}

	public CatalogProduct getProduct(int productId)
	{
		return this.products.get(productId);
	}

	public Collection<CatalogProduct> getAllProducts()
	{
		return this.products.values();
	}

	public void billUser(Session session, CatalogProduct product)
	{
		if(product.isPurchasableBy(session.getUser()))
		{
			if(product.costCredits > 0)
			{
				session.getUser().credits -= product.costCredits;
				session.sendCreditBalance();
			}
			if(product.costPixels > 0)
			{
				session.getUser().pixels -= product.costPixels;
				session.sendPixelBalance();
			}
			session.getUser().saveCurrencies();
		}
	}
	
	public List<InventoryItem> purchaseProduct(CatalogProduct product, User user, String purchaseData, boolean processPurchaseData)
	{
		// Valid args?
		if(product == null || user == null || product.item == null)
		{
			return null;
		}

		// Process purchase data
		if(processPurchaseData)
		{
			try
			{
				purchaseData = this.processDataToState(product, user, purchaseData);
			}
			catch(Exception ex)
			{
				purchaseData = ItemBase.ERROR_DATA;
			}
			if(purchaseData != null && purchaseData.equals(ItemBase.ERROR_DATA))
			{
				this.getLogger().warning(this, String.format("Purchase of product #%d by user #%d failed, bad purchasedata.", product.id, user.id));
				return null;
			}
		}
		
		// Create the shipping items
		List<InventoryItem> shipping = new ArrayList<InventoryItem>(1);
		
		// Effects
		if(product.is(EffectItem.class))
		{
			Session session = this.getGameMgr().getSessionMgr().getSessionByUserId(user.id);
			if(session != null)
			{
				int effectId = GameUtil.parseInt(product.item.data);
				int duration = this.getSystem().getProperties().getInt("game.effects.default.duration");
				session.getEffectInventory().addEffect(effectId, duration);
			}
			return shipping;
		}
		
		// Pets
		else if(product.is(PetItem.class))
		{
			// Parse pet data
			String[] parts = purchaseData.split("\n", 3);
			String name = parts[0];
			int breed = Integer.parseInt(parts[1]);
			String color = parts[2];
			int type = PetUtil.productNameToPetType(product.item.sprite);
			
			// Create the pet for this user, will appear in inventory
			this.getGameMgr().getPetMgr().createPet(user.id, name, type, breed, color);
			
			// TODO: add free food item to shipping (sample!)
			return shipping;
		}
		
		// Start creating the items
		for(int x = 0; x < product.itemAmount; x++)
		{
			// Create the items
			InventoryItem item = new InventoryItem();
			item.definition = product.item;
			item.state = purchaseData;
			if(item.create(user.id))
			{
				shipping.add(item);
				
				// Set post.it default color
				if(item.is(PostItItem.class))
				{
					item.data = "FFFF33";
				}
				
				// Teleporters get a second, linking item
				else if(item.is(TeleporterItem.class))
				{
					// Create the second item
					InventoryItem item2 = new InventoryItem();
					item2.definition = item.definition;
					item2.data = Integer.toString(item.id);
					if(item2.create(user.id))
					{
						shipping.add(item2);
						
						// Update the first item with a reference to the second item
						item.data = Integer.toString(item2.id);
						item.update(user.id);
					}
				}
			}
		}    

		// These are the items that are shipped
		return shipping;
	}
	

	public boolean purchaseAsGift(CatalogProduct product, User user, String purchaseData, int receiverId, String giftNote)
	{
		// Run pre-purchase logic, filter & apply purchase data
		try
		{
			purchaseData = this.processDataToState(product, user, purchaseData);
		}
		catch(Exception ex)
		{
			purchaseData = ItemBase.ERROR_DATA;
		}
		if(purchaseData != null && purchaseData.equals(ItemBase.ERROR_DATA))
		{
			this.getLogger().warning(this, String.format("Gift-purchase of product #%d by user #%d failed, bad purchase data.", product.id, user.id));
			return false;
		}
		
		// Any definitions?
		if(this.presentBoxes.isEmpty())
		{
			this.getLogger().warning(this, "No item definitions with class PresentBoxItem available! Gifts cannot be purchased.");
			return false;
		}
		
		// Create presentbox, state=giftnote, data=purchasedata
		InventoryItem item = GameUtil.getRandomElement(this.presentBoxes).newItem();
		item.state = "!" + giftNote;
		item.data = (product.id + GameUtil.DATA_DELIMITER + purchaseData);
		this.getGameMgr().getItemMgr().storeItem(item, receiverId);

		// Is the receiver online?
		Session receiver = this.getGameMgr().getSessionMgr().getSessionByUserId(receiverId);
		if(receiver != null)
		{
			// Add the item to the inventory component
			receiver.getItemInventory().addItem(item, false);
			receiver.getItemInventory().refresh();

			// Send out a notification
			receiver.sendNotif(_I("You have received a gift!"));
		}

		// Log it?
		if(user != null)
		{
			this.logGift(user.id, receiverId, product, giftNote);
		}
		
		// Purchase OK!
		return true;
	}
	
	private String processDataToState(CatalogProduct product, User user, String purchaseData)
	{
		// Decorations
		if(product.is(DecorationItem.class))
		{
			if(product.item.data.equals("landscape"))
			{
				return Float.toString(Float.parseFloat(purchaseData));
			}
			else
			{
				return Integer.toString(Integer.parseInt(purchaseData));
			}
		}
		
		// Trophies
		else if(product.is(TrophyItem.class))
		{
			return(user.username + (char) 9 + GameUtil.getCurrentDate() + (char) 9 + (purchaseData));
		}
		
		// Pets
		else if(product.is(PetItem.class))
		{
			return purchaseData;
		}
		
		// Nothing special, discard state
		return null;
	}
	
	private void logGift(int userId, int receiverId, CatalogProduct product, String giftNote)
	{
		this.getDb().execute("INSERT INTO user_giftlog(user_id,receiver_id,`timestamp`,product_id,note) VALUES (" + userId + "," + receiverId + ",'" + SQLHelper.getDateTime() + "'," + product.id + ",'" + SQLHelper.escape(giftNote) + "')");
	}
}
