package habbod2.game.rooms.instance;

import habbod2.communication.messages.EServerMessage;
import habbod2.game.rooms.geometry.Direction;
import habbod2.game.rooms.pathfinding.JoehPathfinderStaticImpl;
import habbod2.game.rooms.pathfinding.PathfinderNode;
import habbod2.net.protocol.ServerMessage;

import java.util.ArrayList;
import java.util.List;

public class AvatarEngine
{
	private RoomInstance room;
	
	public AvatarEngine(RoomInstance room)
	{
		this.room = room;
	}
	
	public void pulse()
	{
		// Get door
		RoomTile door = this.room.getMap().getDoor();
		
		// Process all the avatars
		List<RoomAvatar> updates = new ArrayList<RoomAvatar>();
		for(RoomAvatar avatar : this.room.getAllAvatars())
		{
			// Perform other tasks
			avatar.pulse();
			
			// Requires status update?
			if(avatar.requiresUpdate)
			{
				// Update me!
				if(updates.contains(avatar))
				{
					continue;
				}
				else
				{
					updates.add(avatar);
					avatar.requiresUpdate = false;
				}
				
				// Moving?
				if(avatar.hasGoal())
				{
					// At goal?
					if(avatar.at(avatar.goal))
					{
						// Trip complete!
						avatar.clearMovement();
						
						// In door?
						if(avatar.at(door) && avatar instanceof RoomUser)
						{
							((RoomUser)avatar).getSession().kickFromRoom();
						}
					}
					else
					{
						// Determine next tile in path
						RoomTile next = null;
						if(avatar.overrideNextTile && avatar.goal != null)
						{
							// Override used!
							avatar.overrideNextTile = false;

							// Substitute the next step
							next = avatar.goal;
						}
						else
						{
							// Start new path?
							if(avatar.executePathfinder)
							{
								// Shorthand
								RoomTile goal = avatar.goal;
								
								// Redirect?
								if(goal.redirect != null)
								{
									goal = goal.redirect;
									avatar.goal = goal;
								}

								// Open last step tile for pathfinder if the goal is a last step tile
								byte backup = -1;
								if(goal.is(RoomTile.LASTSTEP))
								{
									// No avatar here? (quick check)
									if((!goal.hasAvatar() || goal.allowWalkthrough))
									{
										backup = goal.state;
										goal.state = RoomTile.OPEN;
									}
								}

								// Find path using the beloved Joe 'Joeh' Hegarty 2D A* pathfinder
								boolean noDescLimit = this.room.getInfo().isGuestRoom();
								List<PathfinderNode> path = JoehPathfinderStaticImpl.findPath(avatar.x, avatar.y, goal.x, goal.y, this.room.getMap(), 1.5f, noDescLimit ? 100.0f : 2.0f, avatar.pathfindingAllowedOn);

								// Restore map
								if(backup != -1)
								{
									goal.state = backup;
								}
								
								// Executed pathfinder!
								avatar.executePathfinder = false;

								// Is there a path found?
								if(!path.isEmpty())
								{
									// Remove current tile node
									path.remove(0);

									// Set this path!
									avatar.path = path;
								}
							}

							// Pop the next tile from the path
							if(!avatar.path.isEmpty())
							{
								// Test the next tile
								PathfinderNode tmpNode = avatar.path.remove(0);
								next = this.room.getMap().getTile(tmpNode.x, tmpNode.y);
								if(!next.isPassable(avatar.pathfindingAllowedOn, true))
								{
									next = null;
								}
								else
								{
									// Walking diagonal (hacky wacky isometrics!)
									final boolean ALLOW_DIAGONAL = true;
									if(ALLOW_DIAGONAL && !avatar.path.isEmpty())
									{
										// Peek + test next node (already popped one, so this is like path[1])
										tmpNode = avatar.path.get(0);
										RoomTile peek = this.room.getMap().getTile(tmpNode.x, tmpNode.y);
										if(peek.isPassable(avatar.pathfindingAllowedOn, false))
										{
											// Can't skip more than one tile!
											if(avatar.touches(peek))
											{
												byte xAdd1 = 0;
												byte xAdd2 = 0;
												if(peek.x > avatar.x)
												{
													xAdd1 = -1;
													xAdd2 = 1;
												}
												else
												{
													xAdd1 = 1;
													xAdd2 = -1;
												}

												// Valid shortcut?
												RoomTile left = this.room.getMap().getTile(peek.x + xAdd1, peek.y);
												RoomTile right = this.room.getMap().getTile(avatar.x + xAdd2, avatar.y);
												if(left != null && right != null && left.isPassable(avatar.pathfindingAllowedOn, false) && right.isPassable(avatar.pathfindingAllowedOn, false))
												{
													next = peek;
													avatar.path.remove(0);
												}
											}
										}
									}
								}
							}
						}
						
						// Next tile available?
						if(next == null)
						{
							avatar.clearMovement();
						}
						else
						{
							// Redirect?
							if(next.redirect != null)
							{
								next = next.redirect;
							}
							
							// Set next tile
							avatar.next = next;
							
							// Mounted?
							if(avatar instanceof RoomHuman)
							{
								RoomHuman human = (RoomHuman)avatar;
								if(human.isMounted())
								{
									human.getMount().next = next;
									human.getMount().requiresUpdate = true;
									if(!updates.contains(human.getMount()))
									{
										updates.add(human.getMount());
									}
								}
							}
						}
					}
				}
			}
		}
		
		// Any updates? This includes simple status updates, but also movement:phase2
		if(!updates.isEmpty())
		{
			ServerMessage msg = ServerMessage.create(EServerMessage.UserUpdate);
			msg.appendInt(updates.size());
			for(RoomAvatar avatar : updates)
			{
				if(avatar.next != null)
				{
					avatar.setDirection(Direction.ofMove(avatar.x, avatar.y, avatar.next.x, avatar.next.y));
					avatar.removeNonMoveStatuses();
					avatar.statuses.put("mv", avatar.next.x + "," + avatar.next.y + "," + (avatar.next.height + avatar.getExtraDisplayHeight()));
				}
				avatar.appendStatusString(msg);
				if(avatar.next != null)
				{
					avatar.statuses.remove("mv");
					avatar.setTile(avatar.next, false);
					avatar.next = null;
				}
			}
			this.room.broadcast(msg);
		}
	}
}
