package habbod2.game.rooms.instance;

import habbod2.communication.messages.outgoing.room.action.*;
import habbod2.game.pets.RoomPet;
import habbod2.game.rooms.geometry.Direction;

public abstract class RoomHuman extends RoomAvatar
{
	public static final int CARRYITEM_PULSES = 240; // 120 seconds
	
	public static final int EFFECT_USER = 0;
	public static final int EFFECT_ROOM = 1;
	public static final int EFFECT_GAME = 2;
	public static final int EFFECT_MISC = 3;
	
	public boolean isAsleep;
	public int danceId;
	
	public int carryItemId;
	public int carryItemTimer;
	
	private int[] effectId;
	
	private int warpPhase;
	private RoomTile warpDestination;
	private Direction warpDirection;
	
	private RoomPet mount;
	
	public RoomHuman()
	{
		this.effectId = new int[4];
	}
	
	public void wave()
	{
		this.unidle();
		this.getRoom().broadcast(WaveMessageComposer.compose(this));
	}
	
	public void dance(int danceId)
	{
		this.unidle();
		this.stopCarryItem();
		this.danceId = danceId;
		this.getRoom().broadcast(DanceMessageComposer.compose(this));
	}
	
	public boolean isCarryingItem()
	{
		return (this.carryItemId > 0 && this.carryItemTimer > 0);
	}
	
	public void carryItem(int itemId)
	{
		this.carryItemId = itemId;
		if(itemId > 0)
		{
			this.carryItemTimer = CARRYITEM_PULSES;
		}
		else
		{
			this.carryItemTimer = 0;
		}
		this.getRoom().broadcast(CarryObjectMessageComposer.compose(this));
	}
	
	public void stopCarryItem()
	{
		this.carryItem(0);
	}
	
	public int getEffect()
	{
		// Overrides
		if(this.isMounted())
		{
			return 103;
		}
		
		// Could be a for-loop, but for clarification purposes...
		if(this.effectId[EFFECT_MISC] > 0)
		{
			return this.effectId[EFFECT_MISC];
		}
		else if(this.effectId[EFFECT_GAME] > 0)
		{
			return this.effectId[EFFECT_GAME];
		}
		else if(this.effectId[EFFECT_ROOM] > 0)
		{
			return this.effectId[EFFECT_ROOM];
		}
		else
		{
			return this.effectId[EFFECT_USER];
		}
	}
	
	public void setEffect(int type, int effectId)
	{
		boolean isChange = (effectId != this.effectId[type]);
		if(isChange)
		{
			this.effectId[type] = effectId;
			this.refreshEffect();
		}
	}
	
	public void clearEffect(int type)
	{
		this.setEffect(type, 0);
	}
	
	public void refreshEffect()
	{
		this.getRoom().broadcast(AvatarEffectMessageComposer.compose(this));
	}
	
	public void setAsleep(boolean isAsleep)
	{
		if(this.isAsleep != isAsleep)
		{
			this.isAsleep = isAsleep;
			this.getRoom().broadcast(SleepMessageComposer.compose(this));
		}
	}
	
	@Override
	public void lookTo(RoomTile tile, boolean headOnly)
	{
		// Handle mount
		if(this.isMounted())
		{
			if(headOnly)
			{
				return;
			}
			else
			{
				this.getMount().lookTo(tile, headOnly);
			}
		}
		super.lookTo(tile, headOnly);
	}
	
	public void warp(RoomTile dest)
	{
		this.warp(dest, this.bodyDirection);
	}
	
	public void warp(RoomTile dest, Direction direction)
	{
		if(dest != null)
		{
			this.lockMovement();
			this.warpPhase = 4;
			this.warpDestination = dest;
			this.warpDirection = direction;
		}
	}
	
	public boolean isWarping()
	{
		return this.warpPhase > 0;
	}
	
	public void pulse()
	{
		// RoomAvatar pulse()
		super.pulse();
		
		// Deal with carry item
		if(this.carryItemId > 0)
		{
			this.carryItemTimer--;
			if(this.carryItemTimer <= 0)
			{
				this.stopCarryItem();
			}
		}
		
		// Warping (with sparkles-effect)
		if(this.warpPhase > 0)
		{
			if(this.warpPhase == 4)
			{
				this.setEffect(EFFECT_MISC, 4);
			}
			else if(this.warpPhase == 3)
			{
				this.headDirection = this.bodyDirection = this.warpDirection;
				this.setTile(this.warpDestination, false);
				if(this.isMounted())
				{
					this.mount.setTile(this.warpDestination, false);
					this.mount.headDirection = this.bodyDirection = this.warpDirection;
				}
				this.unlockMovement();
				this.warpDestination = null;
			}
			else if(this.warpPhase == 2)
			{
				this.clearEffect(EFFECT_MISC);
			}
			this.warpPhase--;
		}
	}
	
	@Override
	public void unidle()
	{
		super.unidle();
		this.setAsleep(false);
	}
	
	public RoomPet getMount()
	{
		return this.mount;
	}
	
	public void setMount(RoomPet pet)
	{
		// Mount or dismount?
		if(pet == null)
		{
			if(this.mount != null)
			{
				this.mount.setRider(null);
				this.mount = null;
				this.refresh(true);
				this.refreshEffect();
			}
		}
		else
		{
			this.mount = pet;
			this.mount.setRider(this);
			this.refreshEffect();
			this.bodyDirection = this.headDirection = this.mount.bodyDirection;
			this.setTile(this.mount.getTile(), true);
		}
	}
	
	public boolean isMounted()
	{
		return (this.mount != null);
	}
	
	@Override
	public float getExtraDisplayHeight()
	{
		if(this.isMounted())
		{
			return this.getMount().getInfo().type.mountHeight;
		}
		else
		{
			return 0;
		}
	}
	
	@Override
	public void invalidate()
	{
		this.setMount(null);
	}
}
