package habbod2.game.rooms.instance;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;

import habbod2.Instance;
import habbod2.game.GameManager;
import habbod2.game.pets.*;
import habbod2.game.rooms.*;
import habbod2.game.bots.Bot;
import habbod2.game.bots.RoomBot;
import habbod2.game.items.EItemFlags;
import habbod2.game.items.InventoryItem;
import habbod2.game.items.ItemDefinition;
import habbod2.game.rooms.minigames.Game;
import habbod2.game.rooms.navigator.RoomEvent;
import habbod2.game.sessions.Session;
import habbod2.game.util.SecurityUtil;
import habbod2.net.protocol.ServerMessage;
import habbod2.storage.DatabaseResult;
import habbod2.game.rooms.items.*;
import habbod2.game.rooms.items.floor.*;
import habbod2.game.rooms.items.minigame.ball.BallItem;
import habbod2.game.rooms.items.wall.*;
import habbod2.game.rooms.items.wired.*;
import habbod2.game.rooms.items.wired.triggers.TimerResetTrigger;

import habbod2.communication.messages.outgoing.navigator.*;
import habbod2.communication.messages.outgoing.room.engine.*;

public class RoomInstance implements Runnable
{
	public static final int PULSE_INTERVAL = 500;
	public static final int MAX_IDLE_SECONDS = 15;

	// Core room info
	private Room info;
	private RoomModel model;
	private RoomMap map;
	private RoomEvent event;

	// Instance info
	private ScheduledFuture<?> task;
	private int idleTimer;
	private boolean isUnloading;
	private boolean unloadRequested;

	// Avatars
	private int avatarIdCounter;
	private AvatarEngine avatarEngine;
	private ConcurrentHashMap<String, RoomAvatar> avatars;
	private ConcurrentHashMap<Integer, RoomUser> users;
		
	// Content
	private boolean isLoaded;
	private ConcurrentHashMap<Integer, Long> bans;
	private ConcurrentHashMap<Integer, String> rights;
	private ConcurrentHashMap<Integer, RoomItem> items;
	private ConcurrentHashMap<Integer, RoomWallItem> wallItems;
	private ConcurrentHashMap<Integer, RoomFloorItem> floorItems;
	private List<RoomFloorItem> pendingStateRefresh;
	private RollerEngine rollers;
	private RoomDimmerItem roomDimmer;
	private JukeboxItem jukebox;
	private List<WiredTrigger> triggers;
	private int timerResetTimer;
	private Game minigame;

	public RoomInstance(Room info, RoomModel model)
	{
		// Core room info
		this.info = info;
		this.model = model;
		this.map = model.createMap();
		
		// Avatars
		this.avatarIdCounter = 0;
		this.avatarEngine = new AvatarEngine(this);
		this.avatars = new ConcurrentHashMap<String, RoomAvatar>();
		this.users = new ConcurrentHashMap<Integer, RoomUser>();
		
		// Content
		this.bans = new ConcurrentHashMap<Integer, Long>(0);
		this.rights = new ConcurrentHashMap<Integer, String>(0);
		this.items = new ConcurrentHashMap<Integer, RoomItem>();
		this.wallItems = new ConcurrentHashMap<Integer, RoomWallItem>();
		this.floorItems = new ConcurrentHashMap<Integer, RoomFloorItem>();
		this.pendingStateRefresh = new ArrayList<RoomFloorItem>(1);
		this.triggers = new ArrayList<WiredTrigger>(0);
	}

	public void load() throws Exception
	{
		// Get gamemanager
		GameManager gameMgr = Instance.get().getGameMgr();
		
		// Reload bans & rights
		this.bans.clear();
		this.rights.clear();
		this.rights.putAll(gameMgr.getRoomMgr().getRightsForRoom(this.info.id));
		
		// Clear items
		this.items.clear();
		this.wallItems.clear();
		this.floorItems.clear();
		
		// Load all items for this room
		try(DatabaseResult result = gameMgr.getDb().query("SELECT * FROM items WHERE room_id=" + this.info.id))
		{
			while(result.row())
			{
				// Valid definition?
				ItemDefinition definition = gameMgr.getItemMgr().getDefinition(result.getInt("definition_id"));
				if(definition == null)
				{
					continue;
				}
				
				// Valid room item?
				RoomItem item = definition.createRoomItem();
				if(item == null)
				{
					continue;
				}
				
				// Process from row, according to type
				item.id = result.getInt("id");
				item.state = result.getString("state");
				item.data = result.getString("data");
				if(item instanceof RoomWallItem)
				{
					RoomWallItem wallItem = (RoomWallItem)item;
					wallItem.position = result.getString("wallposition");
					this.wallItems.put(wallItem.id, wallItem);
				}
				else
				{
					RoomFloorItem floorItem = (RoomFloorItem)item;
					floorItem.x = result.getInt("x");
					floorItem.y = result.getInt("y");
					floorItem.z = result.getFloat("z");
					floorItem.rotation = result.getByte("rotation");
					this.floorItems.put(floorItem.id, floorItem);
				}
				
				// Add to all items
				this.items.put(item.id, item);
				item.setRoom(this);
			}
		}
		
		// Secondary pass! (need all items to be loaded)
		for(RoomFloorItem item : this.floorItems.values())
		{
			if(item instanceof WiredItem)
			{
				((WiredItem)item).reloadSettings();
			}
		}

		// Generate map
		this.rebuildFloorMap();
		
		// Loading up a guestroom?
		if(this.info.isGuestRoom())
		{
			// Load/sync event
			RoomEvent event = gameMgr.getRoomEventMgr().getEventForRoom(this.info.id);
			if(event != null)
			{
				this.event = event;
				event.room = this.getInfo();
			}
			
			// Load bots
			List<Bot> bots = gameMgr.getBotMgr().getBotsForRoom(this.info.id);
			for(Bot bot : bots)
			{
				RoomBot avatar = new RoomBot(bot);
				RoomTile tile = this.map.getTile(bot.x, bot.y);
				this.spawnAvatar(avatar, tile);
			}
			
			// Load pets
			List<Pet> pets = gameMgr.getPetMgr().getPetsForRoom(this.info.id);
			for(Pet pet : pets)
			{
				RoomPet avatar = new RoomPet(pet);
				RoomTile tile = this.map.getTile(pet.x, pet.y);
				this.spawnAvatar(avatar, tile);
			}
		}
		
		// Content loaded!
		this.isLoaded = true;
	}

	public void unload()
	{
		// Already unloading?
		if(this.isUnloading)
		{
			return;
		}

		// Flag that this room instance is unloading
		this.isUnloading = true;

		// Sync pets with database
		for(RoomAvatar avatar : this.avatars.values())
		{
			if(avatar instanceof RoomPet)
			{
				((RoomPet)avatar).syncAndSaveLocationInfo();
			}
		}

		// Remove all users
		for(RoomUser user : this.users.values())
		{
			user.getSession().kickFromRoom();
		}
		this.users.clear();
		this.avatars.clear();
		this.rights.clear();
		this.bans.clear();
		if(this.info.usersNow > 0)
		{
			this.info.usersNow = 0;
			this.info.saveUserCount();
		}

		// Save items
		for(RoomItem item : this.items.values())
		{
			if(item.requiresDbUpdate())
			{
				item.update();
			}
		}
		this.items.clear();
		this.wallItems.clear();
		this.floorItems.clear();
	}

	public void spawnAvatar(RoomAvatar avatar, RoomTile tile)
	{
		// Tile exists?
		if(tile == null)
		{
			return;
		}
		
		// Allocate avatar ID
		avatar.avatarId = this.avatarIdCounter++;
		avatar.setRoom(this);
		this.avatars.put(avatar.getAvatarKey(), avatar);

		// Set tile
		avatar.setTile(tile, false);
		
		// Broadcast entry
		this.broadcast(UsersMessageComposer.compose(avatar));

		// Step the tile
		avatar.getTile().stepOn(avatar);
	}

	public void removeAvatar(RoomAvatar avatar)
	{
		// Leave the tile
		avatar.getTile().stepOff(avatar);
		
		// Remove and invalidate for any references floating around
		avatar.invalidate();
		this.avatars.remove(avatar.getAvatarKey());

		// Broadcast removal
		this.broadcast(UserRemoveMessageComposer.compose(avatar));
	}

	public RoomAvatar getAvatarByAvatarKey(String avatarKey)
	{
		return this.avatars.get(avatarKey);
	}

	public RoomAvatar getAvatarByAvatarId(int avatarId)
	{
		for(RoomAvatar avatar : this.avatars.values())
		{
			if(avatar.avatarId == avatarId)
			{
				return avatar;
			}
		}
		return null;
	}

	public RoomUser getUserByUserId(int userId)
	{
		return this.users.get(userId);
	}

	public RoomUser getUserByUsername(String username)
	{
		for(RoomUser user : this.users.values())
		{
			if(user.getInfo().username.equalsIgnoreCase(username))
			{
				return user;
			}
		}
		return null;
	}
	
	public RoomUser addSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.isUnloading)
		{
			return null;
		}
		
		// Already in room?
		if(this.users.containsKey(session.getUser().id))
		{
			return null;
		}

		// Create roomuser with unique virtual ID
		RoomUser user = new RoomUser(session);

		// Position user at spawn position
		RoomTile spawn;
		if(session.authenticatedTeleporterId == 0)
		{
			// Spawn in door
			spawn = this.map.getDoor();
			user.headDirection = user.bodyDirection = this.model.doorDirection;
		}
		else
		{
			// Finalize cross-room teleportation
			RoomItem item = this.getItem(session.authenticatedTeleporterId);
			if(item == null || !(item instanceof TeleporterItem))
			{
				// Transfer corrupted!
				session.authenticatedTeleporterId = 0;
				return null;
			}
			else
			{
				TeleporterItem tele = (TeleporterItem)item;
				tele.arriveUser(user);
				spawn = tele.getTile();
			}
		}

		// Determine privileges if in guestroom
		if(this.info.isGuestRoom())
		{
			boolean isAnyRoomOwner = user.getInfo().hasRight("fuse_is_any_room_owner");
			user.hasRoomOwner = (isAnyRoomOwner || (user.getInfo().id == this.info.ownerId));
			user.hasRoomRights = (user.hasRoomOwner || this.rights.containsKey(user.getInfo().id));
			if(user.hasRoomRights || this.info.allKicking)
			{
				user.addStatus("flatctrl", user.hasRoomOwner ? "useradmin" : null);
			}
		}

		// Spawn avatar
		this.users.put(user.getInfo().id, user);
		this.spawnAvatar(user, spawn);

		// Update current user count
		synchronized(this.info)
		{
			this.info.usersNow = this.users.size();
			this.info.saveUserCount();
		}

		// Done!
		return user;
	}

	public void removeSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(this.isUnloading)
		{
			return;
		}

		// Remove user from collection
		RoomUser user = this.users.remove(session.getUser().id);

		// Actually user removed?
		if(user != null)
		{
			// Quit game session
			user.setPlayer(null);
			
			// Remove avatar
			this.removeAvatar(user);

			// Update current user count
			synchronized(this.info)
			{
				this.info.usersNow = this.users.size();
				this.info.saveUserCount();
			}
		}
	}

	public boolean ringDoorbell(Session ringing)
	{
		boolean heard = false;
		ServerMessage msg = null;
		for(RoomUser user : this.users.values())
		{
			if(user.hasRoomRights)
			{
				if(msg == null)
				{
					msg = DoorbellMessageComposer.compose(ringing.getUser());
				}
				user.getSession().send(msg);
				heard = true;
			}
		}

		// True if atleast one user heard the doorbell and is allowed to answer it
		return heard;
	}

	public void broadcast(ServerMessage msg)
	{
		if(!this.users.isEmpty())
		{
			//Service.get().getLogger().debug(this, "--> " + msg.toString());
			for(RoomUser user : this.users.values())
			{
				user.getSession().getConnection().send(msg);
			}
		}
	}

	public void addUserBan(int userId, int seconds)
	{
		long expireTime = Instance.get().getTime() + (seconds * 1000);
		this.bans.put(userId, expireTime);
	}

	public void removeUserBan(int userId)
	{
		this.bans.remove(userId);
	}

	public boolean userIsBanned(int userId)
	{
		Long expireTime = this.bans.get(userId);
		if(expireTime != null)
		{
			if(Instance.get().getTime() < expireTime)
			{
				return true;
			}
			else
			{
				this.removeUserBan(userId);
			}
		}
		return false;
	}

	public boolean addUserRights(int userId, String username)
	{
		if(this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.put(userId, username);
			Instance.get().getGameMgr().getRoomMgr().addRoomRights(this.info.id, userId);
			
			return true;
		}
	}

	public boolean removeUserRights(int userId)
	{
		if(!this.rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			this.rights.remove(userId);
			Instance.get().getGameMgr().getRoomMgr().removeRoomRights(this.info.id, userId);
			
			return true;
		}
	}

	public boolean userHasRights(int userId)
	{
		return this.rights.containsKey(userId);
	}

	public void rebuildFloorMap()
	{
		// Reset map
		this.map.reset(this.info.allowWalkthrough);

		// Apply floor items to map
		for(RoomFloorItem item : this.floorItems.values())
		{
			this.map.applyFloorItem(item);
		}
		
		// Restore door tile
		RoomTile door = this.map.getDoor();
		door.state = RoomTile.OPEN;
		door.height = this.map.getModel().doorZ;
	}
	
	public RoomItem getItem(int itemId)
	{
		return this.items.get(itemId);
	}
	
	public boolean placeItem(InventoryItem item, int x, int y, byte rotation)
	{
		return this.moveFloorItem(item.toFloorItem(), x, y, rotation, true, true);
	}

	public boolean moveItem(int itemId, int x, int y, byte rotation, boolean doRefresh)
	{
		return this.moveFloorItem(this.floorItems.get(itemId), x, y, rotation, false, doRefresh);
	}

	private boolean moveFloorItem(RoomFloorItem item, int newX, int newY, byte newRotation, boolean isNew, boolean doRefresh)
	{
		// Item given?
		if(item == null)
		{
			return false;
		}
		
		// If new -> should not be in room, if not new -> should be in room
		boolean inRoom = this.items.containsKey(item.id);
		if((isNew && inRoom) || (!isNew && !inRoom))
		{
			return false;
		}

		// Validate new rotation, this can really trouble client and server!
		if(newRotation != 0 && newRotation != 2 && newRotation != 4 && newRotation != 6)
		{
			return false;
		}
		
		// Get the tiles this object would reside on if this item was placed here, check them for 'freeness'
		List<RoomTile> newTiles = this.map.getAffectedTiles(newX, newY, item.definition.length, item.definition.width, newRotation);
		for(RoomTile tile : newTiles)
		{
			// Bad tile?
			if(tile == RoomTile.OUT_OF_RANGE || this.map.getModel().getMap()[tile.x][tile.y] == 'x')
			{
				return false;
			}

			// Is there an avatar on this tile?
			if(tile.hasAvatar())
			{
				// Conflicting rotation?
				if(newRotation == item.rotation && (tile.x != item.x || tile.y != item.y))
				{
					return false;
				}
			}
		}

		// Initial item height is specified by root tile
		float newZ = this.getMap().getTile(newX, newY).height;
		
		// Trying to stack on itself? (by rotating an item that is already in the space)
		if(!isNew)
		{
			if(item.x == newX && item.y == newY && item.z != newZ)
			{
				newZ = item.z;
			}
		}
		
		// Check top items under the item being moved for stacking
		for(RoomTile tile : newTiles)
		{
			// Get top item
			RoomFloorItem top = tile.topItem;
			if(top != null && top != item)
			{
				// Trying to stack an unstackable item?
				if(item.is(EItemFlags.ALLOW_1X1_STACK_ONLY))
				{
					return false;
				}
				
				// Prohibit stacking other than 1x1?
				if(top.is(EItemFlags.ALLOW_1X1_STACK_ONLY))
				{
					// This item is not 1x1?
					if(!(item.definition.length == 1 && item.definition.width == 1))
					{
						return false;
					}
				}
				
				// Can stack on top?
				if(top.is(EItemFlags.CAN_STACK_ON_TOP))
				{
					newZ = top.getTopHeight();
				}
				else
				{
					return false;
				}
			}
		}

		// Impose stacking limit
		int maxStackHeight = Instance.get().getConfig().getInt("game.rooms.stacking.maxheight");
		if(maxStackHeight > 0 && newZ > maxStackHeight)
		{
			newZ = maxStackHeight;
		}

		// Backup the old position
		int oldX = item.x;
		int oldY = item.y;
		byte oldRotation = item.rotation;

		// Set object on new position
		item.x = newX;
		item.y = newY;
		item.z = newZ;
		item.rotation = newRotation;

		// If item is new to room...
		if(isNew)
		{
			// Associate the item with this room
			item.setRoom(this);
			this.items.put(item.id, item);
			this.floorItems.put(item.id, item);

			// Important change, update straight away
			item.update();
		}
		else
		{
			// Just a db update on instance unload will do!
			item.requireDbUpdate();
		}

		// Broadcast placement
		if(doRefresh)
		{
			if(isNew)
			{
				this.broadcast(ObjectAddMessageComposer.compose(item));
			}
			else
			{
				this.broadcast(ObjectUpdateMessageComposer.compose(item));
			}
		}
		
		// Get old tiles
		List<RoomTile> oldTiles = this.map.getAffectedTiles(oldX, oldY, item.definition.length, item.definition.width, oldRotation);
		for(RoomTile tile : oldTiles)
		{
			if(tile.topItem == item && tile.hasAvatar())
			{
				for(RoomAvatar avatar : this.avatars.values())
				{
					if(avatar.at(tile))
					{
						item.onAvatarStepOff(avatar);
					}
				}
			}
		}
		
		// Generate the new map
		this.rebuildFloorMap();

		// Reposition avatars on previous tiles?
		if(!isNew && !(item instanceof BallItem))
		{
			for(RoomTile tile : oldTiles)
			{
				this.repositionAvatarsOnTile(tile);
			}
		}

		// Reposition avatars on new tiles
		for(RoomTile tile : this.map.getAffectedTiles(newX, newY, item.definition.length, item.definition.width, newRotation))
		{
			this.repositionAvatarsOnTile(tile);
		}

		// Move OK!
		return true;
	}

	public boolean placeItem(InventoryItem item, String position)
	{
		return this.moveWallItem(item.toWallItem(), position, true);
	}
	
	public boolean moveItem(int itemId, String newPosition)
	{
		return this.moveWallItem(this.wallItems.get(itemId), newPosition, false);
	}
	
	private boolean moveWallItem(RoomWallItem item, String newPosition, boolean isNew)
	{
		// Item given?
		if(item == null)
		{
			return false;
		}
		
		// If new -> should not be in room, if not new -> should be in room
		boolean inRoom = this.items.containsKey(item.id);
		if((isNew && inRoom) || (!isNew && !inRoom))
		{
			return false;
		}
		
		// Valid position?
		newPosition = SecurityUtil.validateWallPosition(newPosition);
		if(newPosition == null)
		{
			return false;
		}

		// Update the item with new info
		item.setRoom(this);
		item.position = newPosition;
		if(isNew)
		{
			item.update();
		}
		else
		{
			item.requireDbUpdate();
		}

		// Add to the room instance
		this.items.put(item.id, item);
		this.wallItems.put(item.id, item);
		
		// Broadcast placement
		if(isNew)
		{
			this.broadcast(ItemAddMessageComposer.compose(item));
		}
		else
		{
			this.broadcast(ItemUpdateMessageComposer.compose(item));
		}

		// Move OK!
		return true;
	}
	
	public InventoryItem pickupItem(int itemId)
	{
		RoomItem item = this.getItem(itemId);
		if(item == null)
		{
			return null;
		}
		else
		{
			// Remove from room
			this.items.remove(itemId);
			if(item instanceof RoomWallItem)
			{
				this.wallItems.remove(itemId);
				this.broadcast(ItemRemoveMessageComposer.compose((RoomWallItem)item));
			}
			else if(item instanceof RoomFloorItem)
			{
				RoomFloorItem floorItem = (RoomFloorItem)item;
				this.floorItems.remove(itemId);
				this.broadcast(ObjectRemoveMessageComposer.compose(floorItem));
				
				// Step off!
				List<RoomTile> tiles = floorItem.getTiles();
				for(RoomTile tile : tiles)
				{
					if(tile.topItem == floorItem && tile.hasAvatar())
					{
						for(RoomAvatar avatar : this.avatars.values())
						{
							if(avatar.at(tile))
							{
								floorItem.onAvatarStepOff(avatar);
							}
						}
					}
				}
				
				// Rebuild map and reposition avatars
				this.rebuildFloorMap();
				for(RoomTile tile : tiles)
				{
					this.repositionAvatarsOnTile(tile);
				}
			}
			item.clearRoom();
			
			// Return as inventory item
			return item.toInventoryItem();
		}
	}
	
	public void repositionAvatarsOnTile(RoomTile tile)
	{
		if(tile.hasAvatar())
		{
			for(RoomAvatar avatar : this.avatars.values())
			{
				if(avatar.at(tile))
				{
					avatar.reposition();
				}
			}
		}
	}
	
	public void refreshItemState(RoomFloorItem item)
	{
		if(!this.pendingStateRefresh.contains(item))
		{
			this.pendingStateRefresh.add(item);
		}
	}
	
	@Override
	public void run()
	{
		try
		{
			this.pulse();
			if(this.unloadRequested || this.idleTimer > ((RoomInstance.MAX_IDLE_SECONDS * 1000) / RoomInstance.PULSE_INTERVAL))
			{
				Instance.get().getGameMgr().getRoomMgr().unloadRoomInstance(this.info.id);
			}
		}
		catch(Exception ex)
		{
			Instance.get().getLogger().error(this, "Pulse error!", ex);
		}
	}
	
	private void pulse()
	{
		// Pulse the items, process timed items, etc
		this.pulseItems();
		
		// Pulse the avatars
		this.avatarEngine.pulse();
		
		// Pulse the rollers?
		if(this.rollers != null)
		{
			this.rollers.pulse();
		}
		
		// Pulse minigame?
		if(this.minigame != null)
		{
			this.minigame.pulse();
		}
		
		// Another pulse since the wired timer reset!
		this.timerResetTimer++;
		
		// Any items need refreshing?
		if(!this.pendingStateRefresh.isEmpty())
		{
			this.broadcast(ObjectsDataUpdateMessageComposer.compose(this.pendingStateRefresh));
			this.pendingStateRefresh.clear();
		}
		
		// Update idle timer
		if(this.users.size() > 0)
		{
			this.idleTimer = 0;
		}
		else
		{
			this.idleTimer++;
		}
	}

	private void pulseItems()
	{
		for(RoomItem item : this.items.values())
		{
			if(item.requiresPulse())
			{
				try
				{
					item.pulse();
				}
				catch(Exception ex)
				{
					ex.printStackTrace();
				}
			}
		}
	}
	
	public void addTrigger(WiredTrigger trigger)
	{
		this.triggers.add(trigger);
	}
	
	public void removeTrigger(WiredTrigger trigger)
	{
		this.triggers.remove(trigger);
	}
	
	public void executeTriggers(RoomUser user, Class<? extends WiredTrigger> type)
	{
		for(WiredTrigger trigger : this.triggers)
		{
			if(trigger.getClass() == type)
			{
				trigger.execute(user);
			}
		}
	}
	
	public Room getInfo()
	{
		return this.info;
	}
	
	public ScheduledFuture<?> getTask()
	{
		return this.task;
	}
	
	public void setTask(ScheduledFuture<?> task)
	{
		this.task = task;
	}

	public void requestUnload()
	{
		this.unloadRequested = true;
	}
	
	public RoomModel getModel()
	{
		return this.model;
	}

	public RoomMap getMap()
	{
		return this.map;
	}
	
	public Collection<RoomAvatar> getAllAvatars()
	{
		return this.avatars.values();
	}

	public Collection<RoomUser> getAllUsers()
	{
		return this.users.values();
	}

	public Map<Integer, String> getUsersWithRights()
	{
		return this.rights;
	}
	
	public boolean isContentLoaded()
	{
		return this.isLoaded;
	}
	
	public boolean isUnloading()
	{
		return this.isUnloading;
	}
	
	public Collection<RoomItem> getItems()
	{
		return this.items.values();
	}
	
	public Collection<RoomWallItem> getWallItems()
	{
		return this.wallItems.values();
	}
	
	public Collection<RoomFloorItem> getFloorItems()
	{
		return this.floorItems.values();
	}
	
	public RoomEvent getEvent()
	{
		return this.event;
	}
	
	public void setEvent(RoomEvent event)
	{
		this.event = event;
		this.broadcast(RoomEventMessageComposer.compose(event));
	}

	public RollerEngine getRollers()
	{
		return this.rollers;
	}
	
	public void enableRollers()
	{
		this.rollers = new RollerEngine(this);
	}
	
	public RoomDimmerItem getRoomDimmer()
	{
		return this.roomDimmer;
	}
	
	public void setRoomDimmer(RoomDimmerItem roomDimmer)
	{
		this.roomDimmer = roomDimmer;
	}
	
	public JukeboxItem getJukebox()
	{
		return this.jukebox;
	}
	
	public void setJukebox(JukeboxItem jukebox)
	{
		this.jukebox = jukebox;
	}

	public void resetTimer()
	{
		this.timerResetTimer = 0;
		for(WiredTrigger trigger : this.triggers)
		{
			if(trigger instanceof TimerResetTrigger)
			{
				((TimerResetTrigger) trigger).timerReset();
			}
		}
	}
	
	public int getTimerResetTimer()
	{
		return this.timerResetTimer;
	}
	
	public Game getMinigame()
	{
		return this.minigame;
	}
	
	public void setMinigame(Game minigame)
	{
		this.minigame = minigame;
	}
	
	@Override
	public String toString()
	{
		return "RoomInstance #" + this.info.id;
	}
}