package habbod2.game.rooms.instances;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import habbod2.game.rooms.RoomManager;

public class RoomWorker implements Runnable
{
	private int index;
	private RoomManager roomMgr;

	private CopyOnWriteArrayList<Integer> toRemove;
	private ConcurrentHashMap<Integer, RoomInstance> instances;

	private Thread thread;
	private boolean running;

	public RoomWorker(int index, RoomManager roomMgr)
	{
		this.index = index;
		this.roomMgr = roomMgr;
		this.toRemove = new CopyOnWriteArrayList<Integer>();
		this.instances = new ConcurrentHashMap<Integer, RoomInstance>();
	}

	public void start()
	{
		if(!this.running)
		{
			this.running = true;
			this.thread = new Thread(this, this.toString());
			this.thread.start();
		}
	}

	public void stop()
	{
		this.running = false;
		this.thread.interrupt();
		this.instances.clear();
	}

	public void assign(RoomInstance instance)
	{
		instance.setWorkerIndex(this.index);
		this.instances.put(instance.getInfo().id, instance);
	}

	public void unassign(int roomId)
	{
		this.toRemove.add(roomId);
	}

	@Override
	public void run()
	{
		while(this.running)
		{
			// Log start time
			long startTime = System.nanoTime();

			// Pulse all the room instances
			for(RoomInstance instance : this.instances.values())
			{
				try
				{
					instance.pulse();
					if(instance.requiresUnload())
					{
						this.roomMgr.unloadRoomInstance(instance.getInfo().id);
					}
				}
				catch(Exception ex)
				{
					this.roomMgr.getLogger().error(this, String.format("Error pulsing room instance #%d.", instance.getInfo().id), ex);
				}
			}

			// Any roominstances to remove?
			if(!this.toRemove.isEmpty())
			{
				for(Integer roomId : this.toRemove)
				{
					this.instances.remove(roomId);
				}
				this.toRemove.clear();
			}

			// Log the stop time
			long endTime = System.nanoTime();

			// Now sleep the essential amount of ms for a smooth pulse interval
			double sleepTime = (endTime - startTime);
			sleepTime /= 1000000;
			sleepTime = (RoomInstance.PULSE_INTERVAL - sleepTime);
			if(sleepTime > 0.0)
			{
				try
				{
					Thread.sleep((long) sleepTime);
				}
				catch(InterruptedException ex)
				{

				}
			}
		}
	}

	public int index()
	{
		return this.index;
	}

	public RoomManager getMgr()
	{
		return this.roomMgr;
	}

	public int load()
	{
		return this.instances.size();
	}

	public boolean running()
	{
		return this.running;
	}

	@Override
	public String toString()
	{
		return "RoomWorker #" + this.index;
	}
}
