package habbod2.game.rooms.minigames;

import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.ScoreboardItem;
import habbod2.game.rooms.items.minigame.TimerItem;
import habbod2.game.rooms.items.wired.triggers.GameEndsTrigger;
import habbod2.game.rooms.items.wired.triggers.GameStartsTrigger;
import habbod2.game.rooms.minigames.GameTeam;
import habbod2.game.util.GameUtil;

public abstract class Game
{
	private static final int MIN_TEAMS_WITH_PLAYERS = 2;
	
	private int id;
	private RoomInstance room;
	private GameTeam[] teams;
	private boolean isRunning;
	private int startDelay;
	private int timeLeft;
	private boolean decreaseTime;
	
	public Game(RoomInstance room)
	{
		// Set room
		this.room = room;
		
		// Get all team IDs
		GameTeam.ETeam[] teamIds = GameTeam.ETeam.values();
		
		// Create all empty teams
		this.teams = new GameTeam[teamIds.length];
		for(int i = 0; i < teamIds.length; i++)
		{
			this.teams[i] = new GameTeam(teamIds[i], this);
		}
	}
	
	public void destroy()
	{
		this.stop();
		for(GameTeam team : this.teams)
		{
			for(GameUserPlayer player : team.getPlayers())
			{
				player.getUser().setPlayer(null);
				player.getUser().clearEffect(RoomUser.EFFECT_GAME);
			}
		}
	}
	
	public void start()
	{
		// Not already running?
		if(this.isRunning)
		{
			return;
		}
		
		// Requires timer
		TimerItem timer = this.getTimer();
		if(timer == null || timer.getTimeLeft() == 0)
		{
			return;
		}
		
		// Requires at least x teams with players
		int teamsWithPlayers = 0;
		for(GameTeam team : this.teams)
		{
			if(team.getPlayers().size() > 0)
			{
				teamsWithPlayers++;
				continue;
			}
		}
		if(teamsWithPlayers < MIN_TEAMS_WITH_PLAYERS)
		{
			return;
		}
		
		// Prepare for start of new game
		this.id = GameUtil.getRandomInt(1, Integer.MAX_VALUE);
		this.timeLeft = timer.getTimeLeft();
		this.onStart();
		
		// Process items
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof ScoreboardItem)
			{
				((ScoreboardItem)item).setScore(0);
			}
			else if(item instanceof GameStartsTrigger)
			{
				((GameStartsTrigger)item).execute(null);
			}
			else
			{
				this.onStartProcessItem(item);
			}
		}
		
		// Go!
		this.isRunning = true;
	}
	
	protected abstract void onStart();
	
	protected abstract void onStartProcessItem(RoomFloorItem item);
	
	protected void onStartDelayExpired() { }
	
	public void stop()
	{
		// Actually running?
		if(!this.isRunning)
		{
			return;
		}
		
		// Stopped!
		this.isRunning = false;
		this.timeLeft = 0;
		
		// Determine winning team
		GameTeam winner = null;
		for(GameTeam team : this.teams)
		{
			if(team.getScore() > 0 && (winner == null || team.getScore() > winner.getScore()))
			{
				winner = team;
			}
		}
		
		// Is there a winning team?
		if(winner != null)
		{
			// Waving is for winners!
			for(GameUserPlayer player : winner.getPlayers())
			{
				player.getUser().wave();
			}
		}
		
		// Extra events
		this.onStop(winner);
		
		// Process items
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof TimerItem)
			{
				((TimerItem)item).setTimeLeft(0);
			}
			else if(item instanceof GameEndsTrigger)
			{
				((GameEndsTrigger)item).execute(null);
			}
			else
			{
				this.onStopProcessItem(item, winner);
			}
		}
		
		// Reset scores etc
		for(GameTeam team : this.teams)
		{
			team.reset();
		}
	}
	
	protected abstract void onStop(GameTeam winner);
	
	protected abstract void onStopProcessItem(RoomFloorItem item, GameTeam winner);
	
	public void pulse()
	{
		// Actually running?
		if(!this.isRunning)
		{
			return;
		}
		
		// Decrease time every odd pulse
		this.decreaseTime = !this.decreaseTime;
		if(this.decreaseTime)
		{
			// Deal with start delay?
			if(this.startDelay > 0)
			{
				this.startDelay--;
				if(this.startDelay == 0)
				{
					this.onStartDelayExpired();
				}
				return;
			}
			
			// Decrement time left?
			this.setTimeLeft(this.timeLeft - 1);
			if(this.timeLeft <= 0)
			{
				this.stop();
			}
		}
		
		// Process other game logic
		if(this.startDelay == 0)
		{
			this.onPulse();
		}
	}

	protected abstract void onPulse();
	
	public int getId()
	{
		return this.id;
	}

	public RoomInstance getRoom()
	{
		return this.room;
	}
	
	public TimerItem getTimer()
	{
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof TimerItem)
			{
				return (TimerItem)item;
			}
		}
		
		return null;
	}
	
	public GameTeam getTeam(GameTeam.ETeam teamId)
	{
		return this.teams[teamId.ordinal()];
	}
	
	public GameTeam[] getTeams()
	{
		return this.teams;
	}
	
	public int getTimeLeft()
	{
		return this.timeLeft;
	}
	
	public boolean isRunning()
	{
		return (this.isRunning && this.startDelay == 0);
	}

	public void setTimeLeft(int seconds)
	{
		this.timeLeft = seconds;
		
		// Update timer
		TimerItem timer = this.getTimer();
		if(timer != null)
		{
			timer.setTimeLeft(this.timeLeft);
		}
	}
	
	public void setStartDelay(int seconds)
	{
		this.startDelay = seconds;
	}
	
	public abstract GameUserPlayer.GameValues createValuesForPlayer();
}
