package habbod2.game.sessions;

import java.util.List;
import java.util.Map;
import java.util.ArrayList;

import habbod2.System;
import habbod2.core.Logger;
import habbod2.communication.*;
import habbod2.net.protocol.*;
import habbod2.net.NetConnectionHandler;
import habbod2.game.GameManager;
import habbod2.game.availability.AvailableAgainStatus;
import habbod2.game.items.ItemInventory;
import habbod2.game.items.pets.PetInventory;
import habbod2.game.items.trading.Trade;
import habbod2.game.messenger.InstantMessenger;
import habbod2.game.moderation.ModerationBan;
import habbod2.game.rooms.*;
import habbod2.game.rooms.instances.*;
import habbod2.game.rooms.navigator.*;
import habbod2.game.rooms.items.wired.triggers.UserEntersRoomTrigger;
import habbod2.game.users.User;
import habbod2.game.users.badges.BadgeInventory;
import habbod2.game.users.effects.EffectInventory;

import habbod2.communication.messages.outgoing.handshake.*;
import habbod2.communication.messages.outgoing.inventory.purse.*;
import habbod2.communication.messages.outgoing.moderation.*;
import habbod2.communication.messages.outgoing.navigator.*;
import habbod2.communication.messages.outgoing.notifications.*;
import habbod2.communication.messages.outgoing.room.action.*;
import habbod2.communication.messages.outgoing.room.chat.*;
import habbod2.communication.messages.outgoing.room.engine.*;
import habbod2.communication.messages.outgoing.room.permissions.*;
import habbod2.communication.messages.outgoing.room.session.*;
import habbod2.communication.outgoing.availability.*;

public class Session implements NetConnectionHandler.CommunicationHandler
{
	// Core
	public final long id;
	private NetConnectionHandler conn;

	// Message handling
	private long lastMsgTime;
	public boolean pongOk;

	// Session variables
	private User user;
	private int accessLogId;
	private Map<Integer, String> ignoredUsers;
	private InstantMessenger messenger;
	private BadgeInventory badgeInventory;
	private ItemInventory itemInventory;
	private PetInventory petInventory;
	private EffectInventory effectInventory;
	private FavoriteRoomList favoriteRooms;
	private List<Integer> ratedRooms;
	
	// Room instance related variables
	private RoomInstance roomInstance;
	private RoomUser roomUser;
	private RoomEvent roomEvent;
	private Trade trade;
	
	public int authenticatedRoomId;
	public int authenticatedTeleporterId;
	public boolean waitingForDoorbell;
	public long chatTimeoutExpireTime;

	public Session(long sessionId, NetConnectionHandler conn, SessionManager sessionMgr)
	{
		this.id = sessionId;
		this.conn = conn;
		this.pongOk = true;
	}
	
	public void onEnd()
	{
		this.conn.close();
		if(this.authenticated())
		{
			this.onLogout();
			System.get().getGameMgr().getSessionMgr().removeAuthedSession(this);
		}
	}
	
	public void end(String reason)
	{
		System.get().getGameMgr().getSessionMgr().endSession(this.id, reason);
	}

	public void authenticate(String ticket)
	{
		// Get game manager
		GameManager gameMgr = System.get().getGameMgr();
				
		// User exists for this auth ticket?
		User user = gameMgr.getUserMgr().getUserByAuthTicket(ticket);
		if(user == null)
		{
			this.end("invalid auth ticket");
			return;
		}

		/*
		if(true)
		{
			this.send(ModMessageComposer.compose("Babyhandel, oorlogsmisdaden, wapenhandel en vier zedendelicten."));
			return;
		}*/
		
		// Any bans in effect?
		List<ModerationBan> bans = gameMgr.getModerationCenter().getActiveBansForUserOrIp(user.id, this.conn.getIpAddress());
		if(!bans.isEmpty())
		{
			ModerationBan ban = bans.get(0);
			this.send(UserBannedMessageComposer.compose(ban));
			return;
		}

		// Insufficient permissions?
		user.rebuildRights();
		if(user.role == null || !user.rights.contains("fuse_login"))
		{
			this.end(String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
		}
		
		// System is available to this user?
		AvailableAgainStatus status = gameMgr.getAvailabilityMgr().getStatusForUser(user);
		if(status != null)
		{
			this.send(LoginFailedHotelClosedMessageComposer.compose(status.hour, status.minute));
			return;
		}

		// User is online!
		this.user = user;
		this.user.updateLastOnline(true);
		gameMgr.getSessionMgr().addAuthedSession(this);
		
		// Complete the login
		this.onLogin();
		this.send(AuthenticationOKMessageComposer.compose());
	}

	private void onLogin()
	{
		// Get game manager
		GameManager gameMgr = System.get().getGameMgr();
		
		// Reload ignored users
		this.ignoredUsers = gameMgr.getUserMgr().getIgnoredUsersForUser(this.user.id);
		
		// Reload messenger
		this.messenger = new InstantMessenger(this);
		this.messenger.reload();
		
		// Reload badges
		this.badgeInventory = new BadgeInventory(this);
		this.badgeInventory.reload();
		
		// Reload item inventory
		this.itemInventory = new ItemInventory(this);
		this.itemInventory.reload();
		
		// Reload pet inventory
		this.petInventory = new PetInventory(this);
		this.petInventory.reload();
		
		// Reload tags
		this.user.tags = gameMgr.getUserMgr().getTagsForUser(this.user.id);
		
		// Reload effects
		this.effectInventory = new EffectInventory(this);
		this.effectInventory.reload();
		
		// Reload favorite rooms
		this.favoriteRooms = new FavoriteRoomList(this);
		this.favoriteRooms.reload();
		
		// Reload rated rooms
		this.ratedRooms = new ArrayList<Integer>();
		
		// Reload achievements
		this.user.achievements = gameMgr.getAchievementMgr().getAchievementsForUser(this.user.id);
	}

	private void onLogout()
	{
		// Stop trading, leave the room, etc etc
		this.clearRoom();
		
		// End the room event, host has gone offline!
		if(this.roomEvent != null)
		{
			RoomEventManager roomEventMgr = System.get().getGameMgr().getRoomEventMgr();
			roomEventMgr.removeEvent(this.roomEvent);
		}

		// Go offline!
		this.messenger.offline();
		
		// Clear inventory to speed up garbagecollection
		this.itemInventory.clear();
		
		// Save badge inventory
		if(this.badgeInventory.requiresSave())
		{
			this.badgeInventory.save();
		}
		this.badgeInventory.clear();

		// Update last online time
		this.user.updateLastOnline(false);

		// Save misc things
		this.user.saveMisc();
	}

	public void tryGetRoomAccess(int roomId, String password)
	{
		// Leave any room the user might be in now
		this.clearRoom();

		// Get/create room instance
		RoomManager roomMgr = System.get().getGameMgr().getRoomMgr();
		RoomInstance room = roomMgr.getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}
		
		// Connection is open
		if(room.getInfo().isGuestRoom())
		{
			this.send(OpenConnectionMessageComposer.compose());
		}
		
		// Determine whether to skip access checks
		boolean skipAccessChecks = ((this.getUser().id == room.getInfo().ownerId) || this.getUser().rights.contains("fuse_can_enter_any_room") || this.authenticatedTeleporterId > 0);
		if(!skipAccessChecks)
		{
			// No error
			int errorCode = 0;
			
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().rights.contains("fuse_enter_full_rooms"))
				{
					errorCode = 1;
				}
			}

			// User is banned?
			else if(room.userIsBanned(this.getUser().id))
			{
				errorCode = 4;
			}
			
			// Any errors?
			if(errorCode != 0)
			{
				this.send(CantConnectMessageComposer.compose(errorCode));
				this.kickFromRoom();
				return;
			}

			// Password room?
			if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equals(password))
			{
				this.sendGenericError(-100002);
				this.kickFromRoom();
				return;
			}

			// Doorbell room?
			else if(room.getInfo().state == Room.STATE_DOORBELL)
			{
				// Ring the doorbell, await response
				if(room.ringDoorbell(this))
				{
					// Cool, let's wait
					this.waitingForDoorbell = true;

					// Waiting...
					this.send(DoorbellMessageComposer.compose(null));
				}
				else
				{
					this.send(FlatAccessDeniedMessageComposer.compose());
					this.kickFromRoom();
				}
				return;
			}
		}
		
		// Authenticated!
		this.authenticatedRoomId = roomId;
		this.goToRoom(roomId);
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomManager roomMgr = System.get().getGameMgr().getRoomMgr();
		RoomInstance instance = roomMgr.getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}
		
		// Room is ready for user
		this.roomInstance = instance;
		this.send(RoomReadyMessageComposer.compose(instance));
		
		// Send maps
		this.send(HeightMapMessageComposer.compose(instance));
		this.send(FloorHeightMapMessageComposer.compose(instance));
		
		// Send remaining chat timeout seconds
		if(this.chatTimeoutExpireTime > 0)
		{
			int remainingSeconds = (int) ((this.chatTimeoutExpireTime - System.currentTimeMillis()) / 1000);
			if(remainingSeconds > 0)
			{
				this.send(FloodControlMessageComposer.compose(remainingSeconds));
			}
		}
		
		// If entering a guestroom, then send more data, client doesn't ask for it
		Room room = instance.getInfo();
		if(room.isPublicRoom())
		{
			this.send(PublicRoomObjectsMessageComposer.compose(instance.getFloorItems()));
		}
		else if(room.isGuestRoom())
		{
			// Send wallpaper/floor/landscape if not default data
			if(!room.wallpaper.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("wallpaper", room.wallpaper));
			}
			if(!room.floor.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("floor", room.floor));
			}
			if(!room.landscape.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("landscape", room.landscape));
			}
			
			// Send room event
			if(instance.getEvent() != null)
			{
				this.send(RoomEventMessageComposer.compose(instance.getEvent()));
			}
			
			// Send room rating
			if(this.getRatedRooms().contains(room.id) || this.getUser().id == room.ownerId)
			{
				this.send(RoomRatingMessageComposer.compose(room.rating));
			}
			else
			{
				this.send(RoomRatingMessageComposer.compose(RoomRatingMessageComposer.NOT_RATED_YET));
			}
			
			// Send visualization settings
			this.send(RoomVisualizationSettingsMessageComposer.compose(room));

			// Send items
			this.send(ObjectsMessageComposer.compose(instance.getFloorItems()));
			this.send(ItemsMessageComposer.compose(instance.getWallItems()));
		}
		
		// Send avatars
		this.send(UsersMessageComposer.compose(instance.getAllAvatars()));

		// Try spawning room user
		this.roomUser = instance.addSession(this);
		if(this.roomUser == null)
		{
			this.kickFromRoom();
			return;
		}
		
		// User entered room!
		UserEntersRoomTrigger.executeTriggers(this);
		
		// Update messenger status
		this.messenger.updateStatus(false);
			
		// Send permissions
		this.sendRoomPermissions();
			
		// Send statuses of all avatars
		this.send(UserUpdateMessageComposer.compose(instance.getAllAvatars()));
		
		// Send other special user statuses
		for(RoomUser user : instance.getAllUsers())
		{
			if(user.isAsleep)
			{
				this.send(SleepMessageComposer.compose(user));
			}
			if(user.danceId > 0)
			{
				this.send(DanceMessageComposer.compose(user));
			}
			if(user.isCarryingItem())
			{
				this.send(CarryObjectMessageComposer.compose(user));
			}
			if(user.getCurrentEffect() > 0)
			{
				this.send(AvatarEffectMessageComposer.compose(user));
			}
		}
		
		// Log entry
		roomMgr.logRoomEntry(this.user.id, this.roomInstance.getInfo().id);
	}
	
	private void clearRoom()
	{
		// Reset room ID
		if(this.authenticatedTeleporterId == 0)
		{
			this.authenticatedRoomId = 0;
		}

		// In a room?
		if(this.roomInstance != null)
		{
			// Log exit
			RoomManager roomMgr = System.get().getGameMgr().getRoomMgr();
			roomMgr.logRoomExit(this.user.id, this.roomInstance.getInfo().id);
			
			// End trade
			if(this.trade != null)
			{
				this.trade.end(this);
				this.trade = null;
			}
			
			// Remove from room
			this.roomInstance.removeSession(this);
			this.roomInstance = null;
			this.roomUser = null;
			
			// Update messenger
			this.messenger.updateStatus(false);
		}
	}

	public void kickFromRoom()
	{
		this.kickFromRoom(null);
	}

	public void kickFromRoom(String message)
	{
		this.kickFromRoom(message, false);
	}
	
	public void kickFromRoom(String message, boolean isCaution)
	{
		this.clearRoom();
		this.send(CloseConnectionMessageComposer.compose());
		if(message != null && !message.isEmpty())
		{
			this.sendNotif(message, isCaution);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Get logger
		Logger logger = System.get().getLogger();
		
		// Log that this session received a message!
		this.lastMsgTime = System.get().getTime();
		
		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable stuff
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			this.end("MSG_EOF detected in incoming message");
			System.get().getServer().blacklistIp(this.conn.getIpAddress(), "MSG_EOF");
			return;
		}

		// Only handle these messages pre-auth
		if(!this.authenticated() && (msg.type != EClientMessage.InitCrypto && msg.type != EClientMessage.SSOTicket))
		{
			//logger.warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.type, msg.header()));
			return;
		}
		
		// Lookup message handler and execute handling logic
		String name = RequestRegistry.getRequestName(msg.type);
		RequestHandler handler = RequestRegistry.getRequestHandler(msg.type);
		if(handler != null)
		{
			logger.debug(this, "<-- " + msg.toString() + " (" + name + ")");
			try
			{
				// Handle in separate thread or in Netty I/O thread?
				if(handler instanceof DeferredRequestHandler)
				{
					DeferredRequestHandlerExecutor.submit(this, msg, (DeferredRequestHandler) handler);
				}
				else
				{
					handler.handle(this, msg);
				}
			}
			catch(Exception ex)
			{
				logger.error(this, String.format("Error handling msg #%d (\"%s\").", msg.type, name), ex);
			}
		}
		else
		{
			logger.debug(this, "<-- " + msg.toString());
			logger.debug(this, String.format("RequestRegistry misses #%d. (\"%s\" -> %s)", msg.type, msg.header(), name));
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		System.get().getGameMgr().getSessionMgr().endSession(this.id, reason);
	}

	public void send(ServerMessage msg)
	{
		System.get().getLogger().debug(this, "--> " + msg.toString());
		this.conn.send(msg);
	}

	public void sendGenericError(int code)
	{
		this.send(GenericErrorMessageComposer.compose(code));
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null, false);
	}

	public void sendNotif(String text, boolean isCaution)
	{
		this.sendNotif(text, null, isCaution);
	}
	
	public void sendNotif(String text, String url, boolean isCaution)
	{
		this.send(BroadcastMessageComposer.compose(text));
		/*this.msg.init(isCaution ? EServerMessage.Broadcast : EServerMessage.UserBanned);
		this.msg.appendStr(text);
		this.msg.appendStr(url);
		this.send(this.msg);*/
	}

	public void sendSystemNotif(String message)
	{
		this.send(MOTDNotificationMessageComposer.compose(message));
	}
	
	public void sendWhisperNotif(String message)
	{
		if(this.inRoom())
		{
			this.send(WhisperMessageComposer.compose(this.getRoomUser(), message));
		}
	}

	public void sendCreditBalance()
	{
		this.send(CreditBalanceMessageComposer.compose(this.user));
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		this.send(ActivityPointNotificationMessageComposer.compose(this.user.pixels, added));
	}
	
	public void sendPearlBalance()
	{
		this.send(ActivityPointsMessageComposer.compose(this.user.pixels, this.user.pearls));
	}

	public void sendRoomPermissions()
	{
		if(this.roomUser != null)
		{
			if(this.roomUser.hasRoomRights)
			{
				this.send(YouAreControllerMessageComposer.compose());
			}
			if(this.roomUser.hasRoomOwner)
			{
				this.send(YouAreOwnerMessageComposer.compose());
			}
		}
	}

	public NetConnectionHandler getConnection()
	{
		return this.conn;
	}

	public long getLastMsgTime()
	{
		return this.lastMsgTime;
	}

	public boolean authenticated()
	{
		return(this.user != null);
	}

	public User getUser()
	{
		return this.user;
	}

	public int getAccessLogId()
	{
		return this.accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.accessLogId == 0)
		{
			this.accessLogId = accessLogId;
		}
	}

	public Map<Integer, String> getIgnoredUsers()
	{
		return this.ignoredUsers;
	}
	
	public boolean hasIgnoredUser(int userId)
	{
		return this.ignoredUsers.containsKey(userId);
	}
	
	public InstantMessenger getMessenger()
	{
		return this.messenger;
	}
	
	public BadgeInventory getBadgeInventory()
	{
		return this.badgeInventory;
	}
	
	public ItemInventory getItemInventory()
	{
		return this.itemInventory;
	}
	
	public PetInventory getPetInventory()
	{
		return this.petInventory;
	}
	
	public EffectInventory getEffectInventory()
	{
		return this.effectInventory;
	}
	
	public FavoriteRoomList getFavoriteRooms()
	{
		return this.favoriteRooms;
	}
	
	public List<Integer> getRatedRooms()
	{
		return this.ratedRooms;
	}

	public boolean inRoom()
	{
		return(this.roomInstance != null && this.roomUser != null);
	}

	public RoomInstance getRoomInstance()
	{
		return this.roomInstance;
	}

	public RoomUser getRoomUser()
	{
		return this.roomUser;
	}
	
	public RoomEvent getRoomEvent()
	{
		return this.roomEvent;
	}
	
	public void setRoomEvent(RoomEvent event)
	{
		this.roomEvent = event;
	}
	
	public boolean isTrading()
	{
		return (this.trade != null);
	}
	
	public Trade getTrade()
	{
		return this.trade;
	}
	
	public void setTrade(Trade trade)
	{
		this.trade = trade;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
