package habbod2.game.users;

import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;

import habbod2.System;
import habbod2.game.GameComponent;
import habbod2.game.GameManager;
import habbod2.game.sessions.Session;
import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;

public class UserManager extends GameComponent
{
	public UserManager(GameManager gameMgr)
	{
		super(gameMgr);
	}
	
	public void start()
	{
		this.resetDbUserOnlineStatus();
	}
	
	private void resetDbUserOnlineStatus()
	{
		this.getDb().execute("UPDATE users SET online=0");
	}
	
	public User getUserById(int userId)
	{
		return this.getUserById(userId, true);
	}
	
	public User getUserById(int userId, boolean lookupSessions)
	{
		// No need for database lookup if user is online
		if(lookupSessions)
		{
			Session sess = this.getGameMgr().getSessionMgr().getSessionByUserId(userId);
			if(sess != null)
			{
				return sess.getUser();
			}
		}

		// Lookup database...
		DatabaseResult result = this.getDb().query("SELECT * FROM users WHERE id=" + userId);
		User user = result.row() ? this.parseUserFromRow(result) : null;
		result.release();
		
		return user;
	}

	public User getUserByUsername(String username)
	{
		return this.getUserByUsername(username, true);
	}
	
	public User getUserByUsername(String username, boolean lookupSessions)
	{
		// No need for database lookup if user is online
		if(lookupSessions)
		{
			Session sess = this.getGameMgr().getSessionMgr().getSessionByUsername(username);
			if(sess != null)
			{
				return sess.getUser();
			}
		}

		// Lookup database...
		DatabaseResult result = this.getDb().query("SELECT * FROM users WHERE username='" + SQLHelper.escape(username) + "'");
		User user = result.row() ? this.parseUserFromRow(result) : null;
		result.release();
		
		return user;
	}

	public User getUserByAuthTicket(String ticket)
	{
		if(this.getProperties().getBool("game.login.debug"))
		{
			String username = ticket.substring("ST-".length());
			return this.getUserByUsername(username);
		}
		
		DatabaseResult result = this.getDb().query("SELECT * FROM users WHERE auth_ticket='" + SQLHelper.escape(ticket) + "'");
		User user = result.row() ? this.parseUserFromRow(result) : null;
		result.release();
		
		return user;
	}

	public int getUserId(String username)
	{
		// No need for database lookup if user is online
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUsername(username);
		if(sess != null)
		{
			return sess.getUser().id;
		}
		
		// Database lookup...
		DatabaseResult result = this.getDb().query("SELECT id FROM users WHERE username='" + SQLHelper.escape(username) + "'");
		int userId = result.row() ? result.getInt("id") : 0;
		result.release();
		
		return userId;
	}
	
	public String getUsername(int userId)
	{
		// No need for database lookup if user is online
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUserId(userId);
		if(sess != null)
		{
			return sess.getUser().username;
		}
		
		// Database lookup...
		DatabaseResult result = this.getDb().query("SELECT username FROM users WHERE id=" + userId);
		String username = result.row() ? result.getString("username") : null;
		result.release();
		
		return username;
	}
	
	public List<String> getExtraRightsForUser(int userId)
	{
		List<String> rights = new ArrayList<String>();
		
		DatabaseResult result = this.getDb().query("SELECT `right` FROM user_rights WHERE user_id=" + userId);
		while(result.row())
		{
			rights.add(result.getString("right"));
		}
		result.release();
		
		return rights;
	}
	
	public void logCreditTransaction(int userId, int amount, String description)
	{
		this.getDb().execute("INSERT INTO user_transactions(user_id,timestamp,description,amount) VALUES (" + userId + ",'" + SQLHelper.getDateTime() + "','" + SQLHelper.escape(description) + "'," + amount + ")");
	}

	public Map<Integer, String> getIgnoredUsersForUser(int userId)
	{
		Map<Integer, String> userIds = new HashMap<Integer, String>();
		
		DatabaseResult result = this.getDb().query("SELECT u.id,u.username FROM user_ignores i JOIN users u ON u.id = i.ignore_id WHERE i.user_id=" + userId);
		while(result.row())
		{
			userIds.put(result.getInt("id"), result.getString("username"));
		}
		result.release();
		
		return userIds;
	}
	
	public void ignoreUser(int userId, int ignoreId)
	{
		this.getDb().execute("INSERT INTO user_ignores(user_id,ignore_id) VALUES (" + userId + "," + ignoreId + ")");
	}
	
	public void unignoreUser(int userId, int ignoreId)
	{
		this.getDb().execute("DELETE FROM user_ignores WHERE user_id=" + userId + " AND ignore_id=" + ignoreId);
	}
	
	public List<String> getTagsForUser(int userId)
	{
		List<String> tags = new ArrayList<String>();
		
		DatabaseResult result = this.getDb().query("SELECT tag FROM user_tags WHERE user_id=" + userId);
		while(result.row())
		{
			tags.add(result.getString("tag"));
		}
		result.release();
		
		return tags;
	}
	
	public List<WardrobeOutfit> getWardrobeForUser(int userId)
	{
		List<WardrobeOutfit> entries = new ArrayList<WardrobeOutfit>();
		
		DatabaseResult result = this.getDb().query("SELECT slot,figure,gender FROM user_wardrobe WHERE user_id=" + userId);
		while(result.row())
		{
			WardrobeOutfit entry = new WardrobeOutfit();
			entry.slot = result.getInt("slot");
			entry.figure = result.getString("figure");
			entry.gender = result.getString("gender").charAt(0);
			entries.add(entry);
		}
		result.release();
		
		return entries;
	}
	
	public void modifyWardrobeForUser(int userId, int slot, String figure, char gender)
	{
		this.getDb().execute("DELETE FROM user_wardrobe WHERE user_id=" + userId + " AND slot=" + slot);
		this.getDb().execute("INSERT INTO user_wardrobe(user_id,slot,figure,gender) VALUES (" + userId + "," + slot + ",'" + SQLHelper.escape(figure) + "','" + gender + "')");
	}
	
	public void logRoomEntry(int userId, int roomId)
	{
		this.getDb().execute("INSERT INTO user_roomvisits(user_id,room_id,entry) VALUES (" + userId + "," + roomId + ",'" + SQLHelper.getDateTime() + "')");
	}
	
	public void logRoomExit(int userId, int roomId)
	{
		this.getDb().execute("UPDATE user_roomvisits SET `exit` = '" + SQLHelper.getDateTime() + "' WHERE user_id=" + userId + " AND room_id=" + roomId + " AND `exit` IS NULL LIMIT 1");
	}
	
	private User parseUserFromRow(DatabaseResult row)
	{
		User user = new User();
		user.id = row.getInt("id");
		user.username = row.getString("username");
		user.realName = row.getString("realname");
		user.role = System.get().getGameMgr().getAccessCtrl().getUserRole(row.getInt("role_id"));
		user.registeredTime = row.getTime("registered");
		user.lastOnlineTime = row.getTime("lastonline");
		user.onlineNow = row.getBoolean("online");
		user.figure = row.getString("figure");
		user.gender = row.getString("gender").equals("F") ? 'F' : 'M';
		user.motto = row.getString("motto");
		user.credits = row.getInt("credits");
		user.pixels = row.getInt("pixels");
		user.pearls = row.getInt("pearls");
		user.homeRoomId = row.getInt("homeroom_id");
		user.respect = row.getInt("respect");
		user.respectPoints = row.getInt("respect_points");
		user.petRespectPoints = row.getInt("petrespect_points");
		
		return user;
	}
}
