package habbod2;

import java.io.FileInputStream;
import java.util.Properties;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import habbod2.core.Configuration;
import habbod2.core.Locale;
import habbod2.core.ShutdownHook;
import habbod2.core.StatsMonitor;
import habbod2.game.GameManager;
import habbod2.net.NetServer;
import habbod2.rcon.RconServer;
import habbod2.storage.Database;
import habbod2.storage.DatabaseEndpoint;
import habbod2.util.Logger;
import habbod2.util.NamedThreadFactory;
import habbod2.util.ScheduledTask;
import habbod2.util.ScriptHandler;
import habbod2.util.StringUtil;
import habbod2.communication.RequestRegistry;

public class System
{
	private static final String propertiesFile = "habbod2.properties";
	private static final String versionString = "habbod2 v2.4.20-holo 'Silverback'";
	
	private long startTime;
	private Logger logger;
	private Configuration config;
	private Database database;
	private Locale locale;
	private RconServer rcon;
	private NetServer server;
	private GameManager gameMgr;
	private StatsMonitor stats;
	private ScheduledExecutorService scheduler;
	private boolean isShutdown;
	
	public System()
	{
		this.logger = new Logger();
		this.config = new Configuration(this);
		this.locale = new Locale(this);
		this.rcon = new RconServer(this);
		this.server = new NetServer(this);
		this.gameMgr = new GameManager(this);
		this.stats = new StatsMonitor(this);
		this.scheduler = Executors.newScheduledThreadPool(1, new NamedThreadFactory("SystemScheduler #%d"));
	}
	
	public boolean start(String propsFile)
	{
		// Record boot start time
		long bootBegin = this.getTime();
		
		// Set output file for logger
		this.logger.setDatedOutputFile();
		
		// Print version
		this.logger.info(this, String.format("Booting %s...", this.getVersion()));
		
		// Load .properties file
		Properties props = new Properties();
		try
		{
			props.load(new FileInputStream(System.propertiesFile));
		}
		catch(Exception ex)
		{
			this.logger.warning(this, String.format("Could not load %s! Please validate the habbod2 installation.", System.propertiesFile));
			return false;
		}
		
		// Enable debug and stacktraces?
		if(StringUtil.parseInt(props.getProperty("sys.log.debug")) == 1)
		{
			this.logger.setMinImportance(Logger.ELogLevel.Debug);
		}
		this.logger.setDumpStackTraces(StringUtil.parseInt(props.getProperty("sys.log.stacktraces")) == 1);
		
		// Check support for LUA scripting
		if(ScriptHandler.hasNativeLibrary)
		{
			this.logger.info("System", "Lua scripting support available. (native)");
		}
		else
		{
			this.logger.warning("System", "No Lua scripting support available, because the native library 'luajava' is not present in java.library.path.");
		}
		
		// Setup database
		this.database = new Database(this, new DatabaseEndpoint(props.getProperty("db.user", ""), props.getProperty("db.password", ""), props.getProperty("db.url", ""), props.getProperty("db.driver", ""), StringUtil.parseInt(props.getProperty("db.pool.maxsize"))));
		if(!this.database.test())
		{
			return false;
		}
		this.database.getPool().startMonitor();
		if(StringUtil.parseInt(props.getProperty("db.updates.queue.enabled")) == 1)
		{
			this.database.getUpdateQueue().open();
		}
		
		// Load actual configuration
		this.config.reload();
		
		// Reload locale
		this.locale.setLocale(this.config.get("sys.locale", "en-US"));
		
		// Reload + start game manager
		this.gameMgr.updateVersion();
		this.gameMgr.reload();
		this.gameMgr.start();
		
		// Reload request registry
		RequestRegistry.reload();
		
		// Enable RCON?
		if(this.config.getBool("rcon.enabled"))
		{
			// Load commands
			this.rcon.reload();
			
			// Allow local commands?
			if(this.config.getBool("rcon.allowlocal"))
			{
				this.rcon.allowLocalCommands();
			}
			
			// Attempt to start server
			if(!this.rcon.start(this.config.getInt("rcon.port")))
			{
				return false;
			}
		}
		
		// Enable gameserver?
		if(this.config.getBool("net.enabled"))
		{
			// Reload blacklist
			this.server.reloadBlacklist();
			
			// Attempt to start server
			if(!this.server.start(this.config.getInt("net.port")))
			{
				return false;
			}
		}
		
		// Start stats monitor
		int statsIntervalSeconds = this.config.getInt("sys.stats.update.seconds", 30);
		this.scheduleTaskStartNow(this.stats, statsIntervalSeconds * 1000);
		
		// Attach shutdown hook
		this.getVM().addShutdownHook(new ShutdownHook(this));
		
		// Record start time
		this.startTime = this.getTime();
		this.logger.info(this, String.format("Booted in %d ms, ready for connections.", this.startTime - bootBegin));
		
		// OK!
		return true;
	}

	public void stop(String message)
	{
		// Already shutdown?
		if(this.isShutdown)
		{
			return;
		}
		
		// Entering shutdown sequence
		this.isShutdown = true;
		this.logger.info(this, String.format("Shutting down. (reason: %s)", message));
		
		// Stop any running servers first, no new connections!
		this.rcon.stop();
		this.server.stop();
		
		// Shutdown game systems
		this.gameMgr.shutdown();
		
		// Forcefully close any open connections and release resources
		this.server.close();
		
		// Shutdown scheduler
		this.scheduler.shutdownNow();
		
		// Update stats
		this.stats.run();
		
		// Shutdown database
		this.database.getPool().stopMonitor();
		this.database.getUpdateQueue().close();
		
		// Shutdown completed
		this.logger.info(this, String.format("Shutdown completed."));
	}
	
	public void executeTask(Runnable task)
	{
		this.scheduler.execute(task);
	}
	
	public void executeTask(Runnable task, long delay)
	{
		this.scheduler.schedule(task, delay, TimeUnit.MILLISECONDS);
	}
	
	public ScheduledTask scheduleTask(Runnable task, long interval)
	{
		return this.scheduleTask(task, interval, interval);
	}

	public ScheduledTask scheduleTaskStartNow(Runnable task, long interval)
	{
		return this.scheduleTaskWithFirstTime(task, interval, System.currentTimeMillis());
	}
	
	public ScheduledTask scheduleTaskWithFirstTime(Runnable task, long interval, long firstTime)
	{
		long initialDelay = firstTime - System.currentTimeMillis();
		return this.scheduleTask(task, initialDelay, interval);
	}
	
	public ScheduledTask scheduleTask(Runnable task, long initialDelay, long interval)
	{
		return new ScheduledTask(this.scheduler.scheduleAtFixedRate(task, initialDelay, interval, TimeUnit.MILLISECONDS));
	}
	
	public String getVersion()
	{
		return System.versionString;
	}
	
	public final long getTime()
	{
		return java.lang.System.currentTimeMillis();
	}

	public static long currentTimeMillis()
	{
		return java.lang.System.currentTimeMillis();
	}
	
	public long getMemoryUsage()
	{
		Runtime vm = this.getVM();
		return (vm.totalMemory() - vm.freeMemory()) / 1024;
	}
	
	public long getStartTime()
	{
		return this.startTime;
	}
	
	public Runtime getVM()
	{
		return Runtime.getRuntime();
	}
	
	public Logger getLogger()
	{
		return this.logger;
	}
	
	public Configuration getConfig()
	{
		return this.config;
	}
	
	public Database getDb()
	{
		return this.database;
	}
	
	public Locale getLocale()
	{
		return this.locale;
	}
	
	public RconServer getRCON()
	{
		return this.rcon;
	}
	
	public NetServer getServer()
	{
		return this.server;
	}
	
	public GameManager getGameMgr()
	{
		return this.gameMgr;
	}
	
	@Override
	public String toString()
	{
		return "System";
	}
	
	public static abstract class Component
	{
		private final System system;
		
		public Component(System system)
		{
			this.system = system;
		}
		
		public System getSystem()
		{
			return this.system;
		}
		
		public Logger getLogger()
		{
			return this.system.getLogger();
		}
		
		public Configuration getConfig()
		{
			return this.system.getConfig();
		}
		
		public Database getDb()
		{
			return this.system.getDb();
		}
		
		@Override
		public String toString()
		{
			return this.getClass().getSimpleName();
		}
	}
}
