package habbod2.game.pets;

import habbod2.Instance;
import habbod2.communication.messages.outgoing.room.engine.PetExperienceMessageComposer;
import habbod2.communication.messages.outgoing.room.engine.PetLevelNotificationMessageComposer;
import habbod2.game.pets.ai.*;
import habbod2.game.pets.ai.special.*;
import habbod2.game.rooms.Room;
import habbod2.game.rooms.geometry.Direction;
import habbod2.game.rooms.instance.RoomActor;
import habbod2.game.rooms.instance.RoomHuman;
import habbod2.game.rooms.instance.RoomTile;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.sessions.Session;
import habbod2.game.util.GameUtil;

// Emoticons: sml, sad, plf (play), hng (hunger), thr (thirsty), nst (nest)
// Actions: dragon swg (swing on item), wng (fold wings)

public class RoomPet extends RoomActor implements RoomUser.UserObserver
{
	private static final int MOVE_RADIUS = 10;
	
	private Pet pet;
	private RoomHuman rider;
	
	private int talkTimer;
	private int moveTimer;
	private int gestureTimer;
	private int energyTimer;
	
	private String pendingMessage;
	private int pendingMessageTimer;
	
	private ActionHandler action;
	private AccessoryFinder finder;
	
	public RoomPet(Pet pet)
	{
		// Setup actor data
		this.pet = pet;
		this.pathfindingAllowedOn = RoomTile.OPEN;
		if(this.pet.type.canSwim)
		{
			this.pathfindingAllowedOn |= RoomTile.WATER;
		}
		this.finder = new AccessoryFinder(this);
		
		// Initialize
		if(pet.eatTimer == 0) pet.resetEatTimer(true);
		if(pet.drinkTimer == 0) pet.resetDrinkTimer(true);
		if(pet.sleepTimer == 0) pet.resetSleepTimer(true);
	}
	
	public void syncAndSaveLocationInfo(Room room)
	{
		if(room != null && room.ownerId == this.pet.ownerId)
		{
			// Move to room
			this.pet.roomId = room.id;
			this.pet.x = this.x;
			this.pet.y = this.y;
			this.pet.z = this.z;
		}
		else
		{
			// Move to inventory
			this.pet.roomId = 0;
			this.pet.x = 0;
			this.pet.y = 0;
			this.pet.z = 0;
		}
		this.pet.save();
	}
	
	@Override
	public void onUserEnter(RoomUser user)
	{
		if(user.getInfo().id == this.pet.ownerId)
		{
			this.greetOwner();
		}
	}

	@Override
	public void onUserLeave(RoomUser user)
	{
		if(user.getInfo().id == this.pet.ownerId)
		{
			if(this.pet.roomId == 0)
			{
				this.getRoom().removeActor(this);
				user.getSession().getPetInventory().addPet(this.pet);
			}
		}
	}

	@Override
	protected void onChatFromActor(RoomActor from, String text, boolean isShout)
	{
		// Only respond to roonmusers
		if(!(from instanceof RoomUser))
		{
			return;
		}
		
		// Can't respond now?
		if(this.hasRider() || this.action instanceof SleepAction)
		{
			return;
		}

		// Find out if text starts with pet name (and is more than just the name)
		int separatorIndex = text.indexOf(' ');
		if(separatorIndex == -1 || !text.startsWith(this.getInfo().name))
		{
			return;
		}
		
		// Get user
		RoomUser user = (RoomUser)from;
		
		// Get and validate command
		String strCommand = text.substring(separatorIndex + 1);
		if(Character.isUpperCase(strCommand.charAt(0)))
		{
			strCommand = strCommand.toLowerCase();
		}
		PetCommand command = this.pet.type.commands.get(strCommand);
		if(command == null || (command.isOwnerOnly && user.getInfo().id != this.pet.ownerId))
		{
			return;
		}
		
		// Look to the pet?
		if(command.action != "come_here")
		{
			user.lookTo(this, false);
		}
		
		// Run some checks
		if(command.level > this.pet.level || this.pet.happiness < 20 || this.pet.energy < 20 || !GameUtil.getRandomBool(this.pet.getObeyChance()))
		{
			this.pet.modifyEnergy(-5);
			this.sayEventMessage("noObey");
			this.setAction(new TimedStatusAction(this, "sit", 3));
			this.setDirection(Direction.calculate(this.x, this.y, user.x, user.y).invert());
		}
		else
		{
			// Rotate body?
			if(command.action != "forward" && command.action != "lookto_left" && command.action != "lookto_right")
			{
				this.lookTo(user, false);
			}
			
			// Process the command
			this.executeCommand(command, user);
			
			// Take energy
			this.pet.modifyEnergy(-10);
			
			// Give/take experience?
			if(command.experience != 0)
			{
				this.modifyExperience(command.experience);
			}
		}
	}
	
	private void executeCommand(PetCommand command, RoomUser user)
	{
		// Stop following an actor
		this.unfollow();
		
		// Simple status?
		if(command.isStatus)
		{
			this.setAction(new TimedStatusAction(this, command.action, 5));
		}
		else
		{
			// 'Intelligent' logic
			switch(command.action)
			{
				// Stop doing everything ('free')
				case "free":
					this.release();
					break;
					
				// Move to tile in front of user
				case "come_here":
					this.setAction(new InspectActorAction(this, user));
					break;
			
				// Stay at current position?
				case "stay":
				case "stand":
					this.moveTimer = GameUtil.secondsToRoomPulses(60);
					break;
				
				// Follow user around the room
				case "follow_behind":
					this.follow(user, Direction.South);
					break;
				
				// Follow left
				case "follow_left":
					this.follow(user, Direction.West);
					break;
					
				// Follow right
				case "follow_right":
					this.follow(user, Direction.East);
					break;
						
				// Force chatter
				case "voice":
					this.talkTimer = GameUtil.secondsToRoomPulses(1);
					break;
					
				// Find a toy
				case "play":
					this.setAction(new PlayAction(this));
					break;
				
				// Mute for x seconds
				case "shutup":
					this.sayEventMessage("shutup");
					this.talkTimer = GameUtil.secondsToRoomPulses(5 * 60);
					break;
					
				// Return to nest
				case "nest":
					this.setAction(new NestRestAction(this));
					break;
			
				// Eat something
				case "eat":
					this.setAction(new EatAction(this));
					break;
						
				// Drink something
				case "drink":
					this.setAction(new DrinkAction(this));
					break;
	
				// Beg for a treat
				case "beg_treat":
					//this.setAction(new BegForTreatAction(this, user));
					break;
					
				// Move one forward
				case "forward":
					this.setGoal(this.getTileInFront());
					break;
				
				// Rotate to the left
				case "lookto_left":
					this.setDirection(this.bodyDirection.transform(Direction.NorthWest));
					break;
				
				// Rotate to the right
				case "lookto_right":
					this.setDirection(this.bodyDirection.transform(Direction.NorthEast));
					break;
				
				// Play around with ball (with user)
				case "play_ball":
					///this.setAction(new PlayBallAction(this, user));
					break;
					
				// Toggle TV close to user
				case "toggle_tv":
					//this.setAction(new ToggleTvAction(this, user));
					break;
					
				// Dive into some kind of water
				case "dive_water":
					// TODO: find some water (or a water toy) and dive!
					this.say("?????!");
					this.release();
					break;
					
				// Hang to a tree
				case "toy_hang":
					this.setAction(new HangAction(this));
					break;
				
				// Hang to a tree, swing
				case "toy_hang_swing":
					this.setAction(new SwingAction(this));
					break;
					
				// Hang to a tree, spit ring of fire
				case "toy_hang_fire":
					this.setAction(new SwingAction(this));
					break;
			}
		}
	}
	
	private void tickEnergyTimer()
	{
		if(this.energyTimer > 0)
		{
			this.energyTimer--;
		}
		else
		{
			this.pet.modifyEnergy(+1);
			this.energyTimer = GameUtil.secondsToRoomPulses(10);
		}
	}
	
	private void tickMoveTimer()
	{
		if(this.moveTimer > 0)
		{
			this.moveTimer--;
		}
		else
		{
			// Brute force for existing tile
			RoomTile tile = null;
			while(tile == null)
			{
				tile = this.getRoom().getMap().getTile(this.x + GameUtil.getRandomInt(-MOVE_RADIUS, MOVE_RADIUS), this.y + GameUtil.getRandomInt(-MOVE_RADIUS, MOVE_RADIUS));
			}
			
			// Tile reachable?
			if(!tile.isReachableBy(this))
			{
				this.moveTimer = GameUtil.secondsToRoomPulses(1);
			}
			else
			{
				this.setGoal(tile);
				this.moveTimer = GameUtil.secondsToRoomPulses(GameUtil.getRandomInt(5, 30));
			}
		}
	}
	
	private void tickTalkTimer()
	{
		if(this.talkTimer > 0)
		{
			this.talkTimer--;
		}
		else
		{
			String message = this.pet.type.getRandomMessage("chatter");
			if(message != null)
			{
				this.chat(message.replace("%owner%", this.pet.ownerName), false);
			}
			this.talkTimer = GameUtil.secondsToRoomPulses(GameUtil.getRandomInt(30, 100));
		}
	}
	
	private void tickPendingMessageTimer()
	{
		if(this.pendingMessageTimer > 0 && --this.pendingMessageTimer == 0)
		{
			this.chat(this.pendingMessage, false);
			this.pendingMessage = null;
		}
	}
	
	private void tickGestureTimer()
	{
		if(this.gestureTimer > 0)
		{
			this.gestureTimer--;
			if(this.gestureTimer == 0)
			{
				this.statuses.remove("gst");
			}
		}
	}
	
	private void tickEatTimer()
	{
		if(this.pet.eatTimer > 0)
		{
			this.pet.eatTimer--;
			// TODO: account for energy
		}
		else
		{
			if(this.isIdle())
			{
				this.setAction(new EatAction(this));
			}
		}
	}
	
	private void tickDrinkTimer()
	{
		if(this.pet.drinkTimer > 0)
		{
			this.pet.drinkTimer--;
		}
		else
		{
			if(this.isIdle())
			{
				this.setAction(new DrinkAction(this));
			}
		}
	}
	
	private void tickSleepTimer()
	{
		if(this.pet.sleepTimer > 0)
		{
			this.pet.sleepTimer--;
			// TODO: account for sleeping
		}
		else
		{
			if(this.isIdle())
			{
				this.action = new SleepAction(this);
			}
		}
	}
	
	@Override
	public void pulse()
	{
		// Handle low-level logic
		super.pulse();
		
		// Tick internal timers
		this.tickEnergyTimer();
		this.tickGestureTimer();
		this.tickPendingMessageTimer();
		this.tickTalkTimer();
		
		// Tick persistent timers (can spawn {Eat/Drink/Sleep}Action)
		this.tickEatTimer();
		this.tickDrinkTimer();
		this.tickSleepTimer();
		
		// Process action?
		if(this.action != null)
		{
			this.action.pulse();
		}
		
		// Autonomous behavior
		else if(!this.hasRider() && !this.isFollowing())
		{
			this.tickMoveTimer();
		}
	}
	
	public void say(String message)
	{
		// Proceed?
		if(message == null)
		{
			return;
		}
		
		// Flush pending message
		if(this.pendingMessage != null)
		{
			this.chat(this.pendingMessage, false);
		}
		
		// Prepare this message
		message = message.replace("%owner%", this.pet.ownerName);
		
		// Schedule this message
		this.pendingMessage = message;
		this.pendingMessageTimer = GameUtil.secondsToRoomPulses(1);
	}
	
	public void sayEventMessage(String event)
	{
		this.say(this.pet.type.getRandomMessage(event));
	}
	
	public boolean isIdle()
	{
		return (this.action == null && !this.hasRider());
	}
	
	public void release()
	{
		// Clear action?
		if(this.action != null)
		{
			this.action.clear();
			this.action = null;
		}
		
		// Unfollow
		this.unfollow();
		
		// Force move
		this.moveTimer = 1;
	}
	
	public void setAction(ActionHandler action)
	{
		this.release();
		this.action = action;
	}
	
	public void setGesture(String gesture)
	{
		this.addStatus("gst", gesture);
		this.gestureTimer = GameUtil.secondsToRoomPulses(3);
	}
	
	public void modifyExperience(int experience)
	{
		// Record current level
		int previousLevel = this.pet.level;
		
		// Modify experience
		this.pet.experience += experience;
		if(this.pet.experience < 0)
		{
			this.pet.experience = 0;
		}
		this.pet.recalculateLevel();
		this.getRoom().broadcast(PetExperienceMessageComposer.compose(this, experience));
		
		// Pet leveled up?
		if(this.pet.level != previousLevel)
		{
			// Yay!
			this.sayEventMessage("levelUp");
			this.pet.modifyHappiness(+10);
			
			// Owner online?
			Session owner = Instance.get().getGameMgr().getSessionMgr().getSessionByUserId(this.pet.ownerId);
			if(owner != null)
			{
				owner.send(PetLevelNotificationMessageComposer.compose(this.pet));
			}
		}
	}
	
	public void greetOwner()
	{
		this.sayEventMessage("greetOwner");
	}
	
	@Override
	public void invalidate()
	{
		super.invalidate();
		if(this.hasRider())
		{
			this.getRider().setMount(null);
		}
	}
	
	public RoomHuman getRider()
	{
		return this.rider;
	}
	
	public void setRider(RoomHuman human)
	{
		this.rider = human;
		this.unfollow();
		this.clearMovement();
		this.release();
	}
	
	public boolean hasRider()
	{
		return (this.rider != null);
	}
	
	@Override
	public String getActorKey()
	{
		return "p" + this.pet.id;
	}
	
	public Pet getInfo()
	{
		return this.pet;
	}
	
	public AccessoryFinder getFinder()
	{
		return this.finder;
	}
}
