package habbod2.game.rooms.instance;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import habbod2.net.protocol.ServerMessage;
import habbod2.communication.messages.outgoing.room.chat.ChatMessageComposer;
import habbod2.communication.messages.outgoing.room.chat.ShoutMessageComposer;
import habbod2.communication.messages.outgoing.room.engine.UserUpdateMessageComposer;
import habbod2.game.pets.RoomPet;
import habbod2.game.rooms.geometry.Direction;
import habbod2.game.rooms.pathfinding.Pathfinder;
import habbod2.game.rooms.pathfinding.PathfinderNode;
import habbod2.game.util.ChatUtil;

public abstract class RoomActor implements RoomTile.TileObject
{
	// Room
	public int actorId;
	private RoomInstance room;
	private boolean invalidated;
	public int idleTime;
	
	// Position
	public int x;
	public int y;
	public float z;
	public Direction headDirection;
	public Direction bodyDirection;
	
	// Statuses
	public HashMap<String, String> statuses;
	public boolean requiresUpdate;

	// Movement
	public RoomTile goal;
	public RoomTile next;
	private boolean moveLock;
	public boolean overrideNextTile;
	public Pathfinder pathfinder;
	public List<PathfinderNode> path;
	public boolean calculatePath;
	public byte pathfindingAllowedOn;
	public boolean isReverseWalk;
	
	// Following
	private RoomActor follow;
	public Direction followMode;
	
	public boolean isAllFollowMe;

	public RoomActor()
	{
		this.headDirection = Direction.North;
		this.bodyDirection = Direction.North;
		this.statuses = new HashMap<String, String>();
		this.path = Collections.emptyList();
		this.clearMovement();
	}
	
	public RoomInstance getRoom()
	{
		return this.room;
	}

	public void setRoom(RoomInstance room)
	{
		this.room = room;
		this.pathfinder = new Pathfinder(this);
	}

	public void invalidate()
	{
		this.invalidated = true;
	}

	public boolean invalidated()
	{
		return this.invalidated;
	}

	public void pulse()
	{
		// Increase idle-time
		this.idleTime++;
		
		// Follow actor?
		RoomActor follow = this.getFollowing();
		if(follow != null)
		{	
			// Move towards the correct tile if not there already
			RoomTile tile = follow.getAdjacentTile(this.followMode);
			if(this.at(tile))
			{
				this.setDirection(Direction.calculate(this.x, this.y, follow.x, follow.y));
			}
			else
			{
				this.setGoal(tile);
			}
		}
	}

	public void unidle()
	{
		this.idleTime = 0;
	}

	public float getExtraDisplayHeight()
	{
		return 0;
	}
	
	public boolean at(RoomTile tile)
	{
		return(tile != null && this.x == tile.x && this.y == tile.y);
	}
	
	public boolean at(RoomTile.TileObject obj)
	{
		return this.at(obj.getTile());
	}
	
	public boolean touches(RoomTile tile)
	{
		return tile.touches(this);
	}
	
	public boolean touches(RoomTile.TileObject obj)
	{
		return this.touches(obj.getTile());
	}
	
	@Override
	public RoomTile getTile()
	{
		return this.getRoom().getMap().getTile(this.x, this.y);
	}
	
	public RoomTile getTileInFront()
	{
		return this.getAdjacentTile(Direction.North);
	}
	
	public RoomTile getTileBehind()
	{
		return this.getAdjacentTile(Direction.South);
	}
	
	public RoomTile getAdjacentTile(Direction dir)
	{
		dir = this.bodyDirection.transform(dir);
		return this.getRoom().getMap().getTile(this.x + dir.modX, this.y + dir.modY);
	}
	
	public void setDirection(Direction dir)
	{
		boolean changes = (dir != this.bodyDirection);
		this.headDirection = dir;
		this.bodyDirection = dir;
		if(changes)
		{
			this.refresh(false);
		}
	}
	
	public void setGoal(RoomTile goal)
	{
		// Unidle the actor
		this.unidle();
		
		// Get rid of current path
		this.clearMovement();
		
		// Set the goal
		if(goal != null)
		{
			this.goal = goal;
			this.calculatePath = true;
			this.requiresUpdate = true;
		}
	}

	public boolean hasGoal()
	{
		return(this.goal != null);
	}
	
	public boolean isMoving()
	{
		return(this.next != null);
	}

	public void clearMovement()
	{
		this.goal = null;
		this.path.clear();
	}

	public boolean canMove()
	{
		return !this.moveLock;
	}

	public void lockMovement()
	{
		this.moveLock = true;
		this.clearMovement();
	}

	public void unlockMovement()
	{
		this.moveLock = false;
	}

	public void lookTo(RoomTile.TileObject obj, boolean headOnly)
	{
		this.lookTo(obj.getTile(), headOnly);
	}
	
	public void lookTo(RoomTile tile, boolean headOnly)
	{
		// Can't rotate in this context?
		if(tile == null || this.hasGoal() || tile.equals(this.getTile()) || this.statuses.containsKey("lay"))
		{
			return;
		}

		// Face head and body in same direction
		this.headDirection = this.bodyDirection;
		
		// Calculate facing direction to tile
		int curBodyDir = this.bodyDirection.num;
		Direction newBodyDir = Direction.calculate(this.x, this.y, tile.x, tile.y);
		int change = curBodyDir - newBodyDir.num;

		// Apply new rotation
		if(headOnly || this.statuses.containsKey("sit"))
		{
			if(curBodyDir == 0 || curBodyDir == 2 || curBodyDir == 4 || curBodyDir == 6)
			{
				if(change > 0)
				{
					this.headDirection = Direction.get(curBodyDir - 1);
				}
				else if(change < 0)
				{
					this.headDirection = Direction.get(curBodyDir + 1);
				}
			}
		}
		else
		{
			this.headDirection = newBodyDir;
			if(change <= -2 || change >= 2)
			{
				this.bodyDirection = newBodyDir;
			}
		}
		
		// Require update (direction changed)
		this.requiresUpdate = true;
	}

	public void chat(String message, boolean isShout)
	{
		// Unidle
		this.unidle();
		
		// Build servermessage
		int gesture = ChatUtil.getChatGesture(message);
		ServerMessage msg = isShout ? ShoutMessageComposer.compose(this, message, gesture) : ChatMessageComposer.compose(this, message, gesture);
		
		// This is a RoomUser? (real user)
		int myUserId = 0;
		if(this instanceof RoomUser)
		{	
			// Receive own message
			RoomUser user = (RoomUser)this;
			user.getSession().send(msg);
			
			// Record user ID for ignore-checks later
			myUserId = user.getInfo().id;
		}
		
		// Interact with actors
		for(RoomActor actor : this.room.getAllActors())
		{
			// Skip self (already handled)
			if(actor == this)
			{
				continue;
			}
			
			// The actor is a RoomUser too?
			if(actor instanceof RoomUser)
			{
				// Has he/she ignored this RoomUser?
				RoomUser user = (RoomUser)actor;
				if(user.getSession().hasIgnoredUser(myUserId))
				{
					// Skip!
					continue;
				}
				else
				{
					// Receive the message
					user.getSession().send(msg);
				}
			}
			
			// Extra stuff
			actor.onChatFromActor(this, message, isShout);
		}
	}

	protected abstract void onChatFromActor(RoomActor from, String text, boolean isShout);

	public void addStatus(String key)
	{
		this.addStatus(key, null);
	}

	public void addStatus(String key, String value)
	{
		this.statuses.put(key, value);
		this.requiresUpdate = true;
	}

	public void removeStatus(String key)
	{
		this.statuses.remove(key);
		this.requiresUpdate = true;
	}
	
	public void removeNonMoveStatuses()
	{
		this.removeStatus("sit");
		this.removeStatus("lay");
	}	
	
	public void reposition()
	{
		this.setTile(this.getTile(), true);
	}
	
	public void setTile(RoomTile tile, boolean refreshNow)
	{
		// Valid tile?
		if(tile == null)
		{
			return;
		}
		
		// Step off current tile
		this.getTile().stepOff(this);
		
		// Step on new tile
		tile.stepOn(this);
		
		// Delayed update?
		this.refresh(refreshNow);
	}
	
	public void follow(RoomActor actor)
	{
		this.follow(actor, Direction.South);
	}
	
	public void follow(RoomActor actor, Direction mode)
	{
		// Unfollow current actor?
		if(this.isFollowing())
		{
			this.unfollow();
		}
		
		// Follow new actor
		this.follow = actor;
		this.followMode = mode;
	}
	
	public void unfollow()
	{
		this.follow = null;
		this.followMode = null;
	}
	
	public boolean isFollowing()
	{
		return (this.getFollowing() != null);
	}
	
	public RoomActor getFollowing()
	{
		if(this.follow != null && this.follow.invalidated())
		{
			this.follow = null;
		}
		
		return this.follow;
	}
	
	public void refresh(boolean instant)
	{
		if(instant)
		{
			this.room.broadcast(UserUpdateMessageComposer.compose(this));
			this.requiresUpdate = false;
		}
		else
		{
			this.requiresUpdate = true;
		}
	}
	
	public void appendStatusString(ServerMessage msg)
	{
		if(this instanceof RoomPet)
		{
			this.headDirection = this.bodyDirection;
		}
		msg.appendInt(this.actorId);
		msg.appendInt(this.x);
		msg.appendInt(this.y);
		msg.appendStr(this.z + this.getExtraDisplayHeight());
		msg.appendInt(this.headDirection.num);
		msg.appendInt(this.bodyDirection.num);
		for(Entry<String, String> status : this.statuses.entrySet())
		{
			msg.append(status.getKey());
			if(status.getValue() != null)
			{
				msg.append(' ');
				msg.append(status.getValue());
			}
			msg.append('/');
		}
		if(this instanceof RoomPet)
		{
			if(this.getTile().is(RoomTile.WATER))
			{
				msg.append("swm");
				msg.append('/');
			}
		}
		msg.appendStr('/');
	}

	public abstract String getActorKey();
}
