package habbod2.game.rooms.instance;

import habbod2.game.rooms.items.floor.TagPoleItem;
import habbod2.game.rooms.minigames.GameUserPlayer;
import habbod2.game.sessions.Session;
import habbod2.game.users.User;
import habbod2.game.util.FigureUtil;
import habbod2.game.util.GameUtil;
import habbod2.communication.messages.outgoing.room.chat.FloodControlMessageComposer;
import habbod2.communication.messages.outgoing.room.engine.UserChangeMessageComposer;
import habbod2.communication.messages.outgoing.room.session.YouArePlayingGameMessageComposer;
import habbod2.communication.messages.outgoing.users.RespectNotificationMessageComposer;

public class RoomUser extends RoomHuman
{
	private static final int maxIdleAwakeTime = GameUtil.secondsToRoomPulses(5 * 60); // 5 minutes of idling before sleeping
	private static final int maxIdleSleepTime = (maxIdleAwakeTime + GameUtil.secondsToRoomPulses(10 * 60)); // 5 minutes idling -> sleep, 10 more minutes sleeping -> kick

	private Session session;
	public boolean hasRoomOwner;
	public boolean hasRoomRights;
	
	private long chatLastMessageTime;
	private int chatFastMessageCount;
	
	private String figureOverride;
	
	private GameUserPlayer player;
	
	private TagPoleItem tagPole;
	
	public boolean teleportEnabled;
	
	public RoomUser(Session session)
	{
		this.session = session;
		this.pathfindingAllowedOn = (RoomTile.OPEN | RoomTile.WATER);
	}
	
	@Override
	protected void onChatFromActor(RoomActor from, String text, boolean isShout)
	{
		this.lookTo(from, true);
	}
	
	@Override
	public void chat(String text, boolean isShout)
	{
		// Chat timeout in effect?
		long nowTime = System.currentTimeMillis();
		if(this.session.chatTimeoutExpireTime > 0)
		{
			// Calculate remaining seconds
			int chatTimeoutSeconds = (int) ((this.session.chatTimeoutExpireTime - nowTime) / 1000);
			if(chatTimeoutSeconds > 0)
			{
				// Notify client
				this.getSession().send(FloodControlMessageComposer.compose(chatTimeoutSeconds));
				
				// Block chat
				return;
			}
		}
		
		// Increment count if there is less than 1.5 seconds between chat messages
		long timeBetween = nowTime - this.chatLastMessageTime;
		if(timeBetween > 1500)
		{
			this.chatFastMessageCount = 0;
		}
		else
		{
			this.chatFastMessageCount++;
		}
		
		// Forward chat
		super.chat(text, isShout);
		this.chatLastMessageTime = nowTime;
		this.session.chatTimeoutExpireTime = 0;
		
		// Sent 3 or more fast messages in a row?
		if(this.chatFastMessageCount >= 3)
		{
			// Give 30 seconds timeout?
			if(this.getSession().getUser().hasRight("floodcontrol"))
			{
				this.session.chatTimeoutExpireTime = nowTime + (30 * 1000);
			}
		}
	}
	
	@Override
	public void pulse()
	{
		// RoomHuman pulse()
		super.pulse();
		
		// Not sleeping?
		if(!this.isAsleep)
		{
			// Needs to go to sleep?
			if(this.idleTime > maxIdleAwakeTime)
			{
				this.setAsleep(true);
			}
		}
		else
		{
			// Sleeping long enough for kicking?
			if(this.idleTime > maxIdleSleepTime)
			{
				this.getSession().kickFromRoom();
			}
		}
	}
	
	public void refreshAppearance()
	{
		this.getRoom().broadcast(UserChangeMessageComposer.compose(this, false));
	}
	
	public void receiveRespectFrom(RoomUser from)
	{
		if(from.getInfo().respectPoints > 0)
		{
			from.getInfo().respectPoints--;
		}
		this.getInfo().respect++;
		this.getRoom().broadcast(RespectNotificationMessageComposer.compose(this));
		this.refreshAppearance();
		from.wave();
	}
	
	public boolean isKickableBy(RoomUser other)
	{
		// Higher level?
		if(this.getInfo().role.level > other.getInfo().role.level)
		{
			return false;
		}
		else
		{
			// Need higher level to kick room owner
			if(this.hasRoomOwner && other.getInfo().role.level <= this.getInfo().role.level)
			{
				return false;
			}
			else
			{
				// Kickable!
				return true;
			}
		}
	}

	public Session getSession()
	{
		return this.session;
	}
	
	public User getInfo()
	{
		return this.session.getUser();
	}
	
	public String getFigure()
	{
		return FigureUtil.getFigure(this.getInfo().figure, this.figureOverride);
	}
	
	public String getFigureOverride()
	{
		return this.figureOverride;
	}
	
	public void setFigureOverride(String figureOverride)
	{
		this.figureOverride = figureOverride;
		this.refreshAppearance();
	}
	
	public boolean isPlayer()
	{
		return (this.player != null);
	}
	
	public GameUserPlayer getPlayer()
	{
		return this.player;
	}
	
	public void setPlayer(GameUserPlayer player)
	{
		// Dispose old?
		if(this.player != null)
		{
			GameUserPlayer tmp = this.player;
			this.player = null;
			tmp.getTeam().removePlayer(tmp);
		}
		
		// Set new reference
		this.player = player;
		
		// Notify session
		this.session.send(YouArePlayingGameMessageComposer.compose(this.player != null));
	}
	
	public boolean isTagger()
	{
		return (this.tagPole != null);
	}
	
	public void setTagger(TagPoleItem tagPole)
	{
		this.tagPole = tagPole;
		this.setEffect(EFFECT_GAME, tagPole != null ? tagPole.getEffectId() : 0);
	}
	
	public TagPoleItem getTagPole()
	{
		return this.tagPole;
	}
	
	@Override
	public String getActorKey()
	{
		return "u" + this.session.getUser().id;
	}
	
	public interface UserObserver
	{
		public void onUserEnter(RoomUser user);
		public void onUserLeave(RoomUser user);
	}
}
