package prinzip;

import java.io.File;
import java.io.FileInputStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.jongo.Jongo;

import prinzip.accounts.Account;
import prinzip.accounts.AccountManager;
import prinzip.apps.TwitterAppManager;
import prinzip.engine.ClockworkEngine;
import prinzip.util.Twitter4JUtil;
import prinzip.webapi.WebApiServer;

import com.mongodb.CommandResult;
import com.mongodb.MongoClient;

public class Server
{
	private static final Logger logger = Logger.getLogger(Server.class);
	
	private Jongo db;
	private TwitterAppManager appMgr;
	private AccountManager accountMgr;
	private ClockworkEngine engine;
	private WebApiServer webApiServer;
	
	public Server(File file) throws Exception
	{
		// Print network information
		InetAddress host;
		try
		{
			logger.info("looking up localhost...");
			host = InetAddress.getLocalHost();
		}
		catch(UnknownHostException e)
		{
			throw new Exception("cannot determine localhost", e);
		}

		// Load properties
		Properties props = new Properties();
		try
		{
			props.load(new FileInputStream(file));
		}
		catch(Exception e)
		{
			throw new Exception(String.format("cannot load %s", file), e);
		}
		logger.info(String.format("loaded %d entries from %s", props.size(), file));
		
		// Don't spam our logs...
		Twitter4JUtil.silenceLoggers();
		
		// Setup database
		try
		{
			// Try connecting...
			Jongo jongo = new Jongo(new MongoClient().getDB(props.getProperty("db.name")));
			CommandResult stats = jongo.getDatabase().getStats();
			
			// OK!
			logger.info(String.format("opened DB '%s' @ %s, size=%f MiB",
							stats.get("db"),
							stats.get("serverUsed"),
							stats.getDouble("dataSize") / 1024 / 1024));
			this.db = jongo;
		}
		catch(Exception e)
		{
			throw new Exception(String.format("cannot connect to database '%s'", props.getProperty("db.name")), e);
		}
		
		// Load apps
		this.appMgr = new TwitterAppManager(this);
		if(!this.appMgr.reloadApps())
		{
			throw new Exception("no apps configured");
		}
		
		// Apply configuration settings
		Account.applyConfiguration(props);
		this.accountMgr = new AccountManager(this);
		
		// Create scheduler
		this.engine = new ClockworkEngine();
		
		// Start web API server
		/*this.webApiServer = new WebApiServer(Integer.parseInt(props.getProperty("webapi.port")),
				props.getProperty("webapi.allowed.clients"), this);*/
		
		// Log current instance
		logger.info(String.format("running on %s (%s) %s", System.getProperty("os.name"), System.getProperty("os.arch"), System.getProperty("os.version")));
		logger.info(String.format("find me at %s (%s)", host.getHostAddress(), host.getHostName()));
		logger.info(String.format("this is instance '%s'", props.getProperty("instance.name")));
	}
	
	public Jongo getDB()
	{
		return this.db;
	}
	
	public TwitterAppManager getAppMgr()
	{
		return this.appMgr;
	}
	
	public AccountManager getAccountMgr()
	{
		return this.accountMgr;
	}
	
	public ClockworkEngine getEngine()
	{
		return this.engine;
	}
	
	public WebApiServer getWebApiServer()
	{
		return this.webApiServer;
	}
	
	public static class Referent
	{
		private final Server server;
		
		public Referent(Server server)
		{
			this.server = server;
		}
		
		public Server getServer()
		{
			return this.server;
		}
	}
}
