package prinzip.accounts;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.LinkedHashMap;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;
import org.bson.types.ObjectId;
import org.jongo.marshall.jackson.oid.Id;

import prinzip.Server;
import prinzip.accounts.actions.FindTargetsAction;
import prinzip.accounts.actions.FollowAction;
import prinzip.accounts.actions.FollowedBackCheckAction;
import prinzip.accounts.actions.TargetedAccountAction;
import prinzip.accounts.actions.UnfollowAction;
import prinzip.apps.TwitterApp;
import prinzip.util.TimeUtil;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import twitter4j.Twitter;
import twitter4j.TwitterException;
import twitter4j.User;
import twitter4j.auth.AccessToken;

public class Account
{
	private static final Logger logger = Logger.getLogger(Account.class);
	
	public static int TARGETCACHE_SIZE = 0;
	private static long[] FOLLOW_INTERVAL_TIME = {0,0};
	private static long[] UNFOLLOW_INTERVAL_TIME = {0,0};
	private static long[] FOLLOWCHECK_AFTER_TIME = {0,0};
	private static long[] FOLLOWCHECK_INTERVAL_TIME = {0,0};
	
	@Id
	public ObjectId id;
	
	public String username = "";
	public String password = "";
	//public String service = "";
	public String email;
	
	@JsonProperty("uid")
	public long userId;
	
	public int appId;
	
	@JsonProperty("token")
	public String authToken = "";
	
	@JsonProperty("secret")
	public String authSecret = "";
	
	@JsonProperty("keywords")
	public List<String> keywords;
	
	@JsonProperty("lang")
	public String searchLanguage = "";
	
	@JsonProperty("imports")
	public List<Long> importUsers;
	@JsonProperty("imptr")
	public int importPointer;
	
	@JsonProperty("follow")
	public Set<Long> toFollow;
	
	@JsonProperty("unfollow")
	public List<Long> toUnfollow;
	
	@JsonProperty("check")
	public LinkedHashMap<Long, Date> toFollowCheck;
	
	@JsonProperty("back")
	public Set<Long> followedBack;
	
	@JsonProperty("notback")
	public Set<Long> notFollowedBack;
	
	public AccountProblem problem;
	
	// Inner workings
	@JsonIgnore private Twitter twitter;
	@JsonIgnore private AccountManager manager;
	@JsonIgnore private Set<Long> pendingActions;
	@JsonIgnore private int scheduledFollows;
	@JsonIgnore private int scheduledUnfollows;
	@JsonIgnore private int scheduledChecks;
	
	public void save()
	{
		// TODO: Limit database saves
		this.manager.saveAccount(this);
	}
	
	public void initialize(AccountManager manager)
	{
		// Register account manager
		this.manager = manager;
		
		// Handle any missing/new fields
		if(this.keywords == null) this.keywords = new ArrayList<String>();
		if(this.importUsers == null) this.importUsers = new ArrayList<Long>();
		if(this.toFollow == null) this.toFollow = new HashSet<Long>();
		if(this.toUnfollow == null) this.toUnfollow = new ArrayList<Long>();
		if(this.toFollowCheck == null) this.toFollowCheck = new LinkedHashMap<Long, Date>();
		if(this.followedBack == null) this.followedBack = new HashSet<Long>();
		if(this.notFollowedBack == null) this.notFollowedBack = new HashSet<Long>();
		
		//CompatUpdate.update(this);
		
		// Log cache stats
		logger.info(String.format("initialized @%s (f=%d c=%d fb=%d nfb=%d u=%d kw=%d i=%d | r=%.2f)",
				this.username, this.toFollow.size(), this.toFollowCheck.size(),
				this.followedBack.size(), this.notFollowedBack.size(),
				this.toUnfollow.size(), this.keywords.size(), this.importUsers.size(),
				this.calculateFollowedBackRatio()));
	}
	
	public void kickstart()
	{
		// Kickstart any stopped loops
		this.pendingActions = new HashSet<Long>();
		this.nextFollowAction(null);
		this.nextUnfollowAction(null);
		this.nextCheckAction(null);
	}
	
	private void authenticate() throws TwitterException
	{
		// Get a (possibly new & authorized) app
		TwitterApp app = this.getAuthorizedApp();
		
		// Try logging in to the app
		Twitter twitter = app.getTwitter(this);
		User user = twitter.verifyCredentials();
		
		// Detect user ID
		if(this.userId <= 0)
		{
			this.userId = user.getId();
			this.manager.onUserIdDiscovery(this);
			this.save();
		}
		
		// Detect name changes
		if(!user.getScreenName().equals(this.username))
		{
			String oldName = this.username;
			this.username = user.getScreenName();
			this.manager.onUsernameChange(oldName, this.username);
			this.save();
		}
		
		// Logged in!
		this.twitter = twitter;
		logger.debug(String.format("authenticated @%s with app \"%s\"", this.username, app.name));
	}
	
	private TwitterApp getAuthorizedApp() throws TwitterException
	{
		// User has no app?
		TwitterApp app = this.getServer().getAppMgr().getApp(this.appId);
		if(app == null)
		{
			// User actually had an app?
			if(this.appId > 0)
			{
				if(logger.isDebugEnabled()) logger.debug(String.format("app #%d no longer exists!", this.appId));
			}
			app = this.selectNewApp();
		}
		
		return app;
	}
	
	public TwitterApp selectNewApp() throws TwitterException
	{
		TwitterApp newApp = this.getServer().getAppMgr().selectApp();
		if(newApp == null)
		{
			throw new TwitterException("no new app available");
		}
		else
		{
			// Attempt to authorize with this new app
			AccessToken token = newApp.authorize(this);
			
			// Store authorization details for further usage
			this.appId = newApp.appId;
			this.authToken = token.getToken();
			this.authSecret = token.getTokenSecret();
			logger.info(String.format("authorized @%s with app \"%s\"", this.username, newApp.name));
			
			// Dispose of client, need to reauthenticate
			this.twitter = null;
			
			// Save account data
			this.save();
		}
		
		return newApp;
	}
	
	public boolean needsTargets()
	{
		return (this.toFollow.size() < Account.TARGETCACHE_SIZE);
	}
	
	public boolean addFollowTarget(long targetId)
	{
		// Cannot follow self
		if(targetId == this.userId)
		{
			return false;
		}
		
		// Already waiting to follow?
		if(this.toFollow.contains(targetId))
		{
			return false;
		}
		
		// Already followed and waiting to check?
		if(this.toFollowCheck.containsKey(targetId))
		{
			return false;
		}
		
		// Already bothered with?
		if(this.followedBack.contains(targetId) || this.notFollowedBack.contains(targetId))
		{
			return false;
		}
		
		// Check additional info?
		/*
		if(target != null)
		{
			// Is the user speaking a different language?
			if(!this.searchLanguage.isEmpty() && !target.getLang().equals(this.searchLanguage))
			{
				return false;
			}
			
			// Is this a generic, boring user?
			if(target.getDescription().isEmpty() || target.getProfileImageURL().contains("default_profile"))
			{
				return false;
			}
		}*/
		
		// Perfect!
		this.toFollow.add(targetId);
		//if(logger.isDebugEnabled()) logger.debug(String.format("@%s will try to follow #%d", this.username, targetId));
		
		return true;
	}
	
	public void nextFollowAction(TargetedAccountAction previous)
	{
		// Remove from pending
		this.completeTargetedAction(previous);
		
		// Loop still going?
		if(!this.hasProblem() && this.scheduledFollows <= 0)
		{
			// Follow cache empty?
			Long targetId = this.nextActionTargetId(this.toFollow);
			if(targetId == null)
			{
				this.getServer().getEngine().scheduleAfter(10, TimeUnit.MILLISECONDS, new FindTargetsAction(this));
			}
			else
			{
				// Determine a delay and schedule the task
				long followDelay = TimeUtil.selectRandomTime(FOLLOW_INTERVAL_TIME);
				this.getServer().getEngine().scheduleAfter(followDelay, TimeUnit.MILLISECONDS, this.scheduleTargetedAction(new FollowAction(this, targetId)));
			}
		}
	}
	
	public void nextUnfollowAction(TargetedAccountAction previous)
	{
		// Remove from pending
		this.completeTargetedAction(previous);
		
		// Loop still going?
		if(!this.hasProblem() && this.scheduledUnfollows <= 0)
		{
			// Anyone needs infollowing?
			Long targetId = this.nextActionTargetId(this.toUnfollow);
			if(targetId != null)
			{
				long unfollowDelay = TimeUtil.selectRandomTime(UNFOLLOW_INTERVAL_TIME);
				this.getServer().getEngine().scheduleAfter(unfollowDelay, TimeUnit.MILLISECONDS, this.scheduleTargetedAction(new UnfollowAction(this, targetId)));
			}
		}
	}
	
	public void nextCheckAction(TargetedAccountAction previous)
	{
		// Remove from pending
		this.completeTargetedAction(previous);
		
		// Loop still going?
		if(!this.hasProblem() && this.scheduledChecks <= 0)
		{
			// Anyone needs checking?
			Long targetId = this.nextActionTargetId(this.toFollowCheck.entrySet());
			if(targetId != null)
			{
				// Determine a delay and schedule the check (prevent multiple checks from executing too fast after each other)
				Date followedAt = this.toFollowCheck.get(targetId);
				long checkTime = (followedAt.getTime() + TimeUtil.selectRandomTime(FOLLOWCHECK_AFTER_TIME));
				long checkDelay = Math.max((checkTime - System.currentTimeMillis()), TimeUtil.selectRandomTime(FOLLOWCHECK_INTERVAL_TIME));
				this.getServer().getEngine().scheduleAfter(checkDelay, TimeUnit.MILLISECONDS, this.scheduleTargetedAction(new FollowedBackCheckAction(this, targetId)));
			}
		}
	}
	
	private Long nextActionTargetId(Iterable<?> iterable)
	{
		for(Object item : iterable)
		{
			Long targetId = (Long) ((item instanceof Long) ? item : (((Entry<?, ?>)item).getKey()));
			if(!this.pendingActions.contains(targetId))
			{
				return targetId;
			}
		}
		
		return null;
	}
	
	public TargetedAccountAction scheduleTargetedAction(TargetedAccountAction action)
	{
		this.pendingActions.add(action.getTargetId());
		if(action instanceof FollowAction)
		{
			this.scheduledFollows++;
		}
		else if(action instanceof UnfollowAction)
		{
			this.scheduledUnfollows++;
		}
		else if(action instanceof FollowedBackCheckAction)
		{
			this.scheduledChecks++;
		}
		//if(logger.isDebugEnabled()) logger.debug("scheduled " + action);
		
		return action;
	}
	
	public void completeTargetedAction(TargetedAccountAction action)
	{
		if(action != null)
		{
			this.pendingActions.remove(action.getTargetId());
			if(action instanceof FollowAction)
			{
				this.scheduledFollows--;
			}
			else if(action instanceof UnfollowAction)
			{
				this.scheduledUnfollows--;
			}
			else if(action instanceof FollowedBackCheckAction)
			{
				this.scheduledChecks--;
			}
			//if(logger.isDebugEnabled()) logger.debug("completed " + action);
		}
	}
	
	public boolean hasProblem()
	{
		return (this.problem != null);
	}
	
	public void clearProblem()
	{
		// Problem is gone... (atleast for now!)
		this.problem = null;
		this.save();
		
		// Kickstart the loops!
		this.kickstart();
	}
	
	public void reportProblem(AccountProblem problem)
	{
		this.problem = problem;
		this.save();
		
		logger.warn(String.format("@%s has %s problem", this.username, problem));
		
		// TODO: send mail to account's contact email
	}
	
	public AccountManager getManager()
	{
		return this.manager;
	}
	
	public Server getServer()
	{
		return this.manager.getServer();
	}
	
	public synchronized Twitter getTwitter() throws TwitterException
	{
		if(this.twitter == null)
		{
			this.authenticate();
		}
		
		return this.twitter;
	}
	
	@JsonIgnore
	public boolean isAuthenticated()
	{
		return (this.twitter != null);
	}
	
	public List<String> nextSearchKeywords()
	{
		// Make a random selection
		List<String> keywords = new ArrayList<String>(this.keywords);
		Collections.shuffle(keywords);
		
		return keywords;
	}
	
	public long nextImportUser()
	{
		int available = this.importUsers.size();
		if(available > 0)
		{
			if(this.importPointer == available)
			{
				this.importPointer = 0;
			}
			return this.importUsers.get(this.importPointer++);
		}
		else
		{
			this.importPointer = 0;
			return -1;
		}
	}
	
	public float calculateFollowedBackRatio()
	{
		long followedBack = this.followedBack.size();
		if(followedBack > 0)
		{
			long totalFollowed = (this.followedBack.size() + this.notFollowedBack.size());
			return (float)followedBack / (float)totalFollowed;
		}
		else
		{
			return 0.0f;
		}
	}
	
	public static void applyConfiguration(Properties props)
	{
		// Desired amount of targets to cache after a search
		TARGETCACHE_SIZE = Integer.parseInt(props.getProperty("account.targetcache.size", "30"));
		
		// The range of time between two follow() actions for a account
		FOLLOW_INTERVAL_TIME = TimeUtil.parseTimePeriod(props.getProperty("account.follow.interval", "30m..60m"));
		
		// The range of time between two unfollow() actions for a account
		UNFOLLOW_INTERVAL_TIME = TimeUtil.parseTimePeriod(props.getProperty("account.unfollow.interval", "20s..30s"));
		
		// The range of time before a followed user is checked for a follow back
		FOLLOWCHECK_AFTER_TIME = TimeUtil.parseTimePeriod(props.getProperty("account.followcheck.after", "3d..3d"));
		
		// The range of time between two checks
		FOLLOWCHECK_INTERVAL_TIME = TimeUtil.parseTimePeriod(props.getProperty("account.followcheck.interval", "20s..30s"));
		
		logger.info(String.format("followcache size=%d, follow interval=%s, unfollow interval=%s, follow_check after %s (interval = %s)",
				TARGETCACHE_SIZE,
				props.getProperty("account.follow.interval"),
				props.getProperty("account.unfollow.interval"),
				props.getProperty("account.followcheck.after"),
				props.getProperty("account.followcheck.interval")));
	}
}
