package prinzip.accounts;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jongo.MongoCollection;

import org.apache.log4j.Logger;

import prinzip.Server;
import twitter4j.Twitter;

public class AccountManager extends Server.Referent
{
	private static final Logger logger = Logger.getLogger(AccountManager.class);

	private MongoCollection dbAccounts = this.getServer().getDB().getCollection("accounts");

	private Map<Long, Account> accountsByUserId = new HashMap<Long, Account>();
	private Map<String, Account> accountsByUsername = new HashMap<String, Account>();

	public AccountManager(Server server)
	{
		super(server);
		this.dbAccounts.ensureIndex("{username:1}", "{unique:true}");
	}

	public void saveAccount(Account account)
	{
		// Save the ENTIRE document in the database and measure time taken
		// long startTime = System.currentTimeMillis();
		this.dbAccounts.save(account);
		// long endTime = System.currentTimeMillis();

		// TODO: add database performance monitoring?
		// if(logger.isDebugEnabled())
		// logger.debug(String.format("saved @%s in %d milliseconds",
		// account.username, (endTime - startTime)));
	}

	public Account addAccount(String username, String password, String searchLanguage, List<String> keywords, List<Long> importUsers)
	{
		// Already exists?
		if(this.getAccountByUsername(username) != null)
		{
			logger.warn(String.format("cannot add user %s, already exists", username));
			return null;
		}

		// Create new user
		Account account = new Account();
		account.username = username;
		account.password = password;
		account.searchLanguage = searchLanguage;
		account.keywords = keywords;
		account.importUsers = importUsers;
		this.saveAccount(account);

		if(logger.isDebugEnabled())
			logger.debug(String.format("added new user: @%s", account.username));

		// Join circus
		this.joinAccount(account);

		return account;
	}

	public void initializeAccounts()
	{
		// Count the amount of accounts in the database
		long count = this.dbAccounts.count();

		// Initialize all accounts (TODO: load accounts from specified 'clients'
		// only?)
		logger.debug(String.format("initializing %d accounts from DB...", count));
		for(Account account : this.dbAccounts.find().as(Account.class))
		{
			this.joinAccount(account);
		}
	}

	private void joinAccount(Account account)
	{
		// Account has an ID already?
		if(account.userId > 0)
		{
			this.accountsByUserId.put(account.userId, account);
		}
		this.accountsByUsername.put(account.username.toLowerCase(), account);

		// Initialize data
		account.initialize(this);

		// Resume normal operations?
		if(account.hasProblem())
		{
			logger.warn(String.format("@%s has %s problem, not resuming", account.username, account.problem));
		}
		else
		{
			account.kickstart();
		}
	}

	public void onUserIdDiscovery(Account account)
	{
		// Account actually has an ID?
		if(account.userId > 0)
		{
			this.accountsByUserId.put(account.userId, account);
		}
	}

	public void onUsernameChange(String oldName, String newName)
	{
		Account account = this.accountsByUsername.remove(oldName.toLowerCase());
		if(account != null)
		{
			this.accountsByUsername.put(newName.toLowerCase(), account);
			logger.info(String.format("@%s is now named @%s", oldName, newName));
		}
	}

	public Account getAccountById(long userId)
	{
		return this.accountsByUserId.get(userId);
	}

	public Account getAccountByUsername(String username)
	{
		return this.accountsByUsername.get(username.toLowerCase());
	}
	
	public Twitter getUtilityAccount()
	{
		for(Account account : this.accountsByUserId.values())
		{
			if(account.isAuthenticated())
			{
				try
				{
					return account.getTwitter();
				}
				catch(Exception ex)
				{
					ex.printStackTrace();
				}
			}
		}

		return null;
	}
}
