package prinzip.webapi;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.simpleframework.http.Query;
import org.simpleframework.http.Request;
import org.simpleframework.http.Response;
import org.simpleframework.http.Status;
import org.simpleframework.http.core.Container;
import org.simpleframework.http.core.ContainerServer;
import org.simpleframework.transport.connect.Connection;
import org.simpleframework.transport.connect.SocketConnection;

import prinzip.Server;
import prinzip.accounts.Account;

public class WebApiServer extends Server.Referent implements Container
{	
	private static final Logger logger = Logger.getLogger(WebApiServer.class);
	
	private static final String AUTH_CLIENT_HEADER = "X-Prinzip-Client";
	private static final String AUTH_KEY_HEADER = "X-Prinzip-Key";
	private static final boolean ALLOW_QUERYSTRING_AUTHORIZATION = true;
	
	private Map<String, String> allowedClients = new HashMap<String, String>();
	
	public WebApiServer(int port, String allowedClients, Server server) throws IOException
	{
		super(server);
		
		// Parse allowed clients
		String[] clients = allowedClients.split(",");
		for(String client : clients)
		{
			String[] bits = client.split(":");
			this.allowedClients.put(bits[0], bits[1]);
			
			logger.info(String.format("allow \"%s\"", bits[0]));
		}
		
		// Create the actual server, act like a container too
		ContainerServer httpd = new ContainerServer(this);
		Connection connection = new SocketConnection(httpd);
		SocketAddress address = new InetSocketAddress(port);
		connection.connect(address);
		
		logger.info(String.format("started on %s", address));
	}
	
	@Override
	public void handle(Request req, Response res)
	{
		try
		{
			if(this.isAuthorized(req))
			{
				this.handleAuthorizedRequest(req, res);
				if(logger.isDebugEnabled()) logger.debug(String.format("handled authorized request \"%s\" from %s (%s)", req.getPath(), req.getQuery().get("c"), req.getClientAddress().getHostString()));
			}
			else
			{
				res.setStatus(Status.UNAUTHORIZED);
				logger.warn(String.format("denied unauthorized request \"%s\" from %s (%s)", req.getPath(), req.getClientAddress().getHostString(), req.getClientAddress().getHostName()));
			}
			res.close();
		}
		catch(Exception e)
		{
			logger.error(String.format("error processing request \"%s\"", req.getPath()), e);
		}
	}
	
	private boolean isAuthorized(Request req)
	{
		// For testing purposes, also allow authentication in query string
		if(ALLOW_QUERYSTRING_AUTHORIZATION)
		{
			Query query = req.getQuery();
			if(query.containsKey("c") && query.containsKey("k"))
			{
				return query.get("k").equals(this.allowedClients.get(query.get("c")));
			}
		}
		
		// Specified client and key combination must be allowed serverside
		String client = req.getValue(AUTH_CLIENT_HEADER);
		String key = req.getValue(AUTH_KEY_HEADER);
		return (client != null && key != null && key.equals(this.allowedClients.get(client)));
	}
	
	private void handleAuthorizedRequest(Request req, Response res) throws Exception
	{
		String[] segments = req.getPath().getSegments();
		
		String resource = segments[0];
		String name = segments[1];
		
		switch(resource)
		{
			case "account":
			{
				Account ac = this.getServer().getDB().getCollection("accounts").findOne("{username:#}", name).as(Account.class);
				new AccountRequestHandler(ac, this.getServer().getAccountMgr()).handleRequest(req, res);
			}
		}
	}
}
